/**
Action Script library. Accessibility class
Copyright (C) 2005 Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_accessibility.hh"

typedef void (ASAccessibility::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASAccessibility::init,					// 0
	&ASAccessibility::isActive,				// 1
	&ASAccessibility::updateProperties,		// 2
	// Need to be at the end.
	&ASAccessibility::addProperty,		
	&ASAccessibility::registerClass,	
	&ASAccessibility::AStoString,		
	&ASAccessibility::unwatch,			
	&ASAccessibility::valueOf,			
	&ASAccessibility::watch			
};

/**
 * The constructor
 **/
ASAccessibility::ASAccessibility(Script *s, Graphics *g) : ASObject(s,g,L"Accessibility") {
}

ASAccessibility::~ASAccessibility() {
	//puts("~ASAccessibility");
}

/**
 * Return a function descriptor
 */
Function *ASAccessibility::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"Accessibility") 	{func->setFunctionIndex(0); func->setNumParams(0); }
	if (memberName == L"isActive") 			{func->setFunctionIndex(1); func->setNumParams(0); }
	if (memberName == L"updateProperties") 	{func->setFunctionIndex(2); func->setNumParams(0); }
	func = finishFunctionSetup(memberName,3,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASAccessibility::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}


/**
 * No instanciation on this class
 **/
void ASAccessibility::init() {
	wcout << L"Accessibility should not be instanciated!. Use static functions\n";
}

/**
 * A Boolean value of true if there are active Microsoft Active Accessibility (MSAA) clients and the player is running in an environment 
 * that supports communication between Flash Player and accessibility aids, false otherwise.
 **/
void ASAccessibility::isActive() {
	script->push(new BooleanObject(false));
}

/**
 * Does nothing on a system that has no MSAA capabilities
 **/
void ASAccessibility::updateProperties() {
	noReturnValue();
}
