/**
Action Script library. Boolean class implementation.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_boolean.hh"

typedef void (ASBoolean::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASBoolean::init,			// 0
	&ASBoolean::valueOf,		// 1
	// Need to be at the end.
	&ASBoolean::addProperty,		
	&ASBoolean::registerClass,	
	&ASBoolean::AStoString,		
	&ASBoolean::unwatch,			
	&ASBoolean::valueOf,			
	&ASBoolean::watch			
};

/**
 * The constructor
 **/
ASBoolean::ASBoolean(Script *s, Graphics *g) : ASObject(s,g,L"Boolean") {
}
ASBoolean::ASBoolean(Script *s, Graphics *g, bool value) : ASObject(s,g,L"Boolean") {
	this->value = value;
}

/**
 * The destructor
 **/
ASBoolean::~ASBoolean() {
}

/**
 * Return a function descriptor
 */
Function *ASBoolean::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"Boolean")	{func->setFunctionIndex(0); func->setNumParams(1); }
	if (memberName == L"valueOf") 	{func->setFunctionIndex(1); func->setNumParams(0); }
	finishFunctionSetup(memberName,2,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASBoolean::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 * Init a boolean as in JavaScript
 **/
void ASBoolean::init() {
	if (numPassedParams == 0) {
		value = false;
	} else {
		if (script->getRegister(1)->getObjectType() == StringObjectType) {
			value = script->getRegister(1)->toString().size() != 0;
		} else {
			value = (script->getRegister(1)->toNumeric() != 0.0);
		}
	}
}

/**
 *  returns the string representation ("true” or "false”) of the Boolean object.
 **/
void ASBoolean::AStoString() {
	script->push(new StringObject(value ? L"true" : L"false"));
}

/**
 * returns true if the primitive value type of the specified Boolean object is true, false if it is false.
 **/
void ASBoolean::valueOf() {
	script->push(new BooleanObject(value));
}

