/**
Action Script library. MovieCLip class implementation
Copyright (C) 2005 Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "../lib/localFonts.hh"
#include "as_movie_clip.hh"
#include "as_text_field.hh"
#include "as_array.hh"
#include "as_sprite.hh"

typedef void (ASMovieClip::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASMovieClip::attachAudio,				// 0
	&ASMovieClip::attachMovie,				// 1 
	&ASMovieClip::createEmptyMovieClip,		// 2 
	&ASMovieClip::createTextField,			// 3
	&ASMovieClip::duplicateMovieClip,		// 4
	&ASMovieClip::ASgetBounds,				// 5
	&ASMovieClip::getBytesLoaded,			// 6
	&ASMovieClip::getBytesTotal,			// 7
	&ASMovieClip::ASgetDepth,				// 8
	&ASMovieClip::getInstanceAtDepth,		// 9
	&ASMovieClip::getNextHighestDepth,		// 10
	&ASMovieClip::getSWFVersion,			// 11
	&ASMovieClip::getTextSnapshot,			// 12
	&ASMovieClip::getURL,					// 13
	&ASMovieClip::globalToLocal,			// 14
	&ASMovieClip::gotoAndPlay,				// 15
	&ASMovieClip::gotoAndStop,				// 16
	&ASMovieClip::hitTest,					// 17
	&ASMovieClip::loadMovie,				// 18
	&ASMovieClip::loadVariables,			// 19
	&ASMovieClip::localToGlobal,			// 20
	&ASMovieClip::nextFrame,				// 21
	&ASMovieClip::play,						// 22
	&ASMovieClip::prevFrame,				// 23
	&ASMovieClip::removeMovieClip,			// 24
	&ASMovieClip::setMask,					// 25
	&ASMovieClip::startDrag,				// 26
	&ASMovieClip::stop,						// 27
	&ASMovieClip::stopDrag,					// 28
	&ASMovieClip::swapDepths,				// 29
	&ASMovieClip::unloadMovie,				// 30
	&ASMovieClip::beginFill,				// 31
	&ASMovieClip::beginGradientFill,		// 32
	&ASMovieClip::clear,					// 33
	&ASMovieClip::curveTo,					// 34
	&ASMovieClip::endFill,					// 35
	&ASMovieClip::lineStyle,				// 36
	&ASMovieClip::lineTo,					// 37
	&ASMovieClip::moveTo,					// 38
	&ASMovieClip::init,						// 39
	// Events
	&ASMovieClip::onData,					// 40
	&ASMovieClip::onDragOut,				// 41
	&ASMovieClip::onDragOver,				// 42
	&ASMovieClip::onEnterFrame,				// 43
	&ASMovieClip::onKeyDown,				// 44
	&ASMovieClip::onKeyUp,					// 45
	&ASMovieClip::onKillFocus,				// 46
	&ASMovieClip::onLoad,					// 47
	&ASMovieClip::onMouseDown,				// 48
	&ASMovieClip::onMouseMove,				// 49
	&ASMovieClip::onMouseUp,				// 50
	&ASMovieClip::onPress,					// 51
	&ASMovieClip::onRelease,				// 52
	&ASMovieClip::onReleaseOutside,			// 53
	&ASMovieClip::onRollOut,				// 54
	&ASMovieClip::onRollOver,				// 55
	&ASMovieClip::onSetFocus,				// 56
	&ASMovieClip::onUnload,					// 57
	// Need to be at the end.
	&ASMovieClip::addProperty,		
	&ASMovieClip::registerClass,	
	&ASMovieClip::AStoString,		
	&ASMovieClip::unwatch,			
	&ASMovieClip::valueOf,			
	&ASMovieClip::watch
};
/**
 * Return a function descriptor
 */
Function *ASMovieClip::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"attachAudio") 		{func->setFunctionIndex(0); func->setNumParams(1); }
	if (memberName == L"attachMovie") 		{func->setFunctionIndex(1); func->setNumParams(4); }
	if (memberName == L"createEmptyMovieClip"){func->setFunctionIndex(2); func->setNumParams(2); }
	if (memberName == L"createTextField") 	{func->setFunctionIndex(3); func->setNumParams(6); }
	if (memberName == L"duplicateMovieClip"){func->setFunctionIndex(4); func->setNumParams(3); }
	if (memberName == L"getBounds") 		{func->setFunctionIndex(5); func->setNumParams(1); }
	if (memberName == L"getBytesLoaded") 	{func->setFunctionIndex(6); func->setNumParams(0); }
	if (memberName == L"getBytesTotal") 	{func->setFunctionIndex(7); func->setNumParams(0); }
	if (memberName == L"getDepth") 			{func->setFunctionIndex(8); func->setNumParams(0); }
	if (memberName == L"getInstanceAtDepth"){func->setFunctionIndex(9); func->setNumParams(1); }
	if (memberName == L"getNextHighestDepth"){func->setFunctionIndex(10); func->setNumParams(0); }
	if (memberName == L"getSWFVersion") 	{func->setFunctionIndex(11); func->setNumParams(0); }
	if (memberName == L"getTextSnapshot") 	{func->setFunctionIndex(12); func->setNumParams(0); }
	if (memberName == L"getURL") 			{func->setFunctionIndex(13); func->setNumParams(3); }
	if (memberName == L"globalToLocal") 	{func->setFunctionIndex(14); func->setNumParams(1); }
	if (memberName == L"gotoAndPlay") 		{func->setFunctionIndex(15); func->setNumParams(1); }
	if (memberName == L"gotoAndStop") 		{func->setFunctionIndex(16); func->setNumParams(1); }
	if (memberName == L"hitTest") 			{func->setFunctionIndex(17); func->setNumParams(3); }
	if (memberName == L"loadMovie") 		{func->setFunctionIndex(18); func->setNumParams(2); }
	if (memberName == L"loadVariables") 	{func->setFunctionIndex(19); func->setNumParams(2); }
	if (memberName == L"localToGlobal") 	{func->setFunctionIndex(20); func->setNumParams(1); }
	if (memberName == L"nextFrame") 		{func->setFunctionIndex(21); func->setNumParams(0); }
	if (memberName == L"play") 				{func->setFunctionIndex(22); func->setNumParams(0); }
	if (memberName == L"prevFrame") 		{func->setFunctionIndex(23); func->setNumParams(0); }
	if (memberName == L"removeMovieClip") 	{func->setFunctionIndex(24); func->setNumParams(0); }
	if (memberName == L"setMask") 			{func->setFunctionIndex(25); func->setNumParams(1); }
	if (memberName == L"startDrag") 		{func->setFunctionIndex(26); func->setNumParams(5); }
	if (memberName == L"stop") 				{func->setFunctionIndex(27); func->setNumParams(0); }
	if (memberName == L"stopDrag") 			{func->setFunctionIndex(28); func->setNumParams(0); }
	if (memberName == L"swapDepths") 		{func->setFunctionIndex(29); func->setNumParams(1); }
	if (memberName == L"unloadMovie") 		{func->setFunctionIndex(30); func->setNumParams(0); }
	if (memberName == L"beginFill") 		{func->setFunctionIndex(31); func->setNumParams(2); }
	if (memberName == L"beginGradientFill") {func->setFunctionIndex(32); func->setNumParams(5); }
	if (memberName == L"clear") 			{func->setFunctionIndex(33); func->setNumParams(0); }
	if (memberName == L"curveTo") 			{func->setFunctionIndex(34); func->setNumParams(4); }
	if (memberName == L"endFill") 			{func->setFunctionIndex(35); func->setNumParams(0); }
	if (memberName == L"lineStyle") 		{func->setFunctionIndex(36); func->setNumParams(3); }
	if (memberName == L"lineTo") 			{func->setFunctionIndex(37); func->setNumParams(2); }
	if (memberName == L"moveTo") 			{func->setFunctionIndex(38); func->setNumParams(2); }
	if (memberName == L"MovieClip") 		{func->setFunctionIndex(39); func->setNumParams(0); }
	if (memberName == L"onData") 			{func->setFunctionIndex(40); func->setNumParams(1); }
	if (memberName == L"onDragOut") 		{func->setFunctionIndex(41); func->setNumParams(1); }
	if (memberName == L"onDragOver") 		{func->setFunctionIndex(42); func->setNumParams(1); }
	if (memberName == L"onEnterFrame") 		{func->setFunctionIndex(43); func->setNumParams(1); }
	if (memberName == L"onKeyDown") 		{func->setFunctionIndex(44); func->setNumParams(1); }
	if (memberName == L"onKeyUp")	 		{func->setFunctionIndex(45); func->setNumParams(1); }
	if (memberName == L"onKillFocus") 		{func->setFunctionIndex(46); func->setNumParams(1); }
	if (memberName == L"onLoad") 			{func->setFunctionIndex(47); func->setNumParams(1); }
	if (memberName == L"onMouseDown") 		{func->setFunctionIndex(48); func->setNumParams(1); }
	if (memberName == L"onMouseMove") 		{func->setFunctionIndex(49); func->setNumParams(1); }
	if (memberName == L"onMouseUp") 		{func->setFunctionIndex(50); func->setNumParams(1); }
	if (memberName == L"onPress") 			{func->setFunctionIndex(51); func->setNumParams(1); }
	if (memberName == L"onRelease") 		{func->setFunctionIndex(52); func->setNumParams(1); }
	if (memberName == L"onReleaseOutside")	{func->setFunctionIndex(53); func->setNumParams(1); }
	if (memberName == L"onRollOut") 		{func->setFunctionIndex(54); func->setNumParams(1); }
	if (memberName == L"onRollOver") 		{func->setFunctionIndex(55); func->setNumParams(1); }
	if (memberName == L"onSetFocus") 		{func->setFunctionIndex(56); func->setNumParams(1); }
	if (memberName == L"onUnload") 			{func->setFunctionIndex(57); func->setNumParams(1); }
	func = finishFunctionSetup(memberName,58,func);
	return func;
}

/**
 * The constructor
 **/
void ASMovieClip::commonInit() {
	shape = new ShapeWithStyle();
	shape->setNumFillStyles(0);
	shape->setNumLineStyles(0);
	needRedraw = false;
	lastX=lastY= 0;
	isMorphShape = false;
	setMemberNoClone(L"_visible", new BooleanObject(true));
}

ASMovieClip::ASMovieClip(Script *s, Graphics *g) : ASObject(s,g,L"MovieClip") {
	commonInit();
}

ASMovieClip::ASMovieClip(Script *s, Graphics *g, wstring className) : ASObject(s,g,className) {
	commonInit();
}

ASMovieClip::ASMovieClip(ASMovieClip *p) : ASObject(p) {
}

/**
 * Deep copy constructor
 **/
ASInterface* ASMovieClip::getObjectDeepCopy(bool deep) {
	//printf("Clone ASMovieClip::getObjectDeepCopy %x\n",this);
	ASMovieClip *newObj = new ASMovieClip(this);
	newObj->shape= new ShapeWithStyle(shape);
	newObj->lastX=lastX;
	newObj->lastY=lastY;
	newObj->isMorphShape=isMorphShape;
	newObj->ratio=ratio;
	newObj->morphShape=morphShape;
	return newObj;
}

/**
 * The destructor
 **/
ASMovieClip::~ASMovieClip() {
	//printf("~ASMovieClip");
	delete shape;
	shape = NULL;
}

/**
 * Call class built in functions
 * exotic c++ call, but fast
 **/
void ASMovieClip::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 *
 **/
void ASMovieClip::init() {
	printf("ASMovieClip::init");
}

/**
 *  Captures and plays local audio from the microphone hardware.
 **/
void ASMovieClip::attachAudio() {
	puts("TODO - ASMovieClip::attachAudio()");
	noReturnValue();
}

/**
 * Attaches a SWF file in the library.
 **/
void ASMovieClip::attachMovie() {
	// Overloaded in Sprite - should never be called.
}

/**
 * Method; creates an empty movie clip as a child of an existing movie clip.
 **/
void ASMovieClip::createEmptyMovieClip() {
	// Overloaded in Sprite - should never be called.
}

/**
 * Dynamically create a new Text Field
 **/
void ASMovieClip::createTextField() {
	ASTextField *atf = new ASTextField(script,graphics);
	wstring instanceName = script->getRegister(1)->toString();
	int  depth = int(script->getRegister(2)->toNumeric());
	SI32 x = SI32(script->getRegister(3)->toNumeric());
	SI32 y = SI32(script->getRegister(4)->toNumeric());
	int  width  = int(script->getRegister(5)->toNumeric());
	int  height = int(script->getRegister(6)->toNumeric());
	RECT rect;
	rect.xmin = int(x*20.0); rect.ymin = int(y*20.0); rect.xmax = int((x+width)*20.0); rect.ymax = int((y + height)*20.0);
	atf->setBounds(&rect);
	getObjectStore()->setGlobalVariable(instanceName,atf);
	script->getObjectStore()->setObjectAtDepth(depth,atf);
	LocalFontManager *lfm = script->getGraphics()->getLocalFontManager();
	lfm->loadDefaultFont();
	atf->getDirectTextFormat()->setMember(L"size",new IntegerObject(280)); // TODO: call FontManager to get that value
	atf->setMemberNoClone(L"_alpha",new IntegerObject(0xff));
	script->push(atf);
}

/**
 * Duplicates the specified movie clip.
 **/
void ASMovieClip::duplicateMovieClip() {
	if (numPassedParams > 1 ) {
		wstring name = script->getRegister(1)->toString();
		int depth = int(script->getRegister(2)->toNumeric());
		ASInterface *as = (ASInterface*)script->getCurrentContext();
		ASMovieClip *newAs = (ASMovieClip *)as->getObjectDeepCopy(true);
		if (numPassedParams == 3 ) {
			ASObject *initObject = (ASObject *)script->getRegister(3);
			StringValueObjectHashMap *members = initObject->getMembers()->getMembers();
			for (StringValueObjectHashMap::const_iterator i=members->begin(); i!=members->end(); ++i)	newAs->setMember(i->first,i->second);
		}
		newAs->setMovieName(name);
		getObjectStore()->setGlobalVariable(name,newAs);
		script->getObjectStore()->setObjectAtDepth(depth,newAs);
		script->push(newAs);
	} else {
		noReturnValue();
	}
}

/**
 * Returns the minimum and maximum x and y coordinates of a SWF file in a specified coordinate space.
 **/
void ASMovieClip::ASgetBounds() {
	computeBounds();
	if (numPassedParams >0 ) { 
		ASObject* targetCoordinateSpace = (ASObject*)script->getRegister(1);
		// TODO: 
		targetCoordinateSpace = NULL; // Clean compile
	}
	ASObject* aos = new ASObject(script,graphics);
	aos->setMemberNoClone(L"xMin",new FloatObject(bounds.xmin));
	aos->setMemberNoClone(L"xMax",new FloatObject(bounds.xmax));
	aos->setMemberNoClone(L"yMin",new FloatObject(bounds.ymin));
	aos->setMemberNoClone(L"yMax",new FloatObject(bounds.ymax));
	script->push(aos);
}

/**
 * Returns the number of bytes loaded for the specified movie clip.
 **/
void ASMovieClip::getBytesLoaded() {
	puts("TODO - ASMovieClip::getBytesLoaded()");
	noReturnValue();
}

/**
 * Returns the size of the movie clip, in bytes.
 **/
void ASMovieClip::getBytesTotal() {
	puts("TODO - ASMovieClip::getBytesTotal()");
	noReturnValue();
}

/**
 * Returns the depth of a movie clip.
 **/
void ASMovieClip::ASgetDepth() {
	puts("TODO - ASMovieClip::ASgetDepth()");
	noReturnValue();
}

/**
 * Specifies whether a particular depth is already occupied by a movie clip.
 **/
void ASMovieClip::getInstanceAtDepth() {
	puts("TODO - ASMovieClip::getInstanceAtDepth()");
	noReturnValue();
}

/**
 * Specifies a depth value that you can pass to other methods to to ensure that Flash renders the movie clip in front of all
 * other objects in the current movie clip.
 **/
void ASMovieClip::getNextHighestDepth() {
	puts("TODO - ASMovieClip::getNextHighestDepth()");
	noReturnValue();
}

/**
 * An integer that specifies the Flash Player version that was targeted when the SWF file loaded into my_mc was published.
 **/
void ASMovieClip::getSWFVersion() {
	// ????????????????
	script->push(new IntegerObject(-1));
}

/**
 * Returns a TextSnapshot object that contains the text in the static text fields in the specified movie clip.
 **/
void ASMovieClip::getTextSnapshot() {
	puts("TODO - ASMovieClip::getTextSnapshot()");
	noReturnValue();
}

/**
 * Retrieves a document from a URL.
 **/
void ASMovieClip::getURL() {
	puts("TODO - ASMovieClip::getURL()");
	noReturnValue();
}

/**
 * Converts the point object from Stage coordinates to the local coordinates of the specified movie clip.
 * 
 * FIXME: If this tag is called before the first frame rendered, the matrix is not initialized.
 **/
void ASMovieClip::globalToLocal() {
	if (numPassedParams > 0 ) {
		ASObject *a = (ASObject*) script->getRegister(1);
		float x = MEMBER_TONUMERIC_NOCLONE(a,L"x");
		float y = MEMBER_TONUMERIC_NOCLONE(a,L"y");
		doTransform(x,y,invMatrix);
		a->setMemberNoClone(L"x",new FloatObject(x));
		a->setMemberNoClone(L"y",new FloatObject(y));
	}
	noReturnValue();
}

/**
 * Sends the playhead to a specific frame in the movie clip and plays the SWF file.
 * No GotoAndPlay on Movies- Applicable to sprites only.
 **/
void ASMovieClip::gotoAndPlay() {
	noReturnValue();
}

/**
 * Sends the playhead to a specific frame in the movie clip and stops the SWF file.
 * No GotoAndStop on Movies- Applicable to sprites only.
 **/
void ASMovieClip::gotoAndStop() {
	noReturnValue();
}

/**
 * Returns true if bounding box of the specified movie clip intersects the bounding box of the target movie clip.
 **/
void ASMovieClip::hitTest() {
	puts("TODO - ASMovieClip::hitTest()");
	noReturnValue();
}

/**
 * Loads the specified SWF file into the movie clip.
 **/
void ASMovieClip::loadMovie() {
	puts("TODO - ASMovieClip::loadMovie()");
	noReturnValue();
}

/**
 * Loads variables from a URL or other location into the movie clip.
 **/
void ASMovieClip::loadVariables() {
	puts("TODO - ASMovieClip::loadVariables()");
	noReturnValue();
}

/**
 * Converts a point object from the local coordinates of the movie clip to the global Stage coordinates.
 **/
void ASMovieClip::localToGlobal() {
	puts("TODO - ASMovieClip::localToGlobal()");
	noReturnValue();
}

/**
 * Sends the playhead to the next frame of the movie clip.
 **/
void ASMovieClip::nextFrame() {
	puts("TODO - ASMovieClip::nextFrame()");
	noReturnValue();
}

/**
 * Plays the specified movie clip.
 **/
void ASMovieClip::play() {
	setIsStopped(false);
	noReturnValue();
}

/**
 * Sends the playhead to the previous frame of the movie clip.
 **/
void ASMovieClip::prevFrame() {
	puts("TODO - ASMovieClip::prevFrame()");
	noReturnValue();
}

/**
 * Removes the movie clip from the Timeline if it was created with duplicateMovieClip(),MovieClip.duplicateMovieClip(), or MovieClip.attachMovie().
 **/
void ASMovieClip::removeMovieClip() {
	puts("TODO - ASMovieClip::removeMovieClip()");
	noReturnValue();
}

/**
 * Specifies a movie clip as a mask for another movie clip.
 **/
void ASMovieClip::setMask() {
	puts("TODO - ASMovieClip::setMask()");
	noReturnValue();
}

/**
 * Specifies a movie clip as draggable and begins dragging the movie clip.
 **/
void ASMovieClip::startDrag() {
	puts("TODO - ASMovieClip::startDrag()");
	noReturnValue();
}

/**
 * Stops the currently playing SWF file.
 **/
void ASMovieClip::stop() {
	setIsStopped(true);
	noReturnValue();
}

/**
 * Stops the dragging of any movie clip that is being dragged.
 **/
void ASMovieClip::stopDrag() {
	puts("TODO - ASMovieClip::stopDrag()");
	noReturnValue();
}

/**
 * Swaps the depth level of two SWF files.
 **/
void ASMovieClip::swapDepths() {
	puts("TODO - ASMovieClip::swapDepths()");
	noReturnValue();
}

/**
 * Removes a SWF file that was loaded with loadMovie().
 **/
void ASMovieClip::unloadMovie() {
	puts("TODO - ASMovieClip::unloadMovie()");
	noReturnValue();
}

/**
 * Begins drawing a fill on the Stage.
 **/
void ASMovieClip::beginFill() {
	puts("TODO - ASMovieClip::beginFill()");
	noReturnValue();
}

/**
 * Start a gradient fill
 **/
void ASMovieClip::beginGradientFill() {
	if (numPassedParams != 5 ) { noReturnValue(); return ; }  // Wrong number of arguments
	wstring fillType = script->getRegister(1)->toString();

	ASArray *_colors = (ASArray*)script->getRegister(2);
	if (_colors->getThisObjectInstanceName() != L"Array") { noReturnValue(); return ; }  // not an array!
	int nbColors = _colors->ASLength();

	ASArray *_alphas = (ASArray*)script->getRegister(3);
	if (_alphas->getThisObjectInstanceName() != L"Array") { noReturnValue(); return ; }  // not an array!
	int nbAlphas = _alphas->ASLength();

	ASArray *_ratios = (ASArray*)script->getRegister(4);
	if (_ratios->getThisObjectInstanceName() != L"Array") { noReturnValue(); return ; }  // not an array!
	int nbRatios = _ratios->ASLength();

	ScriptObject *_matrix = (ScriptObject*)script->getRegister(5);
	wstring matrixType = _matrix->getMember(L"matrixType")->toString();
	int nbRecord = nbColors > nbAlphas ? nbColors : nbAlphas;
	nbRecord = nbRecord > nbRatios ? nbRecord : nbRatios;

	// Create the gradient records
	GRADIENT *g = new GRADIENT();
	for (int i=0; i< nbRecord; i++) {
		GRADRECORD *gr = new GRADRECORD();
		char tmp[10]; sprintf(tmp,"%d",i);
		wstring wtmp = stringTowstring(tmp);
		int rgb   = int(_colors->getMember(wtmp)->toNumeric());
		RGBA c;
		c.red   = (rgb>>16)&0xff;
		c.green = (rgb>>8)&0xff;
		c.blue  = rgb&0xff;
		c.alpha = int(_alphas->getMember(wtmp)->toNumeric()*2.55);
		gr->setColor(c);
		gr->setRatio(int(_ratios->getMember(wtmp)->toNumeric()));
		g->addGradientRecord(gr);
	}

	// Create the gradient Matrix
	MATRIX  gm;
	gm.hasScale = gm.hasRotate = true;

	// FIXME Color rotation.
	if (matrixType == L"box") {
		gm.scaleX = 4096; // BUG. FIXME - should be deduced from box size - difficult
		gm.scaleY = 4096; // BUG. FIXME
		gm.rotateSkew1 	= 0; //BUG int(40960.0* cos(_matrix->getMember("r")->toNumeric()));
		gm.rotateSkew0	= 0; //BUG int(40960.0* sin(_matrix->getMember("r")->toNumeric()));
		gm.translateX	= int(20.0* _matrix->getMember(L"x")->toNumeric());
		gm.translateY	= int(20.0* _matrix->getMember(L"y")->toNumeric());
	} else {
		gm.scaleX 		= 4096; //int(4096.0* _matrix->getMember("a")->toNumeric());
		gm.scaleY 		= 4096; //int(4096.0* _matrix->getMember("e")->toNumeric());
		gm.rotateSkew0 	= 0;//int(65536.0* _matrix->getMember("d")->toNumeric());
		gm.rotateSkew1	= 0;//int(65536.0* _matrix->getMember("b")->toNumeric());
		gm.translateX	= int(20.0* _matrix->getMember(L"g")->toNumeric());
		gm.translateY	= int(20.0* _matrix->getMember(L"h")->toNumeric());
	}

	FILLSTYLE *fold = shape->getFillStyles();
	int k = shape->getNumFillStyles();
	FILLSTYLE *f2 = new FILLSTYLE[k +1];
	for (int i=0; i< shape->getNumFillStyles(); i++) f2[i] = fold[i];

	f2[k].setGradient(g);
	f2[k].setGradientMatrix(gm);
	f2[k].setFillStyleType(fillType == L"linear" ? 0x10 : 0x12);

	shape->setNumFillStyles(k+1);
	shape->setFillStyles(f2);

	// Add a style change directive
	StyleChangeRecord *st = new StyleChangeRecord();
	st->setStateNewStyles(false);
	st->setStateLineStyle(false);
	st->setStateFillStyle1(false);
	st->setStateFillStyle0(true);
	st->setStateMoveTo(false);
	st->setFillStyle0(k+1); // Starts at 1
	shape->addShapeRecord(st);
	noReturnValue();
}

/**
 * Clear all drawing commands
 **/
void ASMovieClip::clear() {
	delete shape;
	shape = new ShapeWithStyle();
	needRedraw = true;
	noReturnValue();
}

/**
 * Draw a curve
 **/
void ASMovieClip::curveTo() {
	SI32 controlX = SI32(script->getRegister(1)->toNumeric()*20.0);
	SI32 controlY = SI32(script->getRegister(2)->toNumeric()*20.0);
	SI32 anchorX  = SI32(script->getRegister(3)->toNumeric()*20.0);
	SI32 anchorY  = SI32(script->getRegister(4)->toNumeric()*20.0);
	CurvedEdgeRecord *cer = new CurvedEdgeRecord();
	cer->setControlDeltaX(controlX-lastX);
	cer->setControlDeltaY(controlY-lastY);
	cer->setAnchorDeltaX(anchorX-controlX);
	cer->setAnchorDeltaY(anchorY-controlY);
	lastX = anchorX;
	lastY = anchorY;
	shape->addShapeRecord(cer);
	needRedraw = true;
	noReturnValue();
}

/**
 * End Fill
 **/
void ASMovieClip::endFill() {
	// Add a style change directive
	StyleChangeRecord *st = new StyleChangeRecord();
	st->setStateNewStyles(false);
	st->setStateLineStyle(false);
	st->setStateFillStyle1(false);
	st->setStateFillStyle0(true);
	st->setStateMoveTo(false);
	st->setFillStyle0(0); // Starts at 1
	shape->addShapeRecord(st);
	noReturnValue();
}

/**
 * Change current line style
 * Since we convert drawing commands to a Shape Object, we just need:
 * - to add a new LineStyle to the ShapeWithStyle.
 * - to add a styleChangeRecord type entry
 **/
void ASMovieClip::lineStyle() {
	// Add a moveto if shape is empty.
	if (shape->getShapeRecords()->size() == 0) {
		StyleChangeRecord *st = new StyleChangeRecord();
		st->setStateNewStyles(false);
		st->setStateLineStyle(false);
		st->setStateFillStyle1(false);
		st->setStateFillStyle0(false);
		st->setStateMoveTo(true);
		st->setMoveDeltaX(0);
		st->setMoveDeltaY(0);
		shape->addShapeRecord(st);
	}

	// Force a Flush
	//	FlushShapeRecord *st = new FlushShapeRecord();
	//	shape->addShapeRecord(st);	
	LINESTYLE *f = new LINESTYLE[shape->getNumLineStyles() +1];
	for (int i=0; i< shape->getNumLineStyles(); i++) f[i] = shape->getLineStyles()[i];
	int i=shape->getNumLineStyles();
	f[i].width = int(script->getRegister(1)->toNumeric()*20.0);
	if (numPassedParams > 1) {
		UI32 color = UI32(script->getRegister(2)->toNumeric());
		f[i].color.red   = (color>>16)&0xff;
		f[i].color.green = (color>>8)&0xff;
		f[i].color.blue  = color&0xff;
	} else {
		f[i].color.red = f[i].color.green = f[i].color.blue = 0;
	}
	if (numPassedParams > 2) {
		f[i].color.alpha = int(2.55*script->getRegister(3)->toNumeric()+0.1);
	} else {
		f[i].color.alpha = 255;
	}
	//printf("%d %d %d %d",f[i].color.red,f[i].color.green,f[i].color.blue,f[i].color.alpha);
	shape->setNumLineStyles(i+1);
	shape->setLineStyles(f);
	StyleChangeRecord *st = new StyleChangeRecord();
	// Change to new styles ?
	st->setStateNewStyles(false);
	st->setStateLineStyle(true);
	st->setStateFillStyle1(false);
	st->setStateFillStyle0(false);
	st->setStateMoveTo(false);
	st->setLineStyle(i+1); // Starts at 1
	shape->addShapeRecord(st);	
	noReturnValue();
}

/**
 * Draws a line using the current line style.
 **/
void ASMovieClip::lineTo() {
	SI32 x = SI32(script->getRegister(1)->toNumeric()*20.0);
	SI32 y = SI32(script->getRegister(2)->toNumeric()*20.0);
	StraightEdgeRecord *sr = new StraightEdgeRecord();
	sr->setGeneralLineFlag(true);
	sr->setDeltaX(x-lastX);
	sr->setDeltaY(y-lastY);
	shape->addShapeRecord(sr);
	needRedraw = true;
	lastX = x; lastY = y;
	noReturnValue();
}

/**
 * Generate a MoveTo command
 **/
void ASMovieClip::moveTo() {
	SI32 x = SI32(script->getRegister(1)->toNumeric()*20.0);
	SI32 y = SI32(script->getRegister(2)->toNumeric()*20.0);
	StyleChangeRecord *st = new StyleChangeRecord();
	st->setStateNewStyles(false);
	st->setStateLineStyle(false);
	st->setStateFillStyle1(false);
	st->setStateFillStyle0(false);
	st->setStateMoveTo(true);
	st->setMoveDeltaX(x);
	st->setMoveDeltaY(y);
	shape->addShapeRecord(st);
	needRedraw = true;
	lastX = x; lastY = y;
	noReturnValue();
}

/**
 * Convert to Movie Clip to graphic instructions. 
 **/
void ASMovieClip::toGraphic() {
	if (needRedraw) {
		if (!isMorphShape) {
			if (graphics->getSequence(objectUnid) != NULL)	graphics->clearSequence(objectUnid);
			if (graphics->getSequence(objectUnid) == NULL) graphics->processShape(shape,objectUnid,&bounds);
		} else {
			graphics->buildMorphedShape(objectUnid,morphShape,ratio);
		}
		needRedraw = false;
	}
}

/**
 * Set the main object property
 **/
void ASMovieClip::setMainMember(ValueObject  *s) {
	puts("ASTextField::setMainMember WHAT DO DO????????????");
	setMember(L"text",s);
}


/**
 * Get the main object property
 **/
ValueObject* ASMovieClip::getMainMember() {
	return this;
}

/**
 * Register a morph shape. A morphShape is a MovieClip where shapes change over time. The changing factor is the "ratio".
 **/
void ASMovieClip::setMorphShape(DefineMorphShapeTag* t) {
	morphShape = t;
	ratio = 0;
	needRedraw = true;
	isMorphShape = true;
}

/**
 * change the color of a movie clip
 **/
void ASMovieClip::changeColor(RGB color) {
	RGBA c;
	c.red = color.red; c.green = color.green; c.blue= color.blue; c.alpha=255;

	// Update fill & line colors
	FILLSTYLE *f = shape->getFillStyles();
	for (int i=0; i< shape->getNumFillStyles(); i++) {
		f[i].setColor(c);
	}
	LINESTYLE *l = shape->getLineStyles();
	for (int i=0; i< shape->getNumLineStyles(); i++) {
		l[i].color = c;
	}
	if (graphics->getSequence(objectUnid) != NULL)	graphics->clearSequence(objectUnid);
	needRedraw = true;
}

/**
 * Associate the movie clip to a Shape
 */
void ASMovieClip::setShape(ShapeWithStyle *p) {
	delete shape;
	shape = new ShapeWithStyle(p); 
	needRedraw=true; 
}

/**
 * Does sophisticated getMember needs to be implemented for a MovieCLip?. 
 *
 **/
ValueObject* ASMovieClip::getMember(wstring memberName){
	ValueObject *res = getMemberCommon(memberName);
	if (res != NULL) return res;

	if (memberName == L"_width") {
		computeBounds();
		float width = (bounds.xmax-bounds.xmin)/20.0;		
		res = new FloatObject(width);
	} else if (memberName == L"_height") { 
		computeBounds();
		float h = (bounds.ymax-bounds.ymin)/20.0;		
		res = new FloatObject(h);
	} else if (memberName == L"_alpha") { 
		res = new IntegerObject(100);	// It's not CXFORMWITHALPHA. WHat's that alpha value?
	} else {
		res = defaultGetMember(memberName);
	}
	return res;
}

/**
 * Set a property
 **/
void ASMovieClip::setMember(wstring memberName, ValueObject *v) {
	setMovieClipMember(memberName,v);
}

/**
 * Movie clip/Sprite implementation
 **/
void ASMovieClip::setMovieClipMember(wstring memberName, ValueObject *v) {
	if (memberName == L"_X" || memberName == L"_x") {
		transformMatrix.translateX = int (v->toNumeric() * 20.0);
		hasTransformMatrix = true;
	} else if (memberName == L"_Y" | memberName == L"_y") { 
		transformMatrix.translateY = int (v->toNumeric() * 20.0);
		hasTransformMatrix = true;
	} else if (memberName == L"_xscale") {
		double a = v->toNumeric() * 655.36;
		scaleXTransformMatrix(&transformMatrix,a);
		hasTransformMatrix = true;
	} else if (memberName == L"_yscale") { 
		double a = v->toNumeric() * 655.36;
		scaleYTransformMatrix(&transformMatrix,a);
		hasTransformMatrix = true;
	} else if (memberName == L"_currentframe") {
	} else if (memberName == L"_totalframes") {
	} else if (memberName == L"_alpha") { 
	} else if (memberName == L"_visible"){
		setVisible(v->toNumeric() != 0.0f);
	} else if (memberName == L"_width") { 
	} else if (memberName == L"_rotation") {
		double angle = v->toNumeric();
		rotateTransformMatrix(&transformMatrix,angle);
		hasTransformMatrix = true;
	} else if (memberName == L"_target") { 
	} else if (memberName == L"_framesloaded") { 
	} else if (memberName == L"_name") {
	} else if (memberName == L"_droptarget") { 
	} else if (memberName == L"_url") { 
	} else if (memberName == L"_highquality") {
	} else if (memberName == L"_focusrect") { 
	} else if (memberName == L"_soundbuftime") { 
	} else if (memberName == L"_quality") { 
	} else if (memberName == L"_xmouse") { 
	} else if (memberName == L"_ymouse") { 
	} else {
		defaultSetMember(memberName,v);
	}
	setDirty();
}

/**
 * Recompute movieClip bounds
 **/
void ASMovieClip::computeBounds() {
	if (shape == NULL) return;
	float xmin = 100000;
	float xmax =-100000;
	float ymin = 100000;
	float ymax =-100000;
	float x=0,y=0;

	for (list<ShapeRecord*>::const_iterator it = shape->getShapeRecords()->begin(); it != shape->getShapeRecords()->end(); it++) {
		UI8 type = (*it)->getTypeFlag();
		switch (type) {
			case StyleChangeRecordType:	{
				StyleChangeRecord *s = (StyleChangeRecord*) *it;
				if (s->getStateMoveTo()) {
					x = s->getMoveDeltaX();
					y = s->getMoveDeltaY();
				}
				break;
			}
			case StraightEdgeRecordType: { 
				StraightEdgeRecord *s = (StraightEdgeRecord*) *it;
				if (s->getGeneralLineFlag()) {
					x += s->getDeltaX();
					y += s->getDeltaY();
				} else {
					if (s->getVertLineFlag()) {
						y += s->getDeltaVert();
					} else {
						x += s->getDeltaVert();
					}
				}
				break;
			}
			case CurvedEdgeRecordType: {							// The curve is approximated.
				CurvedEdgeRecord *s = (CurvedEdgeRecord*) *it;
				x += s->getControlDeltaX()+ s->getAnchorDeltaX();
				y += s->getControlDeltaY()+ s->getAnchorDeltaY();
				break;
			}
		}
		if (xmin > x) xmin = x;
		if (ymin > y) ymin = y;
		if (xmax < x) xmax = x;
		if (ymax < y) ymax = y;
	}
	if (xmin != 100000 && ymin != 100000) {
		bounds.xmin = int(xmin);
		bounds.xmax = int(xmax);
		bounds.ymin = int(ymin);
		bounds.ymax = int(ymax);
	} else {
		bounds.xmin = 0;
		bounds.xmax = 0;
		bounds.ymin = 0;
		bounds.ymax = 0;
	}
}



void ASMovieClip::onData() { puts("TODO - ASMovieClip::onData()"); }
void ASMovieClip::onDragOut(){ puts("TODO - ASMovieClip::onDragOut()"); }
void ASMovieClip::onDragOver(){ puts("TODO - ASMovieClip::onDragOver()") ; }
void ASMovieClip::onEnterFrame(){ puts("TODO - ASMovieClip::onEnterFrame()") ; }
void ASMovieClip::onKeyDown(){ puts("TODO - ASMovieClip::onKeyDown()"); }
void ASMovieClip::onKeyUp(){ puts("TODO - ASMovieClip::onKeyUp()"); }
void ASMovieClip::onKillFocus(){ puts("TODO - ASMovieClip::onKillFocus()"); }
void ASMovieClip::onLoad(){ puts("TODO - ASMovieClip::onLoad()"); }
void ASMovieClip::onMouseDown(){ puts("TODO - ASMovieClip::onMouseDown()"); }
void ASMovieClip::onMouseMove(){ puts("TODO - ASMovieClip::onMouseMove()"); }
void ASMovieClip::onMouseUp(){ puts("TODO - ASMovieClip::onMouseUp()"); }
void ASMovieClip::onPress(){ puts("TODO - ASMovieClip::onPress()"); }
void ASMovieClip::onRelease(){ puts("TODO - ASMovieClip::onRelease()"); }
void ASMovieClip::onReleaseOutside(){ puts("TODO - ASMovieClip::onReleaseOutside()"); }
void ASMovieClip::onRollOut(){ puts("TODO - ASMovieClip::onRollOut()"); }
void ASMovieClip::onRollOver(){ puts("TODO - ASMovieClip::onRollOver()"); }
void ASMovieClip::onSetFocus(){ puts("TODO - ASMovieClip::onSetFocus()"); }
void ASMovieClip::onUnload(){ puts("TODO - ASMovieClip::onUnload()"); }

/**
 * Associate a movie clip to an action script class.
 * The graphic data are transfered to the action script class.
 * This is used by the action script Object.registerClass().
 **/
void ASMovieClip::associateMovieClip(ASInterface *a) {
	ASMovieClip *m = (ASMovieClip*) a;		// parameter should be a movie clip as required in the specs
	//TODO: finish this association. - extra fields may be required to be copied from the movieClip to the action script object.
	setTransformMatrix(a->getTransformMatrix());
	setHasTransformMatrix(a->getHasTransformMatrix());
	setCxFormWithAlpha(a->getCxFormWithAlpha());
	setHasCxFormWithAlpha(a->getHasCxFormWithAlpha());
}
