/**
Action Script library. Number class implementation
Copyright (C) 2005 Christophe Choquet (cchoquet@users.sourceforge.net)  

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_number.hh"
static char symbols[37] = "0123456789abcdefghijklmnopqrstuvwxyz";
const int MAX_BASE = 36; 

typedef void (ASNumber::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASNumber::init,		// 0
	&ASNumber::valueOf,		// 1
	// Need to be at the end.
	&ASNumber::addProperty,		
	&ASNumber::registerClass,	
	&ASNumber::AStoString,		
	&ASNumber::unwatch,			
	&ASNumber::valueOf,			
	&ASNumber::watch
};

/**
 * The constructor
 *
 **/
ASNumber::ASNumber(Script *s, Graphics *g) : ASObject(s,g,L"Number") {
	setMemberNoClone(L"MAX_VALUE",new DoubleObject(1.79769313486231e+308));
	setMemberNoClone(L"MIN_VALUE",new DoubleObject(4.94065645841247e-324));
	setMemberNoClone(L"NaN",new StringObject(L"NaN"));
	setMemberNoClone(L"NEGATIVE_INFINITY",new StringObject(L"-Infinity"));
	setMemberNoClone(L"POSITIVE_INFINITY",new StringObject(L"Infinity"));
}

ASNumber::ASNumber(Script *s, Graphics *g, double value) : ASObject(s,g,L"Number") {
	this->value = value;
}

/**
 * The destructor
 **/
ASNumber::~ASNumber() {
}

/**
 * Return a function descriptor
 */
Function *ASNumber::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"Number") 	{func->setFunctionIndex(0); func->setNumParams(1); }
	if (memberName == L"valueOf") 	{func->setFunctionIndex(1); func->setNumParams(0); }
	finishFunctionSetup(memberName,2,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASNumber::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}


/**
 * If there is one parameter to the constructor, used it.
 **/
void ASNumber::init() {
	value = 0;
	if (numPassedParams > 0) {
		value = script->getRegister(1)->toNumeric();
	}	
}

/**
 * Convert the number to a string. If a parameter is passed, it specifies the base in 2..36 range.
 **/
void ASNumber::AStoString() {
	DoubleObject *tmp = new DoubleObject(value);
	if (numPassedParams == 0) {
		script->push(new StringObject(tmp->toString()));
	} else {
		int base = int(script->getRegister(1)->toNumeric());
		long v = long(value);
		char *number = (char*) calloc(256,1);
		if( numberToBase(base, v, number) ) {
			script->push(new StringObject(stringTowstring(number)));
		} else {
			script->push(new DoubleObject(value));
		}
		free(number);
	}
	delete tmp;
}

/**
 *
 **/
void ASNumber::valueOf() {
	script->push(new DoubleObject(value));
}

/**
 * Convert a number to base 2..36
 **/
bool ASNumber::numberToBase(int base, long iDec, char* szString) {
	if(base<2||base>MAX_BASE) return false; //Failed
	if(iDec==0){
		strcpy(szString,"0");
		return true;
	}
	
	int count = 0;
	char *chResult = (char*) calloc(256,1);
	while(iDec > 0 && count<255) {
		chResult[count] = symbols[iDec % base];
		iDec = iDec/base;
		count++;
	}
	if (count == 255) return 0;
	
	for (int i=0, j= count-1; j>=0; j--, i++) szString[i] = chResult[j];
	szString[count] = 0;
	free(chResult);
	return true;
}
