/**
Action Script library. Video class implementation
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_video.hh"

/*
 * You may have a look at the file swf.c in the ffmpeg package.
 * This decoder can't be used here because of decoding logic used here.
 * Infos were grabed from this file.
 * The swf specs are not completly accurate: the h263 video format is in fact a FLV1 format.
 *
 **/
extern AVInputFormat flv_iformat;			// From ffmpeg

typedef void (ASVideo::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASVideo::init,			// 0
	&ASVideo::attachVideo,	// 1
	&ASVideo::clear,		// 2
	// Need to be at the end.
	&ASVideo::addProperty,		
	&ASVideo::registerClass,	
	&ASVideo::AStoString,		
	&ASVideo::unwatch,			
	&ASVideo::valueOf,			
	&ASVideo::watch
};

/**
 * The constructor
 **/
ASVideo::ASVideo(Script *s, Graphics *g) : ASObject(s,g,L"Video") {
}

/**
 * The destructor
 **/
ASVideo::~ASVideo() {
	graphics->getTexture()->removeTexture(textureId);
}

/**
 * Return a function descriptor
 */
Function *ASVideo::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"Video") 	{func->setFunctionIndex(0); func->setNumParams(0); }
	if (memberName == L"attachVideo"){func->setFunctionIndex(1); func->setNumParams(1); }
	if (memberName == L"clear") 	{func->setFunctionIndex(2); func->setNumParams(0); }
	finishFunctionSetup(memberName,3,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASVideo::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 *
 **/
void ASVideo::init() {
}

/**
 *
 **/
void ASVideo::attachVideo() {
	puts("TODO - ASVideo::attachVideo()");
	noReturnValue();
}

/**
 * Clears the image currently displayed in the Video object.
 **/
void ASVideo::clear() {
	freeDecoder();
	noReturnValue();
}

/**
 * Convert 
 **/
void ASVideo::toGraphic() {
	if (needRedraw) {
		// ?????	graphics->buildMorphedShape(objectUnid,morphShape,ratio);
		needRedraw = false;
	}
}

/**
 * Initialize a Video Stream from a VideoStreamTag
 **/
void ASVideo::init(DefineVideoStreamTag* t ) {
	puts("ASVideo::init");
	setMemberNoClone(L"deblocking",new IntegerObject(t->getVideoFlagsDeblocking()));
	setMemberNoClone(L"width",new IntegerObject(t->getWidth()));
	setMemberNoClone(L"height",new IntegerObject(t->getHeight()));
	setMemberNoClone(L"smoothing",new BooleanObject(t->getVideoFlagsSmoothing()));
	width2 = nextPowerOf2(t->getWidth());
	height2 = nextPowerOf2(t->getHeight());
	textureId = graphics->getTexture()->allocateTexture24(width2,height2);
	graphics->registerVideoBox(objectUnid,20*t->getWidth(),20*t->getHeight(),textureId);
	// Register all formats and codecs
	av_register_all();
	codecId = t->getCodecID();
	width = t->getWidth();
	height = t->getHeight();

	// Init the codecs
	if (codecId == 2) {	// an H263VIDEOPACKET
	    pFormatCtx = av_alloc_format_context();
		AVStream *st = av_new_stream(pFormatCtx, 0);
#if LIBAVCODEC_VERSION_INT < ((49<<16)+(0<<8)+2)
		pCodecCtx = &st->codec;
#else
		pCodecCtx = st->codec;
#endif
		pCodec=avcodec_find_decoder(CODEC_ID_FLV1);
		pCodecCtx->width = width;
		pCodecCtx->height = height;
		//pCodecCtx->opaque = videoData;
		pFormatCtx->iformat = &flv_iformat;
		pFormatCtx->nb_streams = 1;

	    avcodec_open(pCodecCtx, pCodec);
    	pFrame=avcodec_alloc_frame();
    	pFrameRGB=avcodec_alloc_frame();
	    int numBytes=avpicture_get_size(PIX_FMT_RGB24, pCodecCtx->width, pCodecCtx->height);
		buffer=new uint8_t[numBytes];
		avpicture_fill((AVPicture *)pFrameRGB, buffer, PIX_FMT_RGB24, pCodecCtx->width, pCodecCtx->height);
	} 
	if (codecId == 3) {
		// gzip compressed?
		puts("SCREENVIDEOPACKET video format not supported yes");
	}
}


/**
 * Decode a video Frame
 **/
void ASVideo::decodeFrame(UI16 frameNum,UI8* videoData, UI32 size) {
	if (codecId == 3) return;

	getNextFrame(pFormatCtx, pCodecCtx, videoData, pFrame,size);
	img_convert((AVPicture *)pFrameRGB, PIX_FMT_RGB24, (AVPicture*)pFrame, pCodecCtx->pix_fmt, pCodecCtx->width, pCodecCtx->height);

	// Transfer image to video card
	graphics->getTexture()->video2Texture(textureId,width,height,(char*)pFrameRGB->data[0]);
}


/**
 * Get next video frame - Code adapted from Martin Böhme ffmpeg example
 **/
bool ASVideo::getNextFrame(AVFormatContext *pFormatCtx, AVCodecContext *pCodecCtx,UI8* videoData, AVFrame *pFrame, int size) {
	int	bytesRemaining=size;
	int	bytesDecoded;
	int	frameFinished;

	while(bytesRemaining > 0) {
		// Decode the next chunk of data
		bytesDecoded=avcodec_decode_video(pCodecCtx, pFrame,&frameFinished, videoData, bytesRemaining);
		if(bytesDecoded < 0) {
			puts("ASVideo. Error while decoding frame\n");
			return false;
		}

		bytesRemaining-=bytesDecoded;
		videoData+=bytesDecoded;

		// Did we finish the current frame? Then we can return
		if(frameFinished)	return true;
	}
	return frameFinished!=0;
}


/**
 * Release decoding resources.
 **/
void ASVideo::freeDecoder() {
	// Cleanup buffers & structures
    delete [] buffer;
    av_free(pFrameRGB);
	av_free(pFrame);
	avcodec_close(pCodecCtx);
}
