/**
Action Script library. XMLNode class implementation
Copyright (C) 2005 Christophe Choquet (cchoquet@users.sourceforge.net)  

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_xml_node.hh"

typedef void (ASXmlNode::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASXmlNode::init,			// 0
	&ASXmlNode::appendChild,	// 1
	&ASXmlNode::cloneNode,		// 2
	&ASXmlNode::hasChildNodes,	// 3
	&ASXmlNode::insertBefore,	// 4
	&ASXmlNode::removeNode,		// 5
	// Need to be at the end.
	&ASXmlNode::addProperty,		
	&ASXmlNode::registerClass,	
	&ASXmlNode::AStoString,		
	&ASXmlNode::unwatch,			
	&ASXmlNode::valueOf,			
	&ASXmlNode::watch
};

/**
 * The constructors
 **/
ASXmlNode::ASXmlNode(Script *s, Graphics *g) : ASObject(s,g,L"XMLNode") {
}
ASXmlNode::ASXmlNode(Script *s, Graphics *g, xmlNode *node,xmlDocPtr doc): ASObject(s,g,L"XMLNode") {
	this->node = node;
	this->doc = doc;
}

/**
 * The destructor
 **/
ASXmlNode::~ASXmlNode() {
}

/**
 * Return a function descriptor
 */
Function *ASXmlNode::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"XMLNode") 		{func->setFunctionIndex(0); func->setNumParams(0); }
	if (memberName == L"appendChild") 	{func->setFunctionIndex(1); func->setNumParams(0); } //??
	if (memberName == L"cloneNode")	 	{func->setFunctionIndex(2); func->setNumParams(0); }//??
	if (memberName == L"hasChildNodes") {func->setFunctionIndex(3); func->setNumParams(0); }//??
	if (memberName == L"insertBefore") 	{func->setFunctionIndex(4); func->setNumParams(0); }//??
	if (memberName == L"removeNode") 	{func->setFunctionIndex(5); func->setNumParams(0); }//??
	func = finishFunctionSetup(memberName,6,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASXmlNode::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 *
 **/
void ASXmlNode::init() {
	puts("TODO - ASXmlNode::init()");
}

/**
 *
 **/
void ASXmlNode::appendChild() {
	puts("TODO - ASXmlNode::appendChild()");
	noReturnValue();
}

/**
 *
 **/
void ASXmlNode::cloneNode() {
	puts("TODO - ASXmlNode::cloneNode()");
	noReturnValue();
}

/**
 * Returns true if the specified node has child nodes; otherwise, returns false.
 **/
void ASXmlNode::hasChildNodes() {
	script->push(new BooleanObject(node->children ? true : false));
}

/**
 *
 **/
void ASXmlNode::insertBefore() {
	puts("TODO - ASXmlNode::insertBefore()");
	noReturnValue();
}

/**
 *
 **/
void ASXmlNode::removeNode() {
	puts("TODO - ASXmlNode::removeNode()");
	noReturnValue();
}

/**
 * Display current node value
 **/
void ASXmlNode::AStoString() {
	if (node != NULL ) {
		xmlBufferPtr buffer = xmlBufferCreate();
		if (buffer != NULL) {
			xmlNodeDump(buffer,doc,node,100,0);
			wstring s = stringTowstring((const char*)buffer->content);
			xmlBufferFree(buffer);
			script->push(new StringObject(s));
		} else {
			noReturnValue();
		}
	} else {
		noReturnValue();
	}
}

/**
 * We don't use default member behavior for arrays
 **/
void ASXmlNode::setMember(wstring memberName, ValueObject *v) {
}

/**
 * We don't use default member behavior for arrays
 **/
ValueObject* ASXmlNode::getMember(wstring memberName) {
	ValueObject* res = NULL;
	if (memberName == L"firstChild") {
		res = new ASXmlNode(script,graphics,node->children,doc);
	}
	if (memberName == L"nodeValue") {
		if (node->content != NULL) {
			wstring s = stringTowstring((const char*)node->content);
			res = new StringObject(s);
		} else {
			res = new StringObject(L"null");
		} 
	}
	if (memberName == L"nextSibling") {
		xmlNode *next = node->next;
		printf("NODE : %p\n",next);
		if (next != NULL) res = new ASXmlNode(script,graphics,next,doc);
	}
	return res;
}

/**
 * Get the main object property
 **/
ValueObject* ASXmlNode::getMainMember() {
	AStoString();
	ValueObject *res = script->pop();
	return res;
}
