/**
Macromedia Flash common includes.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#ifndef _COMMON_
#define _COMMON_

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <list>
#include <string>
#include <ext/hash_map>
#include <ext/hash_set>
#include <wchar.h>

using namespace std;
namespace stdext = ::__gnu_cxx;

class HashString {
public:
   int operator()(std::string const &str) const   {
     return stdext::hash<char const *>()(str.c_str());
   }
};

class HashStringCompare {
 public:
   bool operator()(std::string  s1, std::string  s2) const {
     return s1 == s2;
   }
};

class HashUTF8String {
public:
   int operator()(std::wstring const &str) const   {
		unsigned int hash_value = 0;
		int len = str.size();
		const wchar_t *head = str.c_str();
		while (--len >= 0)
			hash_value = (hash_value << 5) - hash_value + *head++;
		return hash_value;
   }
};

class HashUTF8StringCompare {
 public:
   bool operator()(std::wstring  s1, std::wstring  s2) const {
     return s1 == s2;
   }
};

class HashInt {
public:
   int operator()(int i) const   {
     return stdext::hash<int>()(i);
   }
};

class HashIntCompare {
 public:
   bool operator()(int  s1, int  s2) const {
     return s1 == s2;
   }
};

class HashLong {
public:
   long operator()(long i) const   {
     return stdext::hash<long>()(i);
   }
};

class HashLongCompare {
 public:
   bool operator()(long  s1, long  s2) const {
     return s1 == s2;
   }
};

typedef stdext::hash_map<int, int, HashInt,  HashIntCompare> IntIntHashMap;
typedef stdext::hash_map<wstring, int, HashUTF8String, HashUTF8StringCompare> StringIntHashMap;
typedef stdext::hash_map<wstring, float, HashUTF8String, HashUTF8StringCompare> StringFloatHashMap;
typedef stdext::hash_set<int, HashInt,  HashIntCompare> IntHashSet;
typedef stdext::hash_map<int, wstring, HashInt,  HashIntCompare> IntStringHashMap;
typedef stdext::hash_map<int, string, HashInt,  HashIntCompare> ASCIIIntStringHashMap;
typedef stdext::hash_map<string, int, HashString, HashStringCompare> ASCIIStringIntHashMap;
typedef stdext::hash_map<string, float, HashString, HashStringCompare> ASCIIStringFloatHashMap;

#define RGBA2INT(c) ((((((c->red<<8)+c->green)<<8)+c->blue)<<8)+c->alpha)

// Unicode helpers
int wctomb (char* s, wchar_t swc);
void printWstring(wstring w);
void fprintWstring(FILE *f,wstring w);
wstring stringTowstring(const char *s);
wstring utf8Towstring(const char *s,int len=0);
size_t mbrtowc (wchar_t* pwc, const char* s, size_t n);
wstring extentedAsciiTowstring(const char *s,int len=0);
void wstringToUtf8(const wstring src, char *s);
wstring toUpper(wstring in);

// Critial error
void CRITICAL_ERROR(wstring error);
extern bool actionScriptCriticalFailure;
#endif
