/**
Macromedia Flash Maths lib.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#ifndef _MATHS_H_
#define _MATHS_H_

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "common.hh"
#include "swf.hh"

#define getColorf(b) ((float)b/255.0f)

typedef struct {
	int		points[2];
	double	x[2],y[2];
	int		originalIndex[2];
	int		colorsIndex[2];
	LINESTYLE *lineColor;
} Segment;

typedef struct  {
	double	a;
	double	b;
	double	c;
	double	d;
	double	translateX;
	double	translateY;
} NORMMATRIX;

float 	QuadraticBezierInterp(float s, float c, float e, float t);
void 	applyMatrix(MATRIX *m);
void 	getMinMax(int &xmin,int &xmax,int &ymin,int &ymax,Segment *segments,int nbSegments);
int 	getColorIndex(Segment *segments,int nbSegments,float vertex[3][2]);
long 	nextPowerOf2(long in);
NORMMATRIX normalize(MATRIX *m);
NORMMATRIX MatrixInvert(NORMMATRIX *m);
RGBA 	interpolateRGBA(RGBA *c1, RGBA *c2, float t);
MATRIX 	interpolateMATRIX(MATRIX *c1, MATRIX *c2,float t);
GRADIENT *interpolateGRADIENT(GRADIENT *c1, GRADIENT *c2,float t);
GRADIENT *interpolateMORPHGRADIENT(MORPHGRADIENT *m, float t);
bool 	polygon_contains_point_2d(int nbSegments, Segment *s,float xval,float yval );
bool 	triangle_contains_point_2d(float t[3][2],float xval,float yval );
void 	doColorPartition(float *tmpPoints,Segment *tmpSegments,int &tmpNbPoints,int &tmpNbSegments);
RECT 	doTransform(RECT a, float *m);
void 	doTransform(float &x, float &y, float *m);
void	invert4x4Matrix(float *data, float *invmat);
bool 	segmentIntersect(float Ax, float Ay, float Bx, float By, float Cx, float Cy,  float Dx, float Dy,float &Px, float &Py);
bool 	segmentColinear(float Ax, float Ay, float Bx, float By, float Cx, float Cy, float Dx, float Dy);
void 	closePolygon(Segment *segments,int &nbSegments,float *points,int nbPoints);
bool 	isClosedPolygon(Segment *segments,int nbSegments, int nbPoints);
CXFORMWITHALPHA multiplyCxFormWithAlpha(CXFORMWITHALPHA *src1,CXFORMWITHALPHA *src2);
void 	normalizeTransformMatrix(MATRIX *transformMatrix);
void 	rotateTransformMatrix(MATRIX *transformMatrix, double angle);
double 	getTransformationMatrixAngle(MATRIX *transformMatrix);
bool 	doesOverlap(RECT *R1, RECT *r2);
RECT 	doTransform(RECT a,MATRIX *m);
void 	convertMatrix(float *mat, MATRIX *m);
void 	multiplyMatrix(float *res, float *mat1, float *mat2);
void 	scaleXTransformMatrix(MATRIX *transformMatrix, double scale);
void 	scaleYTransformMatrix(MATRIX *transformMatrix, double scale);
#endif
