/**
Macromedia Flash Processed Object Store.
Copyright (C) 2005 Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include <iostream>
#include "object_store.hh"
#include "../ASLib/as_interface.hh"
#include "../ASLib/as_sprite.hh"

extern bool debug;
typedef multimap <long,ScriptObject*>::iterator RCIT; 
typedef pair<RCIT, RCIT> RRange;

/**
 * The contructor
 **/
ObjectStore::ObjectStore(Graphics *graphics) {
	currentTargetPath = L"/";
	this->graphics = graphics;
	globalRegisters = new RegisterSet(4);
	globalRegisters->setGlobal(true);
	poolActionContext = L"_global";
	objectAtDepth.clear();
	global = NULL;
	hasLocalVariableScope = true;
}

/**
 * Destructor
 **/
ObjectStore::~ObjectStore(){
	//puts("~ObjectStore");
	//for (StringFunctionHashMap::const_iterator i=functions.begin(); i!= functions.end(); ++i) {
	//	Function *f = i->second;
	//	if (f != NULL) delete f->getRegisters();
	//}
	functions.clear();

	// Clear local variables
	for (StringValueObjectHashMapHashMap::iterator it=variables.begin(); it != variables.end(); it++) {
		StringValueObjectHashMap *m = (*it).second;
		if (m != NULL) {
			for (StringValueObjectHashMap::iterator it2=m->begin(); it2 != m->end(); it2++) {
				ValueObject *v = (*it2).second;
				removeObject(v);
			}
			m->clear();
			delete m;
		}
	}
	variables.clear();

	// Clear constants
	for (StringIntValueObjectHashMapHashMap::iterator it=constantPools.begin(); it != constantPools.end(); it++) {
		IntValueObjectHashMap *m = (*it).second;
		if (m!=NULL) {
			for (IntValueObjectHashMap::iterator it2=m->begin(); it2 != m->end(); it2++) removeObject((*it2).second);
			m->clear();
			delete m;
		}
	}
	constantPools.clear();

	// clear rendered objects, in two steps (I love C++!)
	// Since ScriptObject members may be either basic type (int,float, string) or complex objects,
	// We need to clear basic objects types first on all objects, before we remove those objects.
	// If we don't do that in two steps,removing a ScriptObject may fails because one of of its members refers to an already removed object in its members
	for (map<ScriptObject*,ScriptObject*>::const_iterator it=allocatedObjects.begin(); it != allocatedObjects.end(); ++it) {
		ScriptObject *m = (*it).second;
		if (m != NULL) m->removeMembers();
	}

	for (map<ScriptObject*,ScriptObject*>::const_iterator it=allocatedObjects.begin(); it != allocatedObjects.end(); ++it) {
		ScriptObject *m = (*it).second;
		if (m != NULL) delete m;
	}

	allocatedObjects.clear();
	delete globalRegisters;
	linkages.clear();
}

/**
 * Go to a new level for local variables
 * Use object pointer address as a unique identifier, 
 * since the movie name may be used multiple times, 
 * if this object is placed more than once.
 **/
void ObjectStore::targetPathDown(ScriptObject *newLevel) {
	currentTargetPathStack.push(currentTargetPath);
	if (currentTargetPath == L"/") currentTargetPath = L"/" + buildKey(newLevel);
	else currentTargetPath = currentTargetPath +  L"/" + buildKey(newLevel);
}

/**
 * Go UP one level for local variables
 **/
void ObjectStore::targetPathUp() {
	currentTargetPath = currentTargetPathStack.top();
	currentTargetPathStack.pop();
	//wstring::size_type po = currentTargetPath.find_last_of('/');
	//if (po != wstring::npos) currentTargetPath = currentTargetPath.substr(0,po);
	if (currentTargetPath == L"") currentTargetPath = L"/" ;
}

/**
 * Find the context of a variable
 * 
 * Try at current local scope, the goes up to /
 * if it is not found, try on the Sprite object
 */
wstring	ObjectStore::findVariableContext(ScriptObject *so, wstring key) {
	key = toUpper(key);
	StringValueObjectHashMap *s = variables[currentTargetPath];
	if (s != NULL && (*s)[key] != NULL) {
		return currentTargetPath;
	}
	wstring sc = currentTargetPath;
	wstring::size_type po = sc.find_last_of('/');
	if (po != wstring::npos) sc = sc.substr(0,po);
	while (po != string::npos) {
		s = variables[sc];
		if (s != NULL && (*s)[key] != NULL) {
			return sc;
		}
		po = sc.find_last_of('/');
		if (po != string::npos && po > 0) sc = sc.substr(0,po-1);
		if (po == 0) po =string::npos; 
	}

	ValueObject * vo = getGlobalVariable(key);
	if (vo != NULL) {
		removeObject(vo);
		// Sometimes we should set the global variable if it does exists,
		// Sometimes a local variable may shadow.
		// rule not understood -> BOGUS
		//return L"/Bogus";
	}
	// if not found, return SpriteObject context
	return L"/"+buildKey(so);
}

/**
 * Add a local variable to the current local level (ie: /A/B/foo)
 **/
void ObjectStore::addLocalVariable(wstring key, ValueObject *v) {
	if (hasLocalVariableScope) {
		StringValueObjectHashMap *s = variables[currentTargetPath];
		if (s == NULL) {
			s = new StringValueObjectHashMap();
			variables[currentTargetPath] = s;
		} else {
			ValueObject *vo = (*s)[toUpper(key)];
			if (vo != NULL) removeObject(vo);
		}
		(*s)[toUpper(key)] = v->clone();
	} else {
		setGlobalVariable(key,v);
	}
}

/**
 * Set a variable.
 * If we are currently processing an onClipEvent, we look for global variables.
 **/
void ObjectStore::setVariable(ScriptObject *so, wstring key, ValueObject *v) {
	if (hasLocalVariableScope) {
		// Check if we reference a global variable
		wstring::size_type pos = key.find(L"_level0/");
		if (pos != wstring::npos) {			// That's global variable
			key = key.substr(pos+8);
			// Check if the variable contains the ":" string. In that case, it's eqyuvalent to a setProperty.
			pos = key.find(L":");
			if (pos != string::npos) {
				wstring var = key.substr(0,pos);
				wstring member = key.substr(pos+1);
				wstring uvar = var;
				pos = uvar.find(L"/");
				if (pos != string::npos) {
					uvar = uvar.substr(0,pos);
					member = var.substr(pos)+L"/"+member;
				}
				if (debug) wcout << L"set global member " << uvar << L"[" << member<<L"]";
				ASInterface *g = (ASInterface*)getGlobalVariable(uvar);
				g->setChildrenProperty(member,v);								// Set a property on a contained object
			} else {
				puts("setVariable :: FIXME!!!!");	
			}
		} else {
			// Update the member if needed.
			if (so != NULL && so->getClassName() == L"Sprite") {
				so->setMember(key,v);
			}	

			wstring ctx = findVariableContext(so,key);
			StringValueObjectHashMap *s = variables[ctx];
			if (s == NULL) {
				s = new StringValueObjectHashMap();
				variables[ctx] = s;
			} 
			ValueObject *ve = (*s)[toUpper(key)];
			if (ve != NULL) {
				if (ve->getObjectType() != ScriptObjectType) delete ve;
				else {
					ASObject * a = (ASObject*) ve;
					wstring className = a->getClassName();
					if (className != L"Sprite" && className != L"MovieClip") {
						if (className == L"Date") delete a;
						else if (className == L"Math") delete a;
			//////			delete a; CHECK conditions to clear
					}
				}
			}
			if(debug) wcout << L"Add local: " << ctx << L"/"<<key;
			(*s)[toUpper(key)] = v->clone();
		}
	} else {
		setGlobalVariable(key,v);
	}
}

/**
 * Remove a local variable
 **/
void ObjectStore::removeLocalVariable(wstring key) {
	key = toUpper(key);
	StringValueObjectHashMap *s = variables[currentTargetPath];
	if (s != NULL) {
		ValueObject *v = (*s)[key];
		if (v !=NULL) removeObject(v);
		(*s).erase(key);
	}
}

/**
 * Read a variable 
 **/
ValueObject *ObjectStore::getVariable(ScriptObject *so, wstring key) {
	if (hasLocalVariableScope) {
		key = toUpper(key);
		//printf("context : %x\n",so);
		wstring scope = findVariableContext(so,key);
		StringValueObjectHashMap *s = variables[scope];
		if (s != NULL) {
			ValueObject *v = (*s)[key];
			if (v != NULL) return v->clone();
		}
		ValueObject *v = so->getMember(key);
		if (v != NULL) return v;
		v = getGlobalVariable(key);
		if (v != NULL) return v;

		if (so->getParent() != NULL) {
			//if (debug)puts("Search for PARENTS");
			// Find All parents. TODO
			ScriptObject *parent = so->getParent();
			v = parent->getMember(key);
			if (v != NULL) return v;
		}
	} else {
		return getGlobalVariable(key);
	}	
	return NULL;
}

/**
 * Absolute path version
 **/
ValueObject*ObjectStore::getVariable(wstring absolutePath) {
	absolutePath = absolutePath.substr(1); 		// Skip /
	ValueObject *so = NULL;
	wstring::size_type po = absolutePath.find('/');

	wstring sc = absolutePath;
	while (po != wstring::npos) {
		sc = absolutePath.substr(0,po);
		if (so == NULL) so = getGlobalVariable(sc);
		else so = getSpriteVariable((ScriptObject*)so,sc);
		absolutePath = absolutePath.substr(po+1);
		po = absolutePath.find('/');
	}
	if (absolutePath.find(L"_root.") == 0) {
		absolutePath = absolutePath.substr(6);
		so = getObjectByDotPath(global,absolutePath);
	} else {
		if (so == NULL) so = getGlobalVariable(absolutePath);
		else so = getSpriteVariable((ScriptObject*)so,absolutePath);
	}
	return so;
}

/**
 * Returns a Global variable
 **/
ValueObject *ObjectStore::getGlobalVariable(wstring key) {
	if (key.find('/') == 0) key = key.substr(1);
	wstring hkey = toUpper(key);
	StringValueObjectHashMap *s = variables[L"/"];
	ValueObject *v = (*s)[key];
	if (v != NULL) return v->clone();
	v = (*s)[hkey];
	if (v != NULL) return v->clone();

	s = variables[L"/"+buildKey(global)];
	if ( s != NULL) {
		v = (*s)[key];
		if (v != NULL) return v->clone();
		v = (*s)[hkey];
		if (v != NULL) return v->clone();
	}
	return NULL;
}

/**
 * Set a Global variable
 **/
void ObjectStore::setGlobalVariable(wstring key, ValueObject *v) {
	wstring hkey = toUpper(key);
	StringValueObjectHashMap *s = variables[L"/"];
	if (s == NULL) {
		s = new StringValueObjectHashMap();
		variables[L"/"] = s;
	} 
	wstring bkey = buildKey((ScriptObject*)v);
	if ((*s)[key] != NULL) removeObject((*s)[key]);
	if ((*s)[hkey] != NULL) removeObject((*s)[hkey]);
	if ((*s)[bkey] != NULL) removeObject((*s)[bkey]);
	(*s)[key] = v->clone();
	(*s)[hkey] = v->clone();
	(*s)[bkey] = v->clone();
}

/**
 * Clean up current level (when we leave a function)
 **/
void ObjectStore::cleanCurrentLevel() {
/*
FIXME: what is the policy of variable cleaning?

wcout << L"clean " << currentTargetPath << L"\n";
	if (currentTargetPath == L"/") return;
	StringValueObjectHashMap *s = variables[currentTargetPath];
	if (s != NULL) {
		for (StringValueObjectHashMap::iterator it2=s->begin(); it2 != s->end(); it2++) {
			ValueObject *v = (*it2).second;
			if (v != NULL) removeObject(v);
		}
		s->clear();
	}
	variables.erase(currentTargetPath);
*/
}

/**
 * Set pool action context - for constants
 **/
void ObjectStore::setPoolActionContext(wstring s) {
	//wcout << L"\nset POOL: " << s;
	poolActionContext = s; 
}

/**
 * Set a constant value (in the Constant Pool).
 */
void ObjectStore::setConstant(int entry, ValueObject *p) {
	IntValueObjectHashMap *m = constantPools[poolActionContext];
	if (m==NULL) {
		m = new IntValueObjectHashMap();
		constantPools[poolActionContext] = m;
	}
	ValueObject *c = (*m)[entry];
	if (c != NULL) delete c;
	(*m)[entry] = p;
}

/**
 * Get constant pool value
 */
ValueObject *ObjectStore::getConstant(int entry) {
	//wcout << L"POOL "<<  poolActionContext << L" ";
	IntValueObjectHashMap *m = constantPools[poolActionContext];
	if (m != NULL) {
		ValueObject *c = (*m)[entry];
		if (c!=NULL) return c->clone();
		return new StringObject(L"undefined");
	} else {
		CRITICAL_ERROR(L" SANITY CHECK: SOMETHING WRONG in ObjectStore::getConstant");
		return new StringObject(L"undefined");
	}
}

/**
 * Removes an object testing if it can be removed (ie not a ScriptObject)
 **/
void ObjectStore::removeObject(ValueObject *v) {
	//wcout << L"Free " << v;
	if (v!=NULL) {
		if (v->getObjectType() != ScriptObjectType) { delete v; v = NULL; }
		else {
			ScriptObject *so = (ScriptObject*) v;
			if (so->getShouldBeRemoved()) { 
				delete so;
 				allocatedObjects.erase(so);
				v = NULL; 
			}		// That's when the object is explicitly removed from actionScript
		}
	}
}

/**
 * Get a function. We make some class transforms if we have base classes.
 **/
Function* ObjectStore::getFunction(wstring className,wstring functionName) {
	if (className == L"_$$$String") className=L"String";
	if (className == L"_$$$Number") className=L"Number";
	Function *ret = functions[className+L"."+functionName];
	if (ret == NULL) ret = functions[L"_global."+functionName];
	return ret;
}

void ObjectStore::addFunction(wstring className,wstring functionName, Function *f) { 
	wstring key = buildKey(f);
	if (constantPools[key] == NULL) {
		f->setConstantPoolKey(key);
		copyCurrentPoolContext(key);
	} else {
		f->setConstantPoolKey(L"");
	}
	functions[className+L"."+functionName] = f;
	allocatedObjects[f] = f;
}

/**
 * returns a movie given its name or its path
 *
 **/
ScriptObject*ObjectStore::getMovieByNameOrPath(wstring namePath) {
	if (namePath.find('/') == 0) namePath = namePath.substr(1);
	wstring::size_type po = namePath.find('/');
	ScriptObject*so = NULL;
	if (po == wstring::npos) {			// it's a simple movie name
		so = (ScriptObject*)getGlobalVariable(namePath);
	} else {
		wstring main = namePath.substr(0,po);
		ASSprite *mo = (ASSprite *)getGlobalVariable(main);
		if (mo != NULL) so = mo->getMovieByNameOrPath(namePath.substr(po+1));
	}
	if (so != NULL && so->getObjectType() != ScriptObjectType) {
		so=NULL;
	}
	return so;
}

/**
 * Get an object when specified as  _root.spriteA.spriteB.buttonC.movieclipD
 **/
ValueObject* ObjectStore::getObjectByDotPath(ScriptObject *root, wstring key) {
	//wcout << key << L"\n";
	wstring::size_type po = key.find('.');
	ValueObject *res = NULL;
	if (po != wstring::npos) {
		root = (ScriptObject*)getGlobalVariable(key.substr(0,po));
		key = key.substr(po+1);
		po = key.find('.');
		while (po != wstring::npos) {
			root = (ScriptObject*) getSpriteVariable(root,key.substr(0,po));
			key = key.substr(po+1);
			po = key.find('.');
		}
		res = getSpriteVariable(root,key);
		if (res == NULL && root->getClassName() == L"Sprite") res = ((ASSprite*)root)->getMovieByNameOrPath(key);
	} else {
		res = getGlobalVariable(key);
	}
	return res;
}

/**
 * Set sprite scope variable
 **/
void ObjectStore::setSpriteVariable(ScriptObject *so,wstring key, ValueObject *v) {
	key = toUpper(key);
	//if(debug)puts("setSpriteVariable\n");
	wstring targetPath = L"/"+buildKey(so);
	StringValueObjectHashMap *s = variables[targetPath];
	if (s == NULL) {
		s = new StringValueObjectHashMap();
		variables[targetPath] = s;
	} 
	(*s)[key] = v->clone();
	(*s)[buildKey(v)] = v->clone();
}

/**
 * get sprite scope variable
 **/
ValueObject*ObjectStore::getSpriteVariable(ScriptObject *so, wstring key) {
	wstring ctx = L"/"+buildKey(so);
	StringValueObjectHashMap *s = variables[ctx];
	if (s == NULL) {
		s = new StringValueObjectHashMap();
		variables[ctx] = s;
	} 

	ValueObject *v = (*s)[toUpper(key)];
	if (v != NULL) return v->clone();
	else return NULL;
}

/**
 * remove a sprite variable
 **/
void ObjectStore::removeSpriteVariable(ScriptObject *so,wstring key){
	key = toUpper(key);
	wstring ctx = L"/"+buildKey(so);
	StringValueObjectHashMap *s = variables[ctx];
	if ((*s)[key] != NULL) s->erase(buildKey((*s)[key]));
	s->erase(key);
}

/**
 * Add a reference to a cloned Object
 **/
void ObjectStore::addClonedObject(int characterId, ScriptObject *so) {
	// Check if already associated.
	RRange range = clonedObjects.equal_range(characterId);
	if (range.first != clonedObjects.end()) {
		for(RCIT i=range.first; i!=range.second; ++i) {
			if (i->second == so) return;
		}
	}
	clonedObjects.insert(make_pair(characterId,so));
}

/**
 * Remove a cloned Object
 **/
void ObjectStore::removeClonedObject(int characterId, ScriptObject *so) {
	RRange range = clonedObjects.equal_range(characterId);
	if (range.first != clonedObjects.end()) {
		RCIT toRemove, empty;
		for (RCIT i=range.first; i!=range.second; ++i) {
			if (i->second == so) {
				toRemove = i;
			}
		}
		if (toRemove != empty) clonedObjects.erase(toRemove);
	}
}

/**
 * Returns the list of cloned object for a given characterId
 **/
list<ScriptObject*>ObjectStore::getClonedObjects(int characterId) {
	list<ScriptObject*> res;
	RRange range = clonedObjects.equal_range(characterId);
	if (range.first != clonedObjects.end()) {
		for(RCIT i=range.first; i!=range.second; ++i) {
			res.push_back(i->second);
		}
	}
	return res;
}

/**
 * Add a removed object - PlaceFlagMove && PlaceFlagHasCharacter
 * For future reuse
 **/
void ObjectStore::addDeletedObject(ScriptObject *parent,int depth, int characterId, ScriptObject *so){
	char tmp[100]; 
	sprintf(tmp,"%p:%d_%d",parent,characterId,depth);
	string key = string(tmp);
	rubishStore[key] = so;
}

/**
 * Get a removed object
 **/
ScriptObject *ObjectStore::getDeletedObject(ScriptObject *parent,int depth, int characterId){
	char tmp[100]; 
	sprintf(tmp,"%p:%d_%d",parent,characterId,depth);
	string key = string(tmp);
	ScriptObject *so = rubishStore[key];
	if (so!= NULL) rubishStore.erase(key);
	return so;
}

/**
 * Remove an object - call the destructor of the object
 **/
void ObjectStore::destroyObject(ScriptObject*s) {
	allocatedObjects.erase(s);
	delete s;
}

/**
 * Remove a global variableREADME
 **/
void ObjectStore::removeGlobalVariable(wstring key) {
	StringValueObjectHashMap *s = variables[L"/"];
	s->erase(toUpper(key));
}

void ObjectStore::unRegisterNewObject(ScriptObject *p) { 
	//allocatedObjects.erase(p); 
	allocatedObjects[p] = NULL;
}

/**
 * Copy the current constant pool context
 * This is used for callback functions
 **/
void ObjectStore::copyCurrentPoolContext(wstring key) {
	IntValueObjectHashMap *m = constantPools[poolActionContext];
	IntValueObjectHashMap *dest = new IntValueObjectHashMap();
	if (m!=NULL) {
		for (IntValueObjectHashMap::iterator i=m->begin(); i != m->end(); i++) {
			ValueObject *v = i->second;
			(*dest)[i->first] = v->clone();
		}
	}
	constantPools[key] = dest;
}

/**
 * Reset all objects - except those created by action script
 **/
void ObjectStore::resetProgram() {
	for (IntScriptObjectMap::iterator i = objectAtDepth.begin(); i != objectAtDepth.end(); i++)
		if (i->second != NULL && ! i->second->getIsASCreated()) objectAtDepth.erase(i);
}

void  ObjectStore::setLinkage(wstring s, ScriptObject *so) {
	linkages[s] = so;
}

wstring	ObjectStore::getLinkage(wstring s) {
	if (linkages[s] == NULL) return s;
	return linkages[s]->getClassName();
}

void ObjectStore::addExportedAsset(wstring name, int characterId) { 
	exportedAssets[name] = characterId; 
	invExportedAssets[characterId] = name;
}

/**
 * Check if we have an Action Script class object for this script object
 * Action script class objects overides the default movieClip class implementation.
 **/
ScriptObject*ObjectStore::getActionScriptObjectClass(ScriptObject *so){
	if (invExportedAssets.find(so->getCharacterId()) == invExportedAssets.end()) return so;
	wstring s = invExportedAssets[so->getCharacterId()];
	if (linkages[s] == NULL) return so;
	return linkages[s];
}

ScriptObject*ObjectStore::getActionScriptObjectClass(wstring s) {
	return linkages[s];
}

// find something cleaner.
wstring	encodeCharacterName(int charId) { 
	char tmp[100]; 
	sprintf(tmp,"_nonAS:%d",charId); 
	return stringTowstring(tmp); 
}
