/**
Macromedia Flash Script execution stack engine.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include <string>
#include <iostream>
#include "script.hh"
#include "ASLib/as_linker.hh"
#include "ASLib/as_array.hh"
#include "ASLib/as_number.hh"
#include "ASLib/as_boolean.hh"
#include "ASLib/as_string.hh"
#include "ASLib/as_arguments.hh"
#include "ASLib/as_sound.hh"
#include "ASLib/as_text_field.hh"
#include "swf_engine.hh"

extern bool debug;
static wstring MovieProperties[]={L"_X",L"_Y",L"_xscale",L"_yscale",L"_currentframe",L"_totalframes",L"_alpha",L"_visible",L"_width",L"_height",
	L"_rotation",L"_target",L"_framesloaded",L"_name",L"_droptarget",L"_url",L"_highquality",L"_focusrect",L"_soundbuftime",
	L"_quality",L"_xmouse",L"_ymouse" };

Script::Script(ObjectStore *p, Graphics *g, SwfEngine *e) {
	objectStore = p;
	timer = e->getTimer();
	currentContext = objectStore->getActionContext();
	setGlobalRegisterBank();
	implicitReturnsTarget.push(NULL);
	implicitReturns.push(NULL);
	this->graphics = g;
	this->engine = e;
	asLinker = new ASLinker(this, graphics);
	gettimeofday(&tStart,NULL);
	rootScript = new ScriptObject(L"_root");
	getObjectStore()->registerNewObject(rootScript);
	swfVersion = e->getHeader().version;
	hasTargetDefined = false;
}

Script::~Script() {
	labels.clear();
	delete asLinker;
}

/**
 * Pushes a value to the stack.
 * A push command may push multiple values on the stack
 **/
bool Script::doActionPush(UI8 **p, int length) {
	bool res = true;
	while (length>0) {
		int code = int(**p);
		(*p)++;
		if (debug) printf("push (T:%d, ",code); fflush(stdout);
		int read = 0;
		switch (code) {
			case 0: { // string literal
				StringObject *c = new StringObject(); 
				read = c->decode(*p);
				execStack.push(c);
				break;
			}
    	    case 1: { // floating-point literal
				FloatObject *c = new FloatObject();
				read = c->decode(*p);
				execStack.push(c);
				break;
			}
    	    case 2: // null
				execStack.push(new StringObject(L"null"));
				read = 0;				
				break;
    	    case 3: // undefined
				execStack.push(new StringObject(L"undefined"));
				read = 0;
				break;
    	    case 4: { // register
				int registerIndex = **p;
				if (debug) printf("r:%d",registerIndex);
				read = 1;
				ValueObject *o = getRegister(registerIndex);
				if (o != NULL) execStack.push(o->clone()); else execStack.push(new StringObject(L"undefined"));
				break;
			}
    	    case 5: { // boolean
				BooleanObject *c = new BooleanObject();
				read = c->decode(*p);
				execStack.push(c);
				break;
			}
    	    case 6: { // double
				DoubleObject *c = new DoubleObject();
				read = c->decode(*p);
				execStack.push(c);
				break;
			}
    	    case 7: { // integer
				IntegerObject *c = new IntegerObject();
				read = c->decode(*p);
				execStack.push(c);
				break;
			}
    	    case 8: { // constant 8
				int cst = int(**p);
				if (debug) printf("c:%d",cst);
				read=1;
				execStack.push(objectStore->getConstant(cst));
				break;
			}
    	    case 9: { // constant 16
				int cst = int(GETUI16(*p)); 
				read=2;
				execStack.push(objectStore->getConstant(cst));
				break;
			}
			default: {
				wcout << L"Data type to pushed non recognized "<< code << L"\n";
				res = false;
			}
		}
		if (debug) wcout<<*(top())<<L") ";
		(*p) += read;
		length -= read+1;
		if (**p == 0 && *(*p+1) == 0) length = 0; // Force end of push sequence
	}
	return res;
}

/**
 * Push a new value on the context stack.
 **/
void Script::actionContextPush(ScriptObject *newObject,wstring movieName) {
	actionContextStack.push(currentContext);
	currentContext = newObject;
}

/**
 * Pop last Action context and restore last movieName path
 ***/
void Script::actionContextPop() {
	if (actionContextStack.size() > 0) 	actionContextStack.pop();
}

/**
 * Pops a value from the stack and discards it.
 **/
void Script::doActionPop(UI8 **p) {
	if (debug) wcout << L"pop";
	ValueObject *p1 = pop();
	removeObject(p1);
}

/**
 * Adds two numbers and pushes the result back to the stack.
 **/
void Script::doActionAdd() {
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	double result = p1->toNumeric()+p2->toNumeric();
	execStack.push(new FloatObject(result));
	if (debug) wcout << "Add" << " -> " << *execStack.top();
	removeObject(p1);
	removeObject(p2);
}

/**
 * subtract two objects (if possible)
 **/
void Script::doActionSubstract() {
	if (debug) wcout << L"Substract";
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	execStack.push(new FloatObject(p2->toNumeric()-p1->toNumeric()));
	removeObject(p1);
	removeObject(p2);
}

/**
 * Multiply two numbers and pushes the result back to the stack.ActionDiv
 **/
void Script::doActionMultiply() {
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	execStack.push(new FloatObject(p2->toNumeric()*p1->toNumeric()));
	if (debug) wcout << L"Multiply ->"<< *execStack.top();
	removeObject(p1);
	removeObject(p2); 
}

/**
 * Divide two numbers and pushes the result back to the stack.
 **/
void Script::doActionDivide() {
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	if (p1->toNumeric() != 0.0) { 
		execStack.push(new FloatObject(p2->toNumeric()/p1->toNumeric()));
	} else {
		execStack.push(new StringObject(L"NaN"));
	}
	if (debug) wcout << L"Divide ->" << *execStack.top();
	removeObject(p1);
	removeObject(p2);
}

/**
 * Compare two numerical objects.
 **/
void Script::doActionEqual() {
	if (debug) wcout << L"Equal";
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	execStack.push(p2->equal(p1));
	removeObject(p1);
	removeObject(p2);
}

/**
 * Compare two numerical objects.
 **/
void Script::doActionLess() {
	if (debug) wcout << L"Less";
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	execStack.push(new IntegerObject(p2->toNumeric() < p1->toNumeric() ? 1 : 0));
	removeObject(p1);
	removeObject(p2);
}

/**
 * Performs a logical AND of two numbers.
 **/
void Script::doActionAnd() {
	if (debug) wcout << L"And";
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	UI32 a1 = UI32(p1->toNumeric());
	UI32 a2 = UI32(p2->toNumeric());
	SI32 a = a1 & a2;
	execStack.push(new IntegerObject(a));
	removeObject(p1);
	removeObject(p2);
}

/**
 * Performs a logical OR of two numbers.
 **/
void Script::doActionOr() {
	if (debug) wcout << L"Or";
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	UI32 a1 = UI32(p1->toNumeric());
	UI32 a2 = UI32(p2->toNumeric());
	SI32 a = a1 | a2;
	execStack.push(new IntegerObject(a));
	removeObject(p1);
	removeObject(p2);
}

/**
 * Performs a logical NOT of two numbers.
 **/
void Script::doActionNot() {
	ValueObject *p1 = pop();
	if (debug) wcout << L"ActionNot "<< *p1;
	bool res;
	if (p1->getObjectType() == ScriptObjectType) {
		res = false;
	} else {
		res = p1->toNumeric() == 0;
	}
	execStack.push(new BooleanObject(res));
	removeObject(p1);
}

/**
 * Do bit and between to unsigned numbers
 **/
void Script::doActionBitAnd() {
	if (debug) wcout << L"BitAnd";
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	UI32 a1 = UI32(p1->toNumeric());
	UI32 a2 = UI32(p2->toNumeric());
	SI32 a = a1 & a2;
	execStack.push(new IntegerObject(a));
	removeObject(p1);
	removeObject(p2);
}

/**
 * Calculates x modulo y.
 **/
void Script::doActionModulo() {
	if (debug) wcout << L"Modulo";
	ValueObject *p2 = pop();
	ValueObject *p1 = pop();
	float a = p1->toNumeric();
	float b = p2->toNumeric();
	if (b != 0) {
		execStack.push(new FloatObject(fmod(a,b)));
	} else {
		execStack.push(new StringObject(L"NaN"));
	}
	removeObject(p1);
	removeObject(p2);
}

/**
 * Signed left shift
 **/
void Script::doActionBitLShift() {
	ValueObject *p2 = pop();
	ValueObject *p1 = pop();
	SI32 a1 = SI32(p1->toNumeric());
	UI32 a2 = UI32(p2->toNumeric()) & 0x1f;
	if (debug) wcout << L"BitLShift "<<a1<<L" << "<<a2;
	a1 <<= a2;
	execStack.push(new IntegerObject(a1));
	removeObject(p1);
	removeObject(p2);
}

/**
 *
 **/
void Script::doActionBitOr() {
	if (debug) wcout << L"BitOr";
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	UI32 a1 = UI32(p1->toNumeric());
	UI32 a2 = UI32(p2->toNumeric());
	SI32 a = a1 | a2;
	execStack.push(new IntegerObject(a));
	removeObject(p1);
	removeObject(p2);
}

/**
 * Push a duplicate of the current return value to the stack
 **/
void Script::doActionPushDuplicate() {
	if (debug) wcout << L"PushDuplicate";
	execStack.push(execStack.top()->clone());
}

/**
 * Initializes an array in a ScriptObject.
 **/
void Script::doActionInitArray() {
	if (debug) wcout << L"InitArray";
	ASArray *array = new ASArray(this,graphics);
	ValueObject *nbElem = pop();
	int nbElement = int(nbElem->toNumeric());
	array->allocate(nbElement);
	removeObject(nbElem);
	for (int i=0; i< nbElement; i++) {
		ValueObject *v = pop();
		IntegerObject *j = new IntegerObject(i);
		array->setMember(j->toString(),v);
		removeObject(j);
		removeObject(v);
	}
	execStack.push(array);
}

/**
 * Exact opposite of ActionLess2
 **/
void Script::doActionGreater() {
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	BooleanObject * r ;
	if (p1->getObjectType() == StringObjectType && p2->getObjectType() == StringObjectType) {
		r = p2->toString() > p1->toString() ? new BooleanObject(true) : new BooleanObject(false);
	} else {
		r = p2->toNumeric() > p1->toNumeric() ? new BooleanObject(true) : new BooleanObject(false);
	}
	execStack.push(r);
	if (debug) wcout << L"Greater "<< *r;
	removeObject(p1);
	removeObject(p2);
}

/**
 * Rigth signed shift
 **/
void Script::doActionBitRShift() {
	ValueObject *p2 = pop();
	ValueObject *p1 = pop();
	SI32 a1 = SI32(p1->toNumeric());
	UI32 a2 = UI32(p2->toNumeric()) & 0x1f;
	if (debug) wcout << L"BitRShift "<< a1 << " >> "<<a2;
	a1 >>= a2;
	execStack.push(new IntegerObject(a1));
	removeObject(p1);
	removeObject(p2);
}

/**
 * The spec is not very clear: Typo?
 * We assume it's unsigned shift
 **/
void Script::doActionBitURShift() {
	ValueObject *p2 = pop();
	ValueObject *p1 = pop();
	UI32 a1 = UI32(p1->toNumeric());
	UI32 a2 = UI32(p2->toNumeric()) & 0x1f;
	if (debug) wcout << L"BitRShift "<< a1 <<L" >> "<<a2;
	a1 >>= a2;
	execStack.push(new IntegerObject(a1));
	removeObject(p1);
	removeObject(p2);
}

/**
 * Bit Xor
 **/
void Script::doActionBitXor() {
	ValueObject *p2 = pop();
	ValueObject *p1 = pop();
	UI32 a1 = UI32(p1->toNumeric());
	double aa = p2->toNumeric();
	UI32 a2 = (UI32)aa;
	SI32 a = a1 ^ a2;
	if (debug) wcout << "BitXor "<< a1 << L" ^ " << a2 << L" -> " << a;
	execStack.push(new IntegerObject(a));
	removeObject(p1);
	removeObject(p2);
}

/**
 * Tests two strings for equality.
 **/
void Script::doActionStringEquals() {
	if (debug) wcout << L"StringEquals";
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	execStack.push(p2->stringEquals(p1));
	removeObject(p1);
	removeObject(p2);
}

/**
 * Computes the length of a string.
 **/
void Script::doActionStringLength() {
	if (debug) wcout << L"StringLength";
	ValueObject *p1 = pop();
	int a = p1->toString().length();
	execStack.push(new IntegerObject(a));
	removeObject(p1);
}

/**
 * Concatenates two strings.
 **/
void Script::doActionStringAdd() {
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	ValueObject *res = p2->add(p1);
	execStack.push(res);
	if (debug) wcout << L"StringAdd " << *p1 << L"+" << *p2 << L"->" << *res;
	removeObject(p1);
	removeObject(p2);
}

/**
 * Extracts a substring from a string.
 **/
void Script::doActionStringExtract() {
	if (debug) wcout << L"StringExtract";
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	ValueObject *p3 = top();
	if (p1->getObjectType() == IntegerObjectType && p2->getObjectType() == IntegerObjectType && p3->getObjectType() == StringObjectType) {
		int count = ((IntegerObject*)p1)->getValue();
		int index = ((IntegerObject*)p2)->getValue();
		((StringObject*)p3)->substring(index,count);
	} else {
		pop();
		removeObject(p3);
		execStack.push(new ValueObject());
	}
	removeObject(p1);
	removeObject(p2);
}

/**
 * String less
 **/
void Script::doActionStringLess() {
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	if (p1->getObjectType() == StringObjectType && p2->getObjectType() == StringObjectType) {
		execStack.push(new BooleanObject(((StringObject*)p2)->stringLess((StringObject*)p1)));
	} else {
		execStack.push(new BooleanObject(false));
	}
	removeObject(p1);
	removeObject(p2);
}

/**
 * Multibyte version: TODO: not sure
 **/
void Script::doActionMBStringLength() {
	doActionStringLength();
}

/**
 * Multibyte version: TODO: not sure
 **/
void Script::doActionMBStringExtract() {
	doActionStringExtract();
}

/**
 * Converts a value to an integer.
 **/
void Script::doActionToInteger() {
	ValueObject *p1 = pop();
	execStack.push(new IntegerObject(int(p1->toNumeric())));
	if (debug) wcout << L"ToInteger ->" << *execStack.top();
	removeObject(p1);
}

/**
 * Converts a character to ASCII code.
 **/
void Script::doActionCharToAscii() {
	ValueObject *p1 = pop();
	if (p1->getObjectType() == StringObjectType) {
		int v = ((StringObject*)p1)->getValue()[0];
		execStack.push(new IntegerObject(v));
	} else {
		execStack.push(new IntegerObject(0));
	}
	removeObject(p1);
}

/**
 * Convert an ASCII code to a character
 **/
void Script::doActionAsciiToChar() {
	ValueObject *p1 = pop();
	if (p1->getObjectType() == IntegerObjectType) {
		UI8 c[2];
		c[0] = ((IntegerObject*)p1)->getValue();
		c[1] = 0;
		execStack.push(new StringObject(utf8Towstring((const char*)c)));
	} else {
		execStack.push(new ValueObject());
	}
	removeObject(p1);
}

/**
 * TODO: fix it
 */
void Script::doActionMBCharToAscii() {
	doActionCharToAscii();
}

void Script::doActionMBAsciiToChar() {
	doActionAsciiToChar();
}

/**
 * Creates new constants in the action pool.
 * NOTES: 
 * The action context for a constant pool is defined a the object type level, not at the object instance level.
 * On the contrary, members are a property on an object instance.
 **/
void Script::doActionConstantPool(UI8 **p) {
	int count = GETUI16(*p);
	(*p)+=sizeof(UI16);
	if (debug) wcout << L"New constants: [" << count << L"] ";
	for (int i=0; i< count; i++) {
		wstring key = utf8Towstring((const char*)*p,strlen((char*)*p));
		objectStore->setConstant(i, new StringObject(key));
		if (debug) wcout << key ;
		(*p)+=strlen((char*)*p)+1;
	}
}

/**
 * Define a new function
 **/
void Script::doActionDefineFunction(UI8 **p) {
	wstring functionName;
	bool functionLiteral = false;
	if (**p == 0) {		// Anonymous
		if (execStack.size()>0) functionName = top()->toString();
		if (functionName.size() == 0) functionName = L"anonymous";
		(*p)++;
		functionLiteral = true;
	} else {
		functionName = utf8Towstring((const char*)*p,strlen((char*)*p));
		(*p)+=strlen((char*)*p)+1;
	}
	Function *f = new Function();
	//if (!functionLiteral) f = new Function(); else f = new FunctionLiteral();

	f->setFunctionName(functionName);
	f->setNumParams(GETUI16(*p));
	(*p)+=sizeof(UI16);
	wstring *s = new wstring[f->getNumParams()];
	if (debug) wcout << L"Define function: " << functionName << L"(";
	for (int i=0; i< f->getNumParams(); i++) {
		s[i] = utf8Towstring((const char*)*p,strlen((char*)*p));
		(*p)+=strlen((char*)*p)+1;
		if (debug) wcout << s[i] << L",";
	}
	f->setParams(s);
	if (debug) wcout << L") ";
	int offset = GETUI16(*p);
	f->setCodeSize(offset);
	(*p)+=sizeof(UI16);
	f->setCode(*p);
	if (functionLiteral) {
		// Create a literal function object
		if (debug) wcout << L"immediate [" << *p << L"]";
		f->setPrototype(f);
		f->setTargetPath(objectStore->getCurrentTargetPath());
		push(f);
	}
	f->setInlined(functionLiteral);
	f->setObjectInstance(currentContext);
	(*p)+=offset;
	objectStore->addFunction(L"_global",functionName,f);

	// Create an associated object property
	ASObject *a = new ASObject(this,graphics,functionName);
	a->setParent((ASInterface*)currentContext);
	a->setMemberNoClone(L"function",f);
	objectStore->setGlobalVariable(functionName,a);
}


/**
 * Reads the next object from the stack (without popping it) and stores it in a register
 **/
void Script::doActionStoreRegister(UI8 **p) {
	int regNum = int(**p);
	(*p)++;
	setRegister(regNum,top());
	if (debug) wcout << L"StoreRegister r:" << regNum << L" = " << *top();
}

/**
 * Call a Method.
 * It does:
 * - get the constructor of the SuperClass or the class if no inheritance
 * - prepare the new execution environment (_doCall)
 * - push the return PC adress
 * - do the call
 **/
void Script::doActionCallMethod(UI8 **p ) {
	StringObject *name = (StringObject*)pop();
	wstring methodName = name->getValue();
	ScriptObject* sc = (ScriptObject*)pop(); 
	if (methodName == L"undefined") {
		methodName = L"prototype"; // If undefined call constructor (this)
	}
	if (sc->toString() == L"undefined") sc = objectStore->getActionContext();
	if (sc->toString() == L"_root") {
		//sc = objectStore->getMovieByNameOrPath(L"_root");	// Do do implement _root nesting
		//if (sc == NULL) {
		//	sc = new ASObject(this,graphics,L"_root");
			//objectStore->setRenderedObject(L"_root",sc);
		//}
	}

	if (debug) wcout << "CallMethod " << *sc << ":" << methodName;
	ValueObject *numParams = (ValueObject *)pop();
	if (debug) wcout << L" args=" << numParams->toNumeric();
	ValueObject** s = new ValueObject*[int(numParams->toNumeric())];
	for (int i=0; i< int(numParams->toNumeric()); i++) {
		s[i] = pop();
		if(debug) wcout<<*(s[i])<<L", ";
	}
	Function* func = NULL;
	bool needRemoveFunc = false;

	// First look in ScriptObject Members
	ValueObject *member = NULL;
	if (sc->getObjectType() == ScriptObjectType) member = sc->getMember(methodName); 

	if (member != NULL) {
		//wcout << L"Member " << member->toString() << L"\n"; fflush(stdout);
		if (member->toString() != L"undefined") func = objectStore->getFunction(L"_global",member->toString());
	}

	// Try in _global
	if (func == NULL) func  = objectStore->getFunction(L"_global",methodName);

	// Then look in by instance Name
	if (func == NULL) func  = objectStore->getFunction(sc->getThisObjectInstanceName(),methodName);

	// Then look for a builtin function
	if (func == NULL) {
		func = _callBuiltInFunction(&sc,methodName);
		// Don't cache this function! We need to transform object types if required
		needRemoveFunc = true;
	}

	// Then try by Object Accessor
	if (func == NULL) {
		if (methodName == L"addProperty") 	func = asLinker->getFunctionMember(L"Object",L"addProperty",NULL);
		if (methodName == L"registerClass") func = asLinker->getFunctionMember(L"Object",L"registerClass",NULL);
		if (methodName == L"toString") 		func = asLinker->getFunctionMember(L"Object",L"toString",NULL);
		if (methodName == L"unwatch") 		func = asLinker->getFunctionMember(L"Object",L"unwatch",NULL);
		if (methodName == L"watch") 		func = asLinker->getFunctionMember(L"Object",L"watch",NULL);
	}

	if (sc->toString() == L"_root") func = NULL; ////////////// FIXME. Implement the _root context
	if (func == NULL) {
		wcout << "Function NOT FOUND!!!!\n";
		prepareExecEnv(sc);
		execStack.push(new ValueObject());
		return;
	}

	if (sc->getThisObjectInstanceName() == methodName) {
		wstring superClassName = sc->getPrototype()->toString();
		Function *subClass = func;
		_doCall(func,int(numParams->toNumeric()),s,sc);
		if (!func->getIsNative()) {
			UI8* e = func->getCode() + func->getCodeSize();
			implicitReturnsTarget.push(e);
			implicitReturns.push(*p);

			// Pass parameters.
			// As we may have neested contructor, we need to pass the original parameters to each nested constructors.
			// The functions registers may be in different orders, so we shuffle them according to function2 definition.
			RegisterSet *rd = func->getRegisters();
			RegisterSet *rs = subClass->getRegisters();
	
			for (int i=0; i< func->getNumParams(); i++) {
				int dreg = func->getParamRegisters()[i];
				int sreg = subClass->getParamRegisters()[i];
				ValueObject * src = rs->getRegister(sreg);
				rd->setRegister(dreg,src);
			}
		} 
		execStack.push(new ValueObject());
	} else {
		// Set the Argument properties
		ASArguments *ag = (ASArguments*)getObjectStore()->getGlobalVariable(L"_ASArgument");
		if (ag == NULL) {
			ag = new ASArguments(this,graphics);
			objectStore->setGlobalVariable(L"_ASArgument",ag);
			objectStore->registerNewObject(ag);
		}
		ag->setMemberNoClone(L"callee",new StringObject(methodName));
		ag->setMemberNoClone(L"caller",new StringObject(L"undefined")); // TODO. Keep a method calling stack
		ag->setMemberNoClone(L"length",new IntegerObject(int(numParams->toNumeric())));

		// If function is Native (built-in), we pass parametersgraphics and set execution context
		// Then call the built-in function
		// We then restore the execution environment to the caller one
		if (func->getIsNative()) {
			_doCall(func,int(numParams->toNumeric()),s,sc);
			asLinker->callFunction(sc,func);
			if (methodName == L"prototype") execStack.push(new ValueObject());
			leaveExecEnv();
		} else {		
			// Jump!
			_doCall(func,int(numParams->toNumeric()),s,sc);
			UI8* e = func->getCode() + func->getCodeSize();
			implicitReturnsTarget.push(e);
			implicitReturns.push(*p);
			execStack.push(new StackPointerObject(*p));
		}
	} 
	if (!func->getIsNative()) *p = func->getCode();

	removeObject(name);
	for (int i=0; i< int(numParams->toNumeric()); i++) 	removeObject(s[i]);
	removeObject(numParams);
	delete []s;
	removeObject(sc);
	if (needRemoveFunc) delete func;
}

/**
 * Defines a local variable
 **/
void Script::doActionDefineLocal() {
	ValueObject *value = pop();
	StringObject *name  = (StringObject*)pop();
	objectStore->addLocalVariable(name->getValue(),value);
	if (debug) wcout<<L"DefineLocal "<<*name<<L":="<<*value;
	removeObject(value);
	removeObject(name);
}

/**
 * Defines a local variable without setting its value.
 **/
void Script::doActionDefineLocal2() {
	if (debug) wcout << L"DefineLocal2 ";
	ValueObject *value = new ValueObject();
	ValueObject *name  = pop();
	wstring key = L"";
	switch (name->getObjectType()) {
		case StringObjectType: 		key = ((StringObject*)name)->getValue(); break;
		case Constant8ObjectType:	key = ((Constant8Object*)name)->getPoolEntry(); CRITICAL_ERROR(L"???");break; // ???
		case Constant16ObjectType: 	key = ((Constant16Object*)name)->getPoolEntry(); CRITICAL_ERROR(L"???");break; // ???
	}
	objectStore->addLocalVariable(key,value);
}

/**
 * Creates a new Object instance and invokes the constructor.
 **/
void Script::doActionNewObject(UI8 **p) { 
	StringObject *objectName  = (StringObject*)pop();
	if (debug) wcout << L"NewObject: "<< objectName->toString();
	IntegerObject *numParams = (IntegerObject *)pop();
	ValueObject** s = new ValueObject*[int(numParams->toNumeric())];
	for (int i=0; i< int(numParams->toNumeric()); i++) {
		s[i] = pop();
		if (debug) wcout<<*(s[i]);
	}
	wstring constructorName = objectStore->getLinkage(objectName->toString());
	if (debug) wcout << L" Function Name: " << constructorName << L" ";
	Function *func = objectStore->getFunction(constructorName,constructorName);
	ScriptObject *so = NULL;
	if (func == NULL) {
		func= asLinker->getFunctionMember(objectName->toString(),objectName->toString(),&so);
		objectStore->addFunction(objectName->toString(),objectName->toString(),func);
	}
	if (func == NULL) {
		func = objectStore->getFunction(L"_global",objectName->toString());
	}
	if (func == NULL) {
		wcout << "***** Constructor for class " << constructorName << L" not found!\n";
	} else {
		// Push the newly constructed object & prepare execution environment.
		if (!func->getIsNative()) {
			// Look is we have an existing object
			so = (ScriptObject*)getEngine()->getGlobal()->getMember(objectName->toString());
			if (so == NULL) {
				so = new ASObject(this,graphics,objectName->toString());
				getObjectStore()->registerNewObject(so);
			}
		} else {
			so = asLinker->getANewClassInstance(this,graphics,objectName->toString());
		}
		so->setPrototype(so);
		so->setParent(currentContext);
		execStack.push(so);
		_doCall(func,numParams->getValue(),s,so);
		if (func ->getIsNative()) {
			asLinker->callFunction(so,func);
			leaveExecEnv();
		} else {
			implicitReturnsTarget.push(func->getCode() + func->getCodeSize());
			implicitReturns.push(*p);
			*p = func->getCode();
		}
	}
	if (debug) wcout <<L" ->"<<*so;

	removeObject(objectName);
	for (int i=0; i< int(numParams->toNumeric()); i++) removeObject(s[i]);
	removeObject(numParams);
	delete []s;
}


/**
 *  We implement FULL ECMA-262 (Section 11.6.1)
 **/
void Script::doActionAdd2() {
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	if (debug) wcout << L"Add2: "<<*p2<<L"+"<<*p1;
	if (p1->getObjectType() == ScriptObjectType || p2->getObjectType() == ScriptObjectType) {
		wstring res = p2->toString()+p1->toString();
		execStack.push(new StringObject(res));		
	} else if (p1->getObjectType() == StringObjectType || p2->getObjectType() == StringObjectType) {
		if (p1->toString() == L"undefined") {
			if (p2->getThisObjectInstanceName() == L"_$$$Number") {
				if (swfVersion < 7) execStack.push(new FloatObject(p2->toNumeric()));
				else execStack.push(new StringObject(L"NaN"));
			} else {
				execStack.push(new StringObject(L"undefined"));
			}
		} else if (p2->toString() == L"undefined") {
			if (p1->getThisObjectInstanceName() == L"_$$$Number") {
				if (swfVersion < 7) execStack.push(new FloatObject(p1->toNumeric()));
				else execStack.push(new StringObject(L"NaN"));
			} else {
				execStack.push(new StringObject(L"undefined"));
			}
		} else {
			execStack.push(new StringObject(p2->toString()+p1->toString()));
		}
	} else {
		execStack.push(new FloatObject(p1->toNumeric()+p2->toNumeric()));
	}
	if (debug) wcout << L" ->" << *execStack.top();
	removeObject(p1);
	removeObject(p2);
}

/**
 * See ECMA-262 Section 11.8.5
 * TODO: implements NaN, infinite.
 **/
void Script::doActionLess2(){
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	BooleanObject * r ;
	if (p1->getObjectType() == StringObjectType && p2->getObjectType() == StringObjectType) {
		r = p2->toString() < p1->toString() ? new BooleanObject(true) : new BooleanObject(false);
	} else {
		r = p2->toNumeric() < p1->toNumeric() ? new BooleanObject(true) : new BooleanObject(false);
	}
	execStack.push(r);
	if (debug) wcout<<L"Less2 "<<*r;
	removeObject(p1);
	removeObject(p2);
}

/**
 * See ECMA-262 equal
 **/
void Script::doActionEquals2() {
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	if (debug) wcout << "Equals2 "<< *p1 << L" == " << *p2;
	if (p1->getObjectType() == StringObjectType ||  p2->getObjectType() == StringObjectType) {
		execStack.push(new BooleanObject(p1->toString() == p2->toString()));
	} else {
		execStack.push(new BooleanObject(p1->toNumeric() == p2->toNumeric()));
	}
	if (debug) wcout << L" ->" << *execStack.top();
	removeObject(p1);
	removeObject(p2);
}

/**
 * Creates a conditional test and branch.
 **/
void Script::doActionIf(UI8 **p){
	ValueObject *p1 = pop();
	bool doJump = false;
	if (p1->getObjectType() == BooleanObjectType) {
		doJump = ((BooleanObject*)p1)->getValue();
	} else {
		doJump = p1->toNumeric() != 0;
	}
	if (doJump) {
		int displacement= GETINT16(*p);
		(*p) += displacement + sizeof(UI16);
		if (debug) wcout << L"If. jump: " << displacement;
	} else {
		(*p) += sizeof(UI16);
		if (debug) wcout << L"If. noJUMP";
	}
	removeObject(p1);
}

/**
 * Pops a value from the stack, converts it to number type, increments it by 1
 **/
void Script::doActionIncrement(){
	if (debug) wcout << L"Increment ";
	ValueObject *p1 = pop();
	execStack.push(new FloatObject(p1->toNumeric()+1));
	removeObject(p1);	
}

/**
 * Pops a value from the stack, converts it to number type, decrements it by 1
 **/
void Script::doActionDecrement() {
	if (debug) wcout << L"Decrement ";
	ValueObject *p1 = pop();
	execStack.push(new FloatObject(p1->toNumeric()-1));
	removeObject(p1);	
}

/**
 * Creates an unconditional branch.
 **/
void Script::doActionJump(UI8 **p) {
	if (debug) wcout << L"Jump : "<<GETINT16(*p);
	(*p) += GETINT16(*p) + sizeof(UI16);
}

/**
 * Returns time since flashPlayer was started
 **/
void Script::doActionGetTime() {
	if (debug) wcout << L"getTime";
	struct timeval tEnd;
	gettimeofday(&tEnd,NULL);
	uint32_t t;
	t = (tEnd.tv_usec - tStart.tv_usec)/1000;
	t += (tEnd.tv_sec - tStart.tv_sec)*1000;
	execStack.push(new IntegerObject(t));
	if (debug) wcout << L"getTime ->" << t ;
}

/**
 * Calculates a random number.
 **/
void Script::doActionRandomNumber(){
	ValueObject *maxValue = pop();
	float val = float(rand()) / float (RAND_MAX) * maxValue->toNumeric();
	execStack.push(new FloatObject(val));
	if (debug) wcout << "Random " << *execStack.top();
}

/**
 * Gets a movie property.
 **/
void Script::doActionGetProperty() {
	ValueObject *sindex = pop();
	ValueObject *target = pop();
	int index = int (sindex->toNumeric());
	// if target is NULL, that defines the current action script. It may be defined as null in a ActionConstantPool tag
	ScriptObject*so = NULL;
	if (target->toString() != L"") so = objectStore->getMovieByNameOrPath(target->toString());
	if( so==NULL) {
		printf("WARN: get property from context ");
		so = currentContext;
	}
	ValueObject *v = so->getMember(MovieProperties[index]);
	if (v == NULL) v = new StringObject(L"undefined");
	execStack.push(v);
	if (debug) wcout << "GetProperty "<< *so << L"[" << MovieProperties[index] << L"] -> " << *top() ;
	removeObject(target);
	removeObject(sindex);
}

/**
 * Set a movie property.
 **/
void Script::doActionSetProperty() {
	ValueObject *value  = pop();
	ValueObject *sindex = pop();
	ValueObject *target = pop();
	int index = int (sindex->toNumeric());
	ScriptObject *so = NULL;
	wstring starget = target->toString();
	
	// Try to get a movie by movie Name
	so = objectStore->getMovieByNameOrPath(starget);
	if (so == NULL) so = currentContext;

	so->setMember(MovieProperties[index],value);
	so->setDirty();
	if (debug) wcout << L"SetProperty " << *so << L"[" << MovieProperties[index] << L"]=" << *value;
	removeObject(value);
	removeObject(sindex);
	removeObject(target);
}

/**
 * Prepare the execution environment to execute a function
 *
 **/
void Script::_doCall(Function* func, int numArgs, ValueObject** args,ScriptObject *scriptObject) {
	if( func == NULL) {
		CRITICAL_ERROR(L"Function not found");
	}
	RegisterSet *rs;
	// If Function2(flash7) allocate new registers set and do register preloading
	if (func->getIsFunction2()) {
		if (func->getRegisters() == NULL) {
			rs = new RegisterSet(func->getRegisterCount()); 
		} else {
			rs = func->getRegisters();
			for (int i=0; i< rs->getRegisterCount(); i++) {
				if (rs->getRegister(i) != NULL) {
					removeObject(rs->getRegister(i));
					rs->setRegister(i,NULL);
				}
			}
		}

		int j = 1;
		if (func->getPreloadThisFlag()) {
			rs->setRegister(j, scriptObject);
			//printf("Preload ThisFlag");
			j++;
		}
		if (func->getPreloadArgumentsFlag()) {
			printf("Preload ArgumentsFlag: todo");
			j++;
		}
		if (func->getPreloadSuperFlag()) {
			rs->setRegister(j, scriptObject->getPrototype());
			//printf("Preload SuperFlag in r%d ",j);
			j++;
		}
		if (func->getPreloadRootFlag()) {
			rs->setRegister(j, rootScript); 
			printf("Preload RootFlag in r:%d ",j);
			j++;
		}
		if (func->getPreloadParentFlag()) {
			printf("Preload ParentFlag: todo");
			j++;
		}
		if (func->getPreloadGlobalFlag()) {
			printf("Preload GlobalFlag: todo");
			j++;
		}

		for (int i=0; i< numArgs; i++) {
			if (i < func->getNumParams()) {
				int reg = func->getParamRegisters()[i];
				if (reg == 0) {
					objectStore->setVariable(scriptObject,func->getParam(i),args[i]);
				} else {
					rs->setRegister(reg,args[i]);
				}
			}
		}
	} else {													// Non Function2 : non flash7 specific 
		for (int i=0; i< numArgs; i++) {
			if (i < func->getNumParams()) {
				objectStore->setVariable(scriptObject,func->getParam(i),args[i]);
			}
			if (debug) wcout<<*(args[i]);
			// Don't delete variables, they will be freed when exiting the func
		}
		rs = objectStore->getGlobalRegisters(); //->clone();
	}

	prepareExecEnv(scriptObject,func->getConstantPoolKey());
	func->setRegisters(rs);
	if (func->getIsFunction2()) setRegisterBank(rs);
	func->setNumPassedParams(numArgs);
	func->setObjectInstance(scriptObject); // required when the class is not instancianted by a new (like string)
}

/**
 * Executes a function.
 **/
void Script::doActionCallFunction(UI8 **p) {
	StringObject *funcName = (StringObject *)pop();
	IntegerObject *numArgs = (IntegerObject*)pop();
	ValueObject** s = new ValueObject*[int(numArgs->toNumeric())];
	for (int i=0; i< int(numArgs->toNumeric()); i++) s[i] = pop();

	if (debug) wcout << L"Call function:" << funcName->getValue() << L" "<< numArgs->getValue() << L"\n";
	ScriptObject *sc = currentContext;
	Function* func = objectStore->getFunction(objectStore->getPoolActionContext(),funcName->getValue());
	if (func == NULL || (func!=NULL && func->getIsNative()) ){		// if native need to reload the target script object.
		func = asLinker->getCallFunction(funcName->getValue(),&sc);
		if (func != NULL) objectStore->addFunction(objectStore->getPoolActionContext(),funcName->getValue(),func);
	}
	
	// Updates the display (independent of the frames per second set for the movie) when you call it within an onClipEvent()
	if (func == NULL && funcName->getValue() == L"updateAfterEvent") {
		wcout << L"execute doDisplay()\n";
		prepareExecEnv(sc);
		engine->doDisplay();
		leaveExecEnv();
		push(new ValueObject());
	} else if (func == NULL && funcName->getValue() == L"alert") {
		wcout << *s[0];
		push(new ValueObject());
	} else {
		if (func == NULL) {
			puts("ERROR function not found!!!!!!!");
			return;
		}
		if (func->getIsNative()) {
			_doCall(func,int(numArgs->toNumeric()),s,sc);
			asLinker->callFunction(sc,func);
			leaveExecEnv();
		} else {
			// Jump!
			_doCall(func,int(numArgs->toNumeric()),s,sc);
			UI8* e = func->getCode() + func->getCodeSize();
			implicitReturnsTarget.push(e);
			implicitReturns.push(*p);
			execStack.push(new StackPointerObject(*p));
			*p = func->getCode();
			if (func->getIsInlined()) objectStore->setCurrentTargetPath(func->getTargetPath());
		} 
	}
	ValueObject *o = (ValueObject*) funcName;
	removeObject(o);
	for (int i=0; i< int(numArgs->toNumeric()); i++) removeObject(s[i]);
	removeObject(numArgs);
	delete []s;
}

/**
 * Enhanced DefineFunction Tag specific to Flash7
 **/
void Script::doActionDefineFunction2(UI8 **p) {
	if (debug) wcout << L"DefineFunction2 ";
	wstring functionName;
	bool immediate = false;
	if (**p == 0) {
		functionName = execStack.size() > 0 ? top()->toString() : L"null";
		if (functionName==L"null") functionName=L"anonymous";
		immediate = true;
		(*p)++;
	} else {
		functionName = utf8Towstring((const char*)*p,strlen((char*)*p));
		(*p)+=strlen((char*)*p)+1;
	}
	Function *f = new Function();
	f->setIsFunction2(true);
	f->setFunctionName(functionName);
	f->setNumParams(GETUI16(*p));
	(*p)+=sizeof(UI16);
	f->setRegisterCount(**p);
	if (debug) wcout << f->getRegisterCount() << L" regs ";
	(*p)++;
	wstring *s = new wstring[f->getNumParams()];
	UI8 caps1 = **p; (*p) ++;
	f->setPreloadParentFlag(((caps1>>7)&1) == 1);
	f->setPreloadRootFlag(((caps1>>6)&1) == 1);
	f->setSuppressSuperFlag(((caps1>>5)&1) == 1);
	f->setPreloadSuperFlag(((caps1>>4)&1) == 1);
	f->setSuppressArgumentsFlag(((caps1>>3)&1) == 1);
	f->setPreloadArgumentsFlag(((caps1>>2)&1) == 1);
	f->setSuppressThisFlag(((caps1>>1)&1) == 1);
	f->setPreloadThisFlag((caps1&1) == 1);
	UI8 caps2 = **p; (*p) ++;
	f->setPreloadGlobalFlag((caps2&1) == 1);
	if (debug) wcout << L"_global" << L":" << functionName << L"()";
	UI8 *paramRegisters = new UI8[f->getNumParams()];
	if (debug) wcout << f->getNumParams() << L" parameters ";
	for (int i=0; i< f->getNumParams(); i++) {
		paramRegisters[i] = **p; 
		(*p)++;
		s[i] = utf8Towstring((const char*)*p,strlen((char*)*p));
		(*p)+=strlen((char*)*p)+1;
		if (debug) wcout << s[i] << L"(" << paramRegisters[i] << "),";
	}
	f->setParams(s);
	f->setParamRegisters(paramRegisters);
	if (debug) wcout << L") ";
	int offset = GETUI16(*p);
	f->setCodeSize(offset);
	(*p)+=sizeof(UI16);
	f->setCode(*p);
	wstring objectClassName = L"_global"; // A function is always global ? currentContext->getThisObjectInstanceName();
	objectStore->addFunction(objectClassName,functionName,f);
	ASObject *so = new ASObject(this,graphics,functionName);
	if (!immediate) {
		(*p)+=offset;
	} else {
		so->setMovieName(functionName);
		execStack.push(so);
		(*p)+=offset;
	}
	objectStore->setGlobalVariable(functionName,so);
}

/**
 * Function return
 **/
void Script::doActionReturn(UI8 **p) {
	if (debug) wcout << "Return";
	ValueObject *ret = pop();

	while (execStack.size() > 0 && top()->getObjectType() != StackPointerObjectType) {
		dumpStack();
		wcout << L"\n *********** Stack ERROR: pop one element:"<<*top()<<L"\n";
		pop();
	}
	StackPointerObject *sp = (StackPointerObject*)pop();
	*p = sp->getValue();
	//if (ret->getObjectType() != ValueObjectType) execStack.push(ret); else delete ret;
	// When should it be kept? Not clear in the specs
	execStack.push(ret);
	leaveExecEnv();
	if (implicitReturns.size() > 0) {
		implicitReturnsTarget.pop();
		implicitReturns.pop();
	}
	delete sp;
}

/**
 * Prepare execution environment
 **/
void Script::prepareExecEnv(ScriptObject *scriptObject) {
	// Set new ActionContext, new PoolContext, new register bank and new LocalVariable scope
	// Push actionContext, poolActionContext and registers. this handle the fact that an object may call another object inside
	// one if a method.
	objectStore->targetPathDown(scriptObject);
	actionContextPush(scriptObject, scriptObject->toString());
	poolActionContextStack.push(objectStore->getPoolActionContext());
	registerBankStack.push(currentRegisterSet);
	objectStore->setActionContext(scriptObject);
	objectStore->setPoolActionContext(scriptObject->getMovieName());
}

/**
 * Same as above, but forced by a function
 **/
void Script::prepareExecEnv(ScriptObject *scriptObject, wstring key) {
	objectStore->targetPathDown(scriptObject);
	actionContextPush(scriptObject, scriptObject->toString());
	poolActionContextStack.push(objectStore->getPoolActionContext());
	registerBankStack.push(currentRegisterSet);
	objectStore->setActionContext(scriptObject);
	objectStore->setPoolActionContext(key);
}

/**
 * Restore execution environment to the caller
 **/
void Script::leaveExecEnv() {
	objectStore->cleanCurrentLevel();	// Clean function local vars
	objectStore->targetPathUp();		// Change variable scope
	if (actionContextStack.size() > 0) objectStore->setActionContext(actionContextStack.top());
	if (registerBankStack.size() > 0) setRegisterBank(registerBankStack.top());
	if (poolActionContextStack.size() > 0) objectStore->setPoolActionContext(poolActionContextStack.top());
	if (actionContextStack.size() > 0) currentContext = actionContextStack.top();
	actionContextPop();
	if (registerBankStack.size() > 0) registerBankStack.pop();
	if (poolActionContextStack.size() > 0) poolActionContextStack.pop();
}

/**
 * Gets a variable value.
 **/
void Script::doActionGetVariable() {
	ValueObject *name = pop();
	wstring skey = name->toString();

	// Look in Local scope
	ValueObject *so = objectStore->getVariable(currentContext,skey);

	// If this is the "this" keywords, it refers to current movie.
	if (so == NULL && skey == L"this") {
		so = currentContext;
	}

	// Then look in global scope
	if (so == NULL) {
		if (skey == L"_global") so = (ScriptObject*) objectStore->getGlobalVariable(skey);
		else {
			ScriptObject *soglobal = (ScriptObject*) objectStore->getGlobalVariable(L"_global");
			so = (ScriptObject*) soglobal->getMember(skey);
		}		
	}

	// Look in _root scope
	if (so == NULL && skey == L"_root") {
		so = (ScriptObject*) objectStore->getGlobalVariable(L"_global");			// TODO: implement _root
	}

	// Then look for a Static class
	if (so == NULL) {
		so = asLinker->getClassSingleton(skey);
	}

	// Then look for a Dynamic class
//	if (so == NULL) {
//		Function *f =  objectStore->getFunction(objectStore->getPoolActionContext(),skey);
//		if (f != NULL) {
//		}
//	}

	// Then look for a Movie
	if (so == NULL) {
		so = objectStore->getMovieByNameOrPath(skey);
	}

	if (skey == L"_parent") { // TODO: Process it at class level
		so = currentContext->getParent();
	}

	// If that's a string, that could be the name of a movie clip. So we check this condition.
	if (so!= NULL && so->getObjectType() == StringObjectType) {
		ScriptObject *k = objectStore->getMovieByNameOrPath(so->toString());
		if (k != NULL) so = k;
	}

	if (so == NULL) {
		if (debug) wcout << L"WARNING doActionGetVariable. Variable not found!\n";
		so = new StringObject(L"undefined");
	}
	execStack.push(so);
	removeObject(name);	
	if (debug) wcout << L"GetVariable "<< skey << L"->" << *(execStack.top()) << L" "<< execStack.top()->toNumeric();
	if (debug) if ( execStack.top()->getObjectType() == ScriptObjectType) wcout << L" " << ((ScriptObject*)execStack.top())->getMovieName();
}

/*
 * Set a variable
 **/
void Script::doActionSetVariable() {
	ValueObject *value  = pop();
	ValueObject *name  = pop();
	if (debug) wcout << "SetVariable " << *name << L" := " << *value << L" ";

	if (name->getObjectType() == StringObjectType) {
		wstring key = ((StringObject*)name)->getValue();

		// Look if the object is an ScriptObject
		ScriptObject *so = objectStore->getMovieByNameOrPath(key);
		if (so == NULL) {
			objectStore->setVariable(currentContext,key, value);
		} else {
			if (debug) wcout<<*so;
			objectStore->setVariable(currentContext,key, value);
			so = objectStore->getMovieByNameOrPath(key);			// Reload in case variabled freed in setVariable.
			if (so->getMovieName() == key) so->setMainMember(value);
			else so->setMember(key,value);
			so->setDirty();
			removeObject(so);
		}

		if (debug) wcout<<L" "<<key<<L"="<<*value;
	} else {
		name->toString();
		printf("&NON StringObjectType! : %d",name->getObjectType()); name->toString();
	}
	removeObject(name);
	removeObject(value);
}

/**
 * Retrieves a named property from an object.
 **/
void Script::doActionGetMember() {
	if (debug) wcout << "GetMember ";
	StringObject *memberName = (StringObject*)pop(); 
	StringObject *scriptObject = (StringObject*)pop();
	wstring skey = scriptObject->toString();
	bool forceDelete = false;

	ScriptObject *so = NULL;
	ValueObject  *v = NULL;
	// We have a true scriptObject, not an alias
	if (scriptObject->getObjectType() == ScriptObjectType) so = (ScriptObject*) scriptObject;

	// If it's global, the member is a global variable ie: an MovieClip.
	if (skey == L"_global") {
		ScriptObject *p = (ScriptObject*) objectStore->getGlobalVariable(memberName->toString());
		if (p!=NULL) {
			so = p;
			v = so;
		} 
	}

	/** 
	* Look at That!	
	* The object parameter may not actually be an “object” type. If the object parameter is a primitive
	* type such as number, boolean or string, it is converted automatically to a temporary wrapper
	* object of class Number, Boolean or String. Thus, methods of wrapper objects may be invoked on
	* values of primitive types. 
	**/
	if (scriptObject->getThisObjectInstanceName() == L"_$$$Number") {
		so = new ASNumber(this,graphics,scriptObject->toNumeric());
		forceDelete = true;
	}
	if (scriptObject->getThisObjectInstanceName() == L"_$$$String") {
		so = new ASString(this,graphics,scriptObject->toString());
		forceDelete = true;
	}
	if (scriptObject->getThisObjectInstanceName() == L"_$$$Boolean") {
		so = new ASBoolean(this,graphics,((BooleanObject*)scriptObject)->getValue());
		forceDelete = true;
	}

	if (so == NULL) CRITICAL_ERROR(L" SANITY CHECK doActionGetMember");
	if (v == NULL) {
		v = so->getMember(memberName->toString());
	}

	if (v == NULL) v = objectStore->getVariable(so,memberName->toString());
	if (v == NULL) v = new StringObject(L"undefined");
	execStack.push(v);
	if (debug)  wcout << *so << L"[" << *memberName << L"]->" << *top() << L"(" << top() << L")"; 
	removeObject(memberName);
	removeObject(scriptObject);
	if (forceDelete) removeObject(so);		// Remove the temporary object wrapper.
}

/**
 * Set a named property to an object.
 **/
void Script::doActionSetMember(){
	if (debug) wcout << "SetMember ";
	ValueObject  *v = pop();
	ValueObject *memberName = (ValueObject*)pop();
	StringObject *scriptObject = (StringObject*)pop();
	wstring skey = scriptObject->toString();

	// wcout << *scriptObject << L"[" << *memberName << L"]="; fflush(stdout);
	// wcout << scriptObject->getObjectType() << L"\n"; fflush(stdout);

	ScriptObject *so = NULL;
	// We have a true scriptObject, not an alias
	if (scriptObject->getObjectType() == ScriptObjectType) so = (ScriptObject*) scriptObject;

	// We have an alias, and we will try to get the appropriate ScriptObject
	if (so == NULL) {
		if (skey == L"_global") so = (ScriptObject*) objectStore->getGlobalVariable(skey);
		else {
			ScriptObject *soglobal = (ScriptObject*) objectStore->getGlobalVariable(L"_global");
			so = (ScriptObject*) soglobal->getMember(skey);
		}
	}
	if (so == NULL) {
		wcout << *scriptObject << *memberName << L"]=" << *v;
		dumpStack();
		CRITICAL_ERROR(L" SANITY CHECK doActionSetMember");
	} else {
		so->setMember(memberName->toString(),v);
		// Check if the member is a named object.
		if (skey == L"_global") {
			ScriptObject *s = objectStore->getMovieByNameOrPath(memberName->toString());
			if (s != NULL) {
				s->setMainMember(v);
			}
		}

		if (debug) wcout << *scriptObject << L"("<<so<<L")" << L"[" << *memberName << L"]=" << *v;
		removeObject(v);
		removeObject(memberName);
		removeObject(scriptObject);
	}
}

/**
 *  Implements the ActionScript extends keyword. ActionExtends creates an inheritance relationship between two classes, 
 * called the subclass and the superclass.
 **/
void Script::doActionExtends() {
	StringObject *superClass = (StringObject *) pop();
	StringObject *subClass = (StringObject *) pop();
	if (debug) wcout << "Class Extends " << superClass->toString() << L" " << subClass->toString() << L"\n";

	ScriptObject *superInstance= (ScriptObject *)objectStore->getGlobalVariable(superClass->toString());
	if (superInstance == NULL) superInstance = asLinker->getClassInstance(superClass->toString());
	if (superInstance == NULL) CRITICAL_ERROR(L"Super class not found!");
	ScriptObject *superClassProto = (ScriptObject *)superInstance->getMember(superClass->toString());
	ScriptObject * sc = new  ASObject(this,graphics,subClass->toString());
	sc->setPrototype(superClassProto);
	sc->setMember(L"constructor",sc);
	getObjectStore()->registerNewObject(sc);
}

/**
 * Resolve a call to a built in function
 * This happens if the funtion is not already found in the Cache
 *
 **/ 
Function *Script::_callBuiltInFunction(ScriptObject **sc,wstring methodName) {
	//printf(" Func "); wcout << (*sc)->getThisObjectInstanceName() << L"\n";
	wstring className = (*sc)->getThisObjectInstanceName();
	// If we pass a basic class, we need to instanciate a high level class and invoke the class constructor to set the main property value.
	if (className == L"_$$$Number") {
		*sc = new ASNumber(this,graphics,(*sc)->toNumeric());
		className= L"Number";
	}		
	if (className == L"_$$$String") {
		*sc = new ASString(this,graphics,(*sc)->toString());
		className= L"String";
	}		
	if (className == L"_$$$Boolean") {
		*sc = new ASBoolean(this,graphics,int((*sc)->toNumeric()) != 0);
		className= L"Boolean";
	}
	if (debug) wcout<<L" Func "<<className<<L" ";
	Function *func = asLinker->getFunctionMember(className,methodName,sc);
	return func;
}

/**
 * Initializes an Object in a ScriptObject.
 **/
void Script::doActionInitObject() {
	if (debug) wcout << L"InitObject ";
	ScriptObject *so = new ScriptObject(L"Object");
	getObjectStore()->registerNewObject(so);
	ValueObject *nbElem = pop();
	int nbProps = int(nbElem->toNumeric());
	for (int i=0; i< nbProps; i++) {
		ValueObject *value = pop();
		ValueObject *key   = pop();
		so->setMember(key->toString(),value);
		if (debug) wcout<<*key<<L":="<<*value<<L" ";
		removeObject(key);
		removeObject(value);
	}
	execStack.push(so);
}

// SWF3 actions
/**
 * Instructs Flash Player to go to the specified frame in the current movie.
 **/
void Script::doActionGotoFrame(UI8 **p){
	int targetFrame = GETUI16(*p);
	(*p) += 2;
	if (debug) wcout << L"GotoFrame " << targetFrame;
	wstring globalMovieName = ((ASObject*)currentContext)->getMovieName();
	if ( globalMovieName == L"_global" || ((ASObject*)currentContext)->getClassName() != L"Sprite"){
		engine->goToMainTrackFrame(targetFrame);
	} else {
		if (debug) wcout << L" current context: " << *currentContext << L" " << ((ASObject*)currentContext)->getMovieName() << L"\n";
		ASInterface *amc = (ASInterface *) currentContext;
		amc->actionGotoFrame(targetFrame);
	}
}

/**
 * Instructs Flash Player to get the URL specified by UrlString. The URL can be of any type, including an HTML file, an image or another SWF movie. 
 * If the movie is playing in a browser, the URL will be displayed in the frame specified by TargetString. The special target names
 * “_level0” and “_level1” are used to load another SWF movie into levels 0 and 1 respectively.
 **/
void Script::doActionGetURL(UI8 **p){
	wstring urlString = utf8Towstring((const char*)(*p));
	*p += 1+strlen((const char*)(*p));
	wstring targetString = utf8Towstring((const char*)(*p));
	*p += 1+strlen((const char*)(*p));
	if (debug) wcout << "ActionGetURL " << urlString << L" " << targetString << L"\n";
	puts("FINISH IT");
}

/**
 * Go on the next frame for the current movie.
 **/
void Script::doActionNextFrame(){
	if (debug) wcout << L"ActionNextFrame";
	wstring globalMovieName = ((ASObject*)currentContext)->getMovieName();
	if ( globalMovieName == L"_global" || ((ASObject*)currentContext)->getClassName() != L"Sprite"){
		engine->next();
	} else {
		ASInterface *amc = (ASInterface *) currentContext;
		amc->actionNextFrame();
	}
};

/**
 * Go on the previous frame for the current movie.
 **/
void Script::doActionPrevFrame(){
	if (debug) wcout << L"ActionPrevFrame";
	wstring globalMovieName = ((ASObject*)currentContext)->getMovieName();
	if ( globalMovieName == L"_global" || ((ASObject*)currentContext)->getClassName() != L"Sprite"){
		engine->actionPreviousFrame();
	} else {
		ASInterface *amc = (ASInterface *) currentContext;
		amc->actionPreviousFrame();
	}
};

/**
 * Instructs Flash Player to start playing at the current frame.
 **/
void Script::doActionPlay(){
	ASInterface *amc = (ASInterface *) currentContext;
	if (debug) wcout << L"doActionPlay on " << amc->getMovieName() << L" (" << amc << L") ";
	if (engine->isSpriteObject(amc)) {
		amc->setIsStopped(false);
	} 
};

/**
 * Instructs Flash Player to stop playing the movie at the current frame.
 **/
void Script::doActionStop(){
	ASInterface *amc = (ASInterface *) currentContext;
	if (debug) wcout << L"doActionStop on " << amc->getMovieName();
	amc->setIsStopped(true);
};

/**
 * Toggle quality between high and low
 **/
void Script::doActionToggleQuality(){
	if (debug) wcout << L"ActionToggleQuality";
	// Not implemented yet
};

/**
 * Instructs Flash Player to stop playing all sounds.
 **/
void Script::doActionStopSounds(){
	wcout << L"ActionStopSounds";
	ASSound *as = (ASSound*)objectStore->getMovieByNameOrPath(L"_SoundManager");
	if (as != NULL) {
		as->stop();
	}
};

/**
 *  Instructs Flash Player to wait until the specified frame; otherwise skips the specified number of actions.
 **/
void Script::doActionWaitForFrame(UI8 **p){
	wcout << L"ActionWaitForFrame ";
	int frame = GETUI16(*p);
	(*p) += 2;
	int skipCount = (*(unsigned char*)p);
	(*p) ++;
	engine->setActionSkip(skipCount);
	engine->setWaitFrame(frame);
};

/**
 * Instructs Flash Player to change the context of subsequent actions, so they apply to a named object (TargetName) rather than the current movie.
 **/
void Script::doActionSetTarget(UI8 **p){
	wstring targetName = utf8Towstring((const char*)(*p));
	*p += 1+strlen((const char*)(*p));

	// Merge paths.
	wstring a = getObjectStore()->getCurrentTargetPath();
	wstring b = targetName;
	wstring::size_type pos = b.find(L"../");
	while (pos != wstring::npos) {
		wstring::size_type pos2 = a.find_last_of('/');
		if (pos2 != wstring::npos) a = a.substr(0,pos2);
		b = b.substr(3);
		pos = b.find(L"..");
	}
	if (b.find('/') == 0) b = b.substr(1);
	wstring finalPath = a+L"/"+b;
	if (targetName == L"") finalPath=L"/";

	if (targetName != L"") {
		ASInterface *amc = (ASInterface *) getObjectStore()->getVariable(finalPath);
		if (amc != NULL) {
			actionContextPush(amc,amc->getMovieName());
			hasTargetDefined = true;
		}
		wcout << amc << L" ";
	} else {
		actionContextPop();
		hasTargetDefined = false;
	}
	if (debug) wcout << "ActionSetTarget ->" << finalPath ;
};

/**
 * Instructs Flash Player to go to frame associated with the specified label. A label can be attached to a frame 
 * with the FrameLabel tag.
 **/
void Script::doActionGoToLabel(UI8 **p){
	wstring labelName = utf8Towstring((const char*)(*p));
	*p += 1+strlen((const char*)(*p));
	if (debug) wcout << L"doActionGoToLabel " << labelName << L"\n";
	int frame = getLabel(labelName);
	if (debug) wcout << L"doActionGoToLabel " << labelName << L" -> frame " << frame << L"\n";

	// All sounds must be stopped completely before the jump
	ASSound *as = (ASSound*)objectStore->getGlobalVariable(L"_SoundManager");
	if (as != NULL)
		as->stop();
	engine->goToMainTrackFrame(frame);
};

/**
 * Calls a subroutine.
 **/
void Script::doActionCall(){
	ValueObject *frameOrLabel = pop();
	if (frameOrLabel->getObjectType() == StringObjectType) {
		wstring label = frameOrLabel->toString();
		int frame = getLabel(label);
		// TODO: get actions on this frame and execute it!
		frame = 0; // Clean compile
	} else {
		int frame = int(frameOrLabel->toNumeric());
		// TODO: get actions on this frame and execute it!
		frame = 0; // Clean compile
	}
	removeObject(frameOrLabel);
	puts("FINISH ActionCall");
};

/**
 * Get URL Stack based.
 **/
void Script::doActionGetURL2(UI8 **p){
	int props = *(*p);
	(*p) ++;
	ValueObject  *target = pop();
	ValueObject  *url = pop();

	props =0 ; //clean compile
	removeObject(target);
	removeObject(url);
	puts("TODO: GetURL2");
};

/**
 * Goes to Frame stack based.
 * 
 * TODO: Implement composed targets
 **/
void Script::doActionGotoFrame2(UI8 **p){
	int props = **p;
	(*p) ++;
	int sceneBias = 0;
	if ((props & 2) != 0) {
		sceneBias = GETUI16(*p);
		(*p) += 2;
	}

 	// All sounds must be stopped completely before the jump
 	ASSound *as = (ASSound*)objectStore->getGlobalVariable(L"_SoundManager");
 	as->stop();
	//ASSound::stop adds a true boolean to the stack. This is not needed here
	pop();

	ValueObject *frameOrLabel = pop();
	if (frameOrLabel->getObjectType() == StringObjectType) {
		wstring label = frameOrLabel->toString();
		if (label[0] >= '0' && label[0]<='9') {
			int frame = getLabel(label);
			engine->goToMainTrackFrame(frame+1+sceneBias);
		} else {
			// TODO: goto Label
		}
	} else {
		int frame = int(frameOrLabel->toNumeric());
		engine->goToMainTrackFrame(frame+1+sceneBias);
	}
	removeObject(frameOrLabel);
	if (debug) wcout << L"ActionGotoFrame2: TODO: /MovieClip:Label format";
};

/**
 * Set current context, stack based.
 **/
void Script::doActionSetTarget2(){
	ValueObject *target = pop();
	removeObject(target);
	puts("TODO: ActionSetTarget2");
};

/**
 * Clone a Sprite
 *
 **/
void Script::doActionCloneSprite(){
	ValueObject *vtarget = pop();
	ValueObject *vdepth = pop();
	ValueObject *vsource = pop();

	if (debug) wcout << L"ActionCloneSprite "<< *vtarget << L" " << *vdepth<< L" " << *vsource;
	int dest  = int(vtarget->toNumeric());
	ASObject *amc = NULL;
	if (vsource->getObjectType() == ScriptObjectType) {
		amc = (ASObject *) vsource;
	} else {
		amc = (ASObject *)objectStore->getGlobalVariable(vsource->toString());
	}
	if (amc != NULL) {
		ASObject *dobj = (ASObject *) amc->getObjectDeepCopy(true);
		dobj->setMovieName(vdepth->toString());
		dobj->setVisible(true);
		dobj->setASCreated(true);
		dobj->setDepth(dest);
		getObjectStore()->setObjectAtDepth(dest,dobj);
		objectStore->setGlobalVariable(vdepth->toString(),dobj);
	} else {
		wcout << L"Script::doActionCloneSprite FAILS!!!!!!!!\n";
	}
};

/**
 * Remove a cloned sprite
 **/
void Script::doActionRemoveSprite(){
	ValueObject *v = pop();
	ASInterface *script = NULL;
	if (v->getObjectType() != ScriptObjectType) {
		script = (ASInterface *)objectStore->getGlobalVariable(v->toString());
	} else {
		script = (ASInterface *) v;
	}
	if (script != NULL) {
		objectStore->removeObjectAtDepth(script->getDepth());
		objectStore->destroyObject(script);
	}
	if (debug) wcout << L"RemoveSprite ";
}

/**
 * Start dragging on a movie clip.
 **/
void Script::doActionStartDrag(){
	ValueObject *target = pop();
	ValueObject *lockCenter = pop();
	ValueObject *contraints = pop();
	if (contraints->toNumeric() == 0) {
		ValueObject *y2 = pop();
		ValueObject *x2 = pop();
		ValueObject *y1 = pop();
		ValueObject *x1 = pop();

		removeObject(x1);
		removeObject(x2);
		removeObject(y1);
		removeObject(y2);
	}	
	removeObject(target);
	removeObject(lockCenter);
	removeObject(contraints);
	puts("TODO: FINISH ActionStartDrag");
};

/**
 * End th edrag operation in progress in any.
 **/
void Script::doActionEndDrag(){
	puts("TODO: FINISH ActionEndDrag");
};

/**
 * Wait for a frame to be loaded, stack based.
 **/
void Script::doActionWaitForFrame2(UI8 **p){
	int skipCount = (*(unsigned char*)p);
	(*p) ++;
	ValueObject *frame = pop();
	skipCount = 0; // clean compile

	removeObject(frame);
	puts("TODO: implement ActionWaitForFrame2");
};

/**
 * Sends a debugging output string.
 **/
void Script::doActionTrace(){
	ValueObject *value = pop();
	if (debug) wcout << L"ActionTrace ->" << value->toString();
	removeObject(value);
};

/**
 * Deletes a named property from a ScriptObject.
 **/
void Script::doActionDelete(){
	ValueObject *property = pop();
	ValueObject *object = pop();

	removeObject(property);
	removeObject(object);
	puts("TODO: FINISH doActionDelete");
};

/**
 * Deletes the variables of a thread or Flash Player.
 **/
void Script::doActionDelete2(){
	ValueObject *name = pop();
	removeObject(name);
	puts("TODO: FINISH ActionDelete2");
};

/**
 * Obtains the names of all “slots” in use in an ActionScript object—that is, for an object obj, all
 * names X that could be retrieved with the syntax obj.X. It is used to implement the ActionScriptfor / in loop.
 **/
void Script::doActionEnumerate(){
	if (debug) wcout << L"ActionEnumerate";
	ValueObject *name = pop();
	ScriptObject *p1 = getObjectStore()->getMovieByNameOrPath(name->toString());
	if (p1 != NULL && p1->getObjectType() == ScriptObjectType) {
		ScriptObject *sc = (ScriptObject*)p1; 
		ObjectMembers *objectMembers = sc->getMembers();
		push(new StringObject(L"null"));
		StringValueObjectHashMap *entries = objectMembers->getMembers();
		for (StringValueObjectHashMap::iterator it = entries->begin(); it != entries->end(); it++) {
			wstring s = (*it).first;
			push(new StringObject(s));
		}
	} else {
		push(new StringObject(L"null"));
	}
	removeObject(name);
};

/**
 * Invokes a constructor function to create a new object.
 **/
void Script::doActionNewMethod(UI8 **p){
	if (debug) wcout  << L"ActionNewMethod ";
	ValueObject *methodName = pop();
	ScriptObject *sc = (ScriptObject *)pop();
	wcout << methodName->toString() << L" " << sc->getThisObjectInstanceName();
	ValueObject *numParams = (ValueObject *)pop();
	if (debug) wcout << L" args=" << numParams->toNumeric();
	ValueObject** s = new ValueObject*[int(numParams->toNumeric())];
	for (int i=0; i< int(numParams->toNumeric()); i++) {
		s[i] = pop();
		if(debug) wcout<<*(s[i])<<L", ";
	}

	Function *func = objectStore->getFunction(L"_global",methodName->toString());
	ScriptObject *so;
	if (func == NULL) {
		wcout << "***** Constructor for class " << methodName->toString() << L" not found!\n";
		so = new ASObject(this,graphics,methodName->toString());
		execStack.push(so);
	} else {
		so = sc; 
		// If object exist, don't create it again
		//new ASObject(this,graphics,methodName->toString());
		//so->setMember(L"prototype",so);
		//so->setMember(L"constructor",so);
		execStack.push(so);
		_doCall(func,int(numParams->toNumeric()),s,so);
		implicitReturnsTarget.push(func->getCode() + func->getCodeSize());
		implicitReturns.push(*p);
		*p = func->getCode();
	}

	// Register this new object
	ScriptObject *soglobal = (ScriptObject*) objectStore->getGlobalVariable(L"_global");
	soglobal->setMember(objectStore->buildKey(so),so);

	removeObject(methodName);
	for (int i=0; i< int(numParams->toNumeric()); i++) removeObject(s[i]);
	delete []s;
	removeObject(sc);
};

/**
 * If the object in the stack is a “movieclip”, then the object’s target path is pushed on the stack in
 * dot notation. If the object is not a movie clip, the result is the "undefined" type rather than the
 * movie clip target path.
 **/
void Script::doActionTargetPath(){
	ValueObject *value = pop();
	if (value->getObjectType() == ScriptObjectType) {
		ASInterface *so = (ASInterface*) value;
		if (so->getClassName() == L"MovieClip") {
			execStack.push(new StringObject(L"TODO: MOVIECLIP PATH"));
		} else execStack.push(new StringObject(L"undefined"));
	} else execStack.push(new StringObject(L"undefined"));
	
	removeObject(value);
};

/**
 * Defines a “With” block of script.
 **/
void Script::doActionWith(UI8 **p){
	int tmpSize = GETUI16(*p);
	(*p) += 2;
	UI8* code = new UI8[tmpSize+1];
	UI8* pcode = code;
	for (int i=0; i< tmpSize; i++) code[i]=(*p)[i];
	code[tmpSize] = 0;
	(*p) += tmpSize;

	ScriptObject *sc = (ScriptObject*)pop();
	if (debug) wcout << L"Excute ActionWith for " << sc << L" " << sc->getMovieName() << L"\n\n";
	wstring currContext = objectStore->getPoolActionContext();
	prepareExecEnv(sc);
	objectStore->setPoolActionContext(currContext); // Always correct ?

	while (*pcode != 0) {
		engine->actionDispatch(&pcode);
	}
	leaveExecEnv();
	delete []code;
	if (debug) wcout << L"end action with\n";
};

/**
 * Converts the object on the top of the stack into a number, and pushes the number back to the stack.
 **/
void Script::doActionToNumber(){
	if (debug) wcout << L"ToNumber";
	ValueObject *value = pop();
	execStack.push(new DoubleObject(value->toNumeric()));
	removeObject(value);
};

/**
 * Converts the object on the top of the stack into a string, and pushes the string back to the stack.
 **/
void Script::doActionToString(){
	if (debug) wcout << L"ToString";
	ValueObject *p1 = pop();
	execStack.push(new StringObject(p1->toString()));
	removeObject(p1); 
};

/**
 * Returns type of Object
 **/
void Script::doActionTypeOf() {
	ValueObject *p1 = pop();
	wstring res = L"undefined";
	wstring type = p1->getThisObjectInstanceName();
	if (type == L"_$$$Number"  || type == L"Number")  res = L"number";
	if (type == L"_$$$Boolean" || type == L"Boolean") res = L"boolean";
	if (type == L"_$$$String"  || type == L"String")  res = L"string";
	if (type == L"Object") res = L"object";
	if (type == L"Sprite" || type == L"MovieClip") res = L"movieclip";
	if (type == L"Function" || type == L"function" || type == L"FunctionLiteral") res = L"function";
	if (p1->getObjectType() == ValueObjectType) res = L"null";
	if (p1->getObjectType() == ScriptObjectType && res == L"undefined") res = L"object";
	if (debug) wcout << L"ActionTypeOf ->" << res;
	push(new StringObject(res));
}

/**
 * Swaps the top two ScriptAtoms on the stack.
 **/
void Script::doActionStackSwap(){
	if (debug) wcout << L"ActionStackSwap";
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	execStack.push(p1);
	execStack.push(p2);
};

/**
 * 
 **/
void Script::doDoInitAction(){
	puts("TODO:");
};

/**
 *
 **/
void Script::doActionInstanceOf(){
	puts("TODO:");
};

/**
 * Obtains the names of all “slots” in use in an ActionScript object—that is, for an object obj, all
 * names X that could be retrieved with the syntax obj.X. It is used to implement the ActionScript
 * for / in loop.
 **/
void Script::doActionEnumerate2(){
	ValueObject *p1 = pop();
	if (debug) wcout << L"ActionEnumerate2 on "<< *p1;
	push(new StringObject(L"null"));
	if (p1->getObjectType() == ScriptObjectType) {
		ScriptObject *sc = (ScriptObject*)p1; 
		ObjectMembers *objectMembers = sc->getMembers();
		StringValueObjectHashMap *entries = objectMembers->getMembers();
		for (StringValueObjectHashMap::iterator it = entries->begin(); it != entries->end(); it++) {
			wstring s = (*it).first;
			if (s != L"_ROOT" && s != L"_root" ) {
				if (debug) wcout << s << L"   " ;
				push(new StringObject(s));
			}
		}
	}
	removeObject(p1);
/**
	According to MX Flash, it look it returns nothing:
	result = new Number();
	for (props in result) {
		trace (props);
	}

	 else if (p1->getThisObjectInstanceName() == L"_$$$Number") {
		push(new FloatObject(p1->toNumeric()));
	} else if (p1->getThisObjectInstanceName() == L"_$$$String") {
		push(new StringObject(p1->toString()));
	}
*/
};

/**
 * Similar to ActionEquals2, but the two arguments must be of the same type in order to be considered equal. 
 * Implements the ‘===’ operator from the ActionScript language.
 **/
void Script::doActionStrictEquals(){
	ValueObject *p1 = pop();
	ValueObject *p2 = pop();
	bool res = false;
	if (debug) wcout << "StrictEquals "<< *p1 << L" === " << *p2;
	if (p1->getObjectType() == StringObjectType ||  p2->getObjectType() == StringObjectType) {
		wstring s1 = p1->toString();
		wstring s2 = p2->toString();
		if (s1 == L"null") s1 = L"undefined";
		if (s2 == L"null") s2 = L"undefined";
		res = s1 == s2;
	} else {
		res = p1->toNumeric() == p2->toNumeric();
	}
	if (p1->getObjectType() != p2->getObjectType()) res = false;
	execStack.push(new BooleanObject(res));

	removeObject(p1);
	removeObject(p2);
};

/**
 *
 **/
void Script::doActionStringGreater(){
	puts("TODO:");
};

/**
 *
 **/
void Script::doActionCastOp(){
	puts("TODO:");
};

/**
 *
 **/
void Script::doActionImplementsOp(){
	puts("TODO:");
};

/**
 *
 **/
void Script::doActionTry(){
	puts("TODO:");
};

/**
 *
 **/
void Script::doActionThrow(){
	puts("TODO:");
};


/**
 * DEBUG - Mortem analysis
 **/
void Script::dumpStack() {
	wcout << L"\n******** DUMP Nb elements: " << execStack.size() << L"\n";
	if (execStack.size() > 0) {
		wcout<<L"Top of Stack:"<<*(execStack.top()) << L"\n";

		execStack.pop();
		while (execStack.size() > 0) {
			wcout<<*(execStack.top()) << L"\n";
			execStack.pop();
		}
	}
	wcout << L"\n"; fflush(stdout);
}

EventManager*Script::getEventManager() { 
	return engine->getEventManager();
}

void Script::removeObject(ValueObject *v) { 
	objectStore->removeObject(v);
}

ValueObject * Script::pop() { 
	ValueObject *p = NULL;
	if (execStack.size() > 0) {
		p = execStack.top(); 
		execStack.pop();
	} else {
		p = new ValueObject();
	}
	return p; 
}

/**
 * Get a label for a GotoLabel/2 instruction
 * If the label is not found, reparse the flash file to find it (the label may be ahead)
 **/
int Script::getLabel(wstring labelName) { 
	if (labels.find(labelName) == labels.end()) {
		engine->scanLabels();
	}
	return labels[labelName]; 
}

void  Script::setRegisterBank(RegisterSet *s) { 
	//printf("Change register set to %x\n",s);
	currentRegisterSet = s; 
}

/**
 * returns a singleton instance.
 **/
ScriptObject* Script::getSingleton(wstring className) {
	return asLinker->getClassSingleton(className);
}

