/**
Macromedia Flash File decoder TAGS definition.
Copyright (C) 2004  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#ifndef _SWF_TAGS_
#define _SWF_TAGS_
#include "common.hh"
#include "swf.hh"

static pthread_mutex_t  audioSynchro = PTHREAD_MUTEX_INITIALIZER;	// Protect ~ASSound while SDL callback in progress.

/**
 * The base Tag class
 **/
class BaseTag {
public:
	void setType(int type) {tagType=type; };
	int  getType() { return tagType; };
	virtual ~BaseTag() { }
	UI16	getCurSpriteId() { return curSpriteId; }
	void	setCurSpriteId(UI16 p) { curSpriteId =p; }	
private:
	int 	tagType;
	UI16	curSpriteId;
};

/**
 * Tag class implementations
 **/
class  SetBackgroundColorTag : public BaseTag {
public:
	SetBackgroundColorTag() { setType(stagSetBackgroundColor); }
	void setBackgroundColor(UI8 red, UI8 green, UI8 blue) { backgroundColor.red = red; backgroundColor.green = green; backgroundColor.blue = blue; }
	UI8	 getRed()   { return backgroundColor.red; }
	UI8	 getGreen() { return backgroundColor.green; }
	UI8	 getBlue()  { return backgroundColor.blue; }
	RGB	 getColor() { return backgroundColor; }
private:
	RGB backgroundColor;
};

/**
 * A DoAction Tag 
 **/
class DoActiontag : public BaseTag {
public:
	DoActiontag() { 
		setType(stagDoAction); 
		code = NULL;
	}
	~DoActiontag() {
		if (code != NULL) delete []code;
	}
	void setCode(UI8 *p) { code = p; }
	UI8  *getCode() { return code; }
private:
	UI8* code;
};

/**
 * A shape tag 
 * 
 * shapeWidthStyle may be a mixed list of StyleChangeRecord, StraightEdgeRecord,CurvedEdgeRecord,EndShapeRecord
 **/
class DefineShapeTag : public BaseTag {
public:
	DefineShapeTag() { setType(stagDefineShape); };
	virtual ~DefineShapeTag() { 
		delete shapeWithStyle;
	};
	UI16	getShapeId() { return shapeId; };
	RECT	*getShapeBounds() { return &shapeBounds; };
	ShapeWithStyle *getShapeWithStyle() { return shapeWithStyle; };
	void	setShapeId(UI16 p) { shapeId = p; };
	void	setShapeBounds(RECT p) { shapeBounds = p; };
	void	setShapeWithStyle(ShapeWithStyle *p) { shapeWithStyle = p; };

private:
	UI16	shapeId;
	RECT	shapeBounds;
	ShapeWithStyle *shapeWithStyle;
};

/**
 * Place object tag
 **/
class PlaceObjectTag : public BaseTag {
public:
	PlaceObjectTag() { setType(stagPlaceObject); };
	~PlaceObjectTag() { };
	UI16	getDepth() { return depth; };
	UI16	getCharacterId() { return characterId; };
	MATRIX	*getMatrix() { return &matrix; };
	CXFORM  *getColorTransform() { return &colorTransform; };
	void	setDepth(UI16 p) { depth= p; };
	void	setCharacterId(UI16 p) { characterId= p; };
	void	setMatrix(MATRIX p) { matrix= p; };
	void	setColorTransform(CXFORM p) { colorTransform= p; };
private:
	UI16	depth;
	UI16	characterId;
	MATRIX	matrix;
	CXFORM  colorTransform;
};


/**
 * Place object II Tag
 **/
class PlaceObject2Tag : public BaseTag {
public:
	PlaceObject2Tag() { 	
		setType(stagPlaceObject2); 
		name = NULL;
		clipActions = NULL;
	};
	~PlaceObject2Tag() {
		if (name != NULL) delete name;
		if (clipActions != NULL) delete clipActions;
	};
	bool	getPlaceFlagHasClipActions() { return placeFlagHasClipActions; };
	bool	getPlaceFlagHasClipDepth() { return placeFlagHasClipDepth; };
	bool	getPlaceFlagHasName() { return placeFlagHasName; };
	bool	getPlaceFlagHasRatio() { return placeFlagHasRatio; };
	bool	getPlaceFlagHasColorTransform() { return placeFlagHasColorTransform; };
	bool	getPlaceFlagHasMatrix() { return placeFlagHasMatrix; };
	bool	getPlaceFlagHasCharacter() { return placeFlagHasCharacter; };
	bool	getPlaceFlagMove() { return placeFlagMove; };
	UI16	getDepth() { return depth; };
	UI16	getCharacterId() { return characterId; };
	MATRIX	*getMatrix() { return &matrix; };
	CXFORMWITHALPHA *getColorTransform() { return &colorTransform; };
	UI16	getRatio() { return ratio; };
	STRING 	*getName() { return name; };
	UI16	getClipDepth() { return clipDepth; };
	ClipActions *getClipActions() { return clipActions; };
	void	setPlaceFlagHasClipActions(bool p) { placeFlagHasClipActions= p; };
	void	setPlaceFlagHasClipDepth(bool p) { placeFlagHasClipDepth= p; };
	void	setPlaceFlagHasName(bool p) { placeFlagHasName= p; };
	void	setPlaceFlagHasRatio(bool p) { placeFlagHasRatio= p; };
	void	setPlaceFlagHasColorTransform(bool p) { placeFlagHasColorTransform= p; };
	void	setPlaceFlagHasMatrix(bool p) { placeFlagHasMatrix= p; };
	void	setPlaceFlagHasCharacter(bool p) { placeFlagHasCharacter= p; };
	void	setPlaceFlagMove(bool p) { placeFlagMove= p; };
	void	setDepth(UI16 p) { depth= p; };
	void	setCharacterId(UI16 p) { characterId= p; };
	void	setMatrix(MATRIX p) { matrix= p; };
	void	setColorTransform(CXFORMWITHALPHA p) { colorTransform= p; };
	void	setRatio(UI16 p) { ratio= p; };
	void 	setName(STRING *p) { name= p; };
	void	setClipDepth(UI16 p) { clipDepth= p; };
	void	setClipActions(ClipActions *p) { clipActions= p; };
private:
	bool	placeFlagHasClipActions;
	bool	placeFlagHasClipDepth;
	bool	placeFlagHasName;
	bool	placeFlagHasRatio;
	bool	placeFlagHasColorTransform;
	bool	placeFlagHasMatrix;
	bool	placeFlagHasCharacter;
	bool	placeFlagMove;
	UI16	depth;
	UI16	characterId;
	MATRIX	matrix;
	CXFORMWITHALPHA colorTransform;
	UI16	ratio;
	STRING 	*name;
	UI16	clipDepth;
	ClipActions *clipActions;
};

/**
 * a set tab index
 **/
class SetTabIndexTag  : public BaseTag {
public:
	SetTabIndexTag() { setType(stagSetTabIndex); };
	UI16	getDepth() { return depth;};
	UI16	getTabIndex() { return tabIndex;};
	void	setDepth(UI16 p) { depth = p;};
	void	setTabIndex(UI16 p) { tabIndex = p;};
private:
	UI16	depth;
	UI16	tabIndex;
};

/**
 * A Show frame tag
 **/
class ShowFrameTag : public BaseTag {
public:
	ShowFrameTag() { setType(stagShowFrame); frameNumber = 0; }
	void setFrameNumber(int num){ frameNumber = num; }
	int	getFrameNumber(){ return frameNumber; }
private:
	int frameNumber;
};

/**
 * An end Tag
 **/
class EndTag : public BaseTag {
public:
	EndTag() { setType(stagEnd); }
};

/**
 * A define font tag
 **/
class DefineFontTag : public BaseTag {
public:
	DefineFontTag() { setType(stagDefineFont); offsetTable = NULL; glyphShapeTable=NULL; }
	virtual ~DefineFontTag() { 
		if (glyphShapeTable != NULL) {
			delete []glyphShapeTable;
		}		
		if (offsetTable != NULL) delete []offsetTable;
	}
	UI16	getFontID() { return fontID; };
	UI16* 	getOffsetTable() { return offsetTable; }
	Shape*	getGlyphShapeTable() { return glyphShapeTable; }
	void	setFontID(UI16 p) { fontID =p; };
	void 	setOffsetTable(UI16* p) { offsetTable =p; }
	void	setGlyphShapeTable(Shape* p) { glyphShapeTable =p; }
	int 	getNGlyphs() { return offsetTable[0]; }
private:
	UI16	fontID;
	UI16	*offsetTable;
	Shape	*glyphShapeTable;
};

/**
 * A define font info tag
 **/
class DefineFontInfoTag : public BaseTag {
public:
	DefineFontInfoTag() { setType(stagDefineFontInfo); fontName=NULL;  codeTable=NULL;}
	~DefineFontInfoTag() { 
		if (fontName != NULL) free(fontName);
		if (codeTable != NULL) delete []codeTable;
	}
	UI16	getFontID() { return fontID;};
	char*  	getFontName() { return fontName;};
	bool	getFontFlagsSmallText() { return fontFlagsSmallText;};
	bool	getFontFlagsShiftJIS() { return fontFlagsShiftJIS;};
	bool	getFontFlagsANSI() { return fontFlagsANSI;};
	bool	getFontFlagsItalic() { return fontFlagsItalic;};
	bool	getFontFlagsBold() { return fontFlagsBold;};
	bool	getFontFlagsWideCodes() { return fontFlagsWideCodes;};
	UI16	*getCodeTable() { return codeTable;};
	int 	getCodeTableLen() { return codeTableLen; };
	void	setFontID(UI16 p) { fontID = p;};
	void  	setFontName(char*  p) { fontName = p;};
	void	setFontFlagsSmallText(bool p) { fontFlagsSmallText = p;};
	void	setFontFlagsShiftJIS(bool p) { fontFlagsShiftJIS = p;};
	void	setFontFlagsANSI(bool p) { fontFlagsANSI = p;};
	void	setFontFlagsItalic(bool p) { fontFlagsItalic = p;};
	void	setFontFlagsBold(bool p) { fontFlagsBold = p;};
	void	setFontFlagsWideCodes(bool p) { fontFlagsWideCodes = p;};
	void	setCodeTable(UI16 *p) { codeTable = p;};
	void 	setCodeTableLen(int p) { codeTableLen = p; };
private:
	UI16	fontID;
	char*  	fontName;
	bool	fontFlagsSmallText;
	bool	fontFlagsShiftJIS;
	bool	fontFlagsANSI;
	bool	fontFlagsItalic;
	bool	fontFlagsBold;
	bool	fontFlagsWideCodes;
	int 	codeTableLen;
	UI16	*codeTable;
};

/**
 * extends DefineFontInfoTag by adding LanguageCode
 **/
class DefineFontInfo2Tag : public DefineFontInfoTag {
public:
	DefineFontInfo2Tag() { setType(stagDefineFontInfo2); }
	~DefineFontInfo2Tag() { }
	LANGCODE getLanguageCode()  { return languageCode; };
	void	setLanguageCode(LANGCODE p)  { languageCode =p; };
private:
	LANGCODE languageCode;
};

/**
 * A define text tag
 **/
class DefineTextTag : public BaseTag {
public:
	DefineTextTag() { setType(stagDefineText); };
	virtual ~DefineTextTag() { 
		list<TEXTRECORD*>::iterator it;
		for (it = textRecords.begin(); it != textRecords.end(); it++) 	delete (*it);
		textRecords.clear(); 
	};
	UI16	getCharacterID() { return characterID;};
	RECT	*getTextBounds() { return &textBounds;};
	MATRIX	*getTextMatrix() { return &textMatrix;};
	UI8		getGlyphBits() { return glyphBits;};
	UI8		getAdvanceBits() { return advanceBits;};
	list<TEXTRECORD *>*getTextRecords() { return &textRecords;};
	void	setCharacterID(UI16 p) { characterID =p;};
	void	setTextBounds(RECT p) { textBounds =p;};
	void	setTextMatrix(MATRIX p) { textMatrix =p;};
	void	setGlyphBits(UI8 p) { glyphBits =p;};
	void	setAdvanceBits(UI8 p) { advanceBits =p;};
	void	addTextRecord(TEXTRECORD *p) { textRecords.push_back(p);};
private:
	UI16	characterID;
	RECT	textBounds;
	MATRIX	textMatrix;
	UI8		glyphBits;
	UI8		advanceBits;
	list<TEXTRECORD*> textRecords;
};

/**
 * A define II text tag
 **/
class DefineText2Tag : public DefineTextTag {
public:
	DefineText2Tag() { setType(stagDefineText2); };
private:
};

/**
 * A define sprite tag
 **/
class DefineSpriteTag : public BaseTag {
public:
	DefineSpriteTag() { setType(stagDefineSprite) ;}
	~DefineSpriteTag() { 
		list<BaseTag*>::iterator it;
		for (it = tags.begin(); it != tags.end(); it++) 	delete (*it);
		tags.clear(); 
	};
	UI16	getSpriteID() { return spriteID;}
	UI16	getFrameCount() { return frameCount; }
	void	setSpriteID(UI16 p) { spriteID = p;}
	void	setFrameCount(UI16 p) { frameCount = p; }
	list<BaseTag*> *getTags() { return &tags; }
	void	addTag(BaseTag *p) { tags.push_back(p); }
private:
	UI16	spriteID;
	UI16	frameCount;
	list<BaseTag*> tags;
};

/**
 * a remove object II tag
 **/
class RemoveObject2Tag : public BaseTag {
public:
	RemoveObject2Tag() { setType(stagRemoveObject2); }
	~RemoveObject2Tag() {}
	UI16	getDepth() { return depth; };
	void	setDepth(UI16 p) { depth = p; };	
private:
	UI16	depth;
};

/**
 * a define bits lossless tag
 **/
class DefineBitsLosslessTag : public BaseTag {
public:
	DefineBitsLosslessTag() { setType(stagDefineBitsLossless); zlibBitmapData = NULL;};
	~DefineBitsLosslessTag() { if (zlibBitmapData != NULL) delete []zlibBitmapData; };
	UI16	getCharacterID() { return characterID; };
	UI8		getBitmapFormat() { return bitmapFormat; };
	UI16	getBitmapWidth() { return bitmapWidth; };
	UI16	getBitmapHeight() { return bitmapHeight; };
	UI8		getBitmapColorTableSize() { return bitmapColorTableSize; };
	UI8		*getZlibBitmapData() { return zlibBitmapData; };
	void	setCharacterID(UI16 p) { characterID =p; };
	void	setBitmapFormat(UI8	p) { bitmapFormat =p; };
	void	setBitmapWidth(UI16 p) { bitmapWidth =p; };
	void	setBitmapHeight(UI16 p) { bitmapHeight =p; };
	void	setBitmapColorTableSize(UI8 p) { bitmapColorTableSize =p; };
	void	setZlibBitmapData(UI8 *p) { zlibBitmapData =p; };
private:
	UI16	characterID;
	UI8		bitmapFormat;
	UI16	bitmapWidth;
	UI16	bitmapHeight;
	UI8		bitmapColorTableSize;
	UI8		*zlibBitmapData;
};

/**
 * a define font II tag
 **/
class DefineFont2Tag : public BaseTag {
public:
	DefineFont2Tag() { 
		setType(stagDefineFont2); 
		fontName = NULL; 
		offsetTable= NULL; 
		codeTable = NULL; 
		glyphShapeTable = NULL; 
		fontAdvanceTable = NULL; 
		fontBoundsTable = NULL;
		fontKerningTable = NULL;
	}
	~DefineFont2Tag() {
		if (fontName != NULL) delete [] fontName;
		if (offsetTable != NULL) delete []offsetTable;
		if (codeTable != NULL) delete [] codeTable;
		if (glyphShapeTable != NULL) delete []glyphShapeTable;
		if (fontAdvanceTable != NULL) delete []fontAdvanceTable;
		if (fontBoundsTable != NULL) delete []fontBoundsTable;
		if (fontKerningTable != NULL) delete []fontKerningTable;
	}
	UI16	getFontID() { return fontID;};
	bool	getFontFlagsHasLayout() { return fontFlagsHasLayout;};
	bool	getFontFlagsShiftJIS() { return fontFlagsShiftJIS;};
	bool	getFontFlagsSmallText() { return fontFlagsSmallText;};
	bool	getFontFlagsANSI() { return fontFlagsANSI;};
	bool	getFontFlagsWideOffsets() { return fontFlagsWideOffsets;};
	bool	getFontFlagsWideCodes() { return fontFlagsWideCodes;};
	bool	getFontFlagsItalic() { return fontFlagsItalic;};
	bool	getFontFlagsBold() { return fontFlagsBold;};
	LANGCODE getLanguageCode() { return languageCode;};
	UI8		getFontNameLen() { return fontNameLen;};
	UI8	*	getFontName() { return fontName;};
	UI16	getNumGlyphs() { return numGlyphs;};
	UI32*	getOffsetTable() { return offsetTable;};
	UI32	getCodeTableOffset() { return codeTableOffset;};
	Shape*	getGlyphShapeTable() { return glyphShapeTable;};
	UI16*	getCodeTable() { return codeTable;};
	SI16	getFontAscent() { return fontAscent;};
	SI16	getFontDescent() { return fontDescent;};
	SI16	getFontLeading() { return fontLeading;};
	SI16* 	getFontAdvanceTable() { return fontAdvanceTable;};
	RECT*	getFontBoundsTable() { return fontBoundsTable;};
	UI16	getKerningCount() { return kerningCount;};
	KERNINGRECORD* getFontKerningTable() { return fontKerningTable;};
	void	setFontID(UI16 p) { fontID =p;};
	void	setFontFlagsHasLayout(bool p) { fontFlagsHasLayout =p;};
	void	setFontFlagsShiftJIS(bool p) { fontFlagsShiftJIS =p;};
	void	setFontFlagsSmallText(bool p) { fontFlagsSmallText =p;};
	void	setFontFlagsANSI(bool p) { fontFlagsANSI =p;};
	void	setFontFlagsWideOffsets(bool p) { fontFlagsWideOffsets =p;};
	void	setFontFlagsWideCodes(bool p) { fontFlagsWideCodes =p;};
	void	setFontFlagsItalic(bool p) { fontFlagsItalic =p;};
	void	setFontFlagsBold(bool p) { fontFlagsBold =p;};
	void	setLanguageCode(LANGCODE p) { languageCode =p;};
	void	setFontNameLen(UI8 p) { fontNameLen =p;};
	void	setFontName(UI8	* p) { fontName =p;};
	void	setNumGlyphs(UI16 p) { numGlyphs =p;};
	void	setOffsetTable(UI32* p) { offsetTable =p;};
	void	setCodeTableOffset(UI32 p) { codeTableOffset =p;};
	void	setGlyphShapeTable(Shape* p) { glyphShapeTable =p;};
	void	setCodeTable(UI16* p) { codeTable =p;};
	void	setFontAscent(SI16 p) { fontAscent =p;};
	void	setFontDescent(SI16 p) { fontDescent =p;};
	void	setFontLeading(SI16 p) { fontLeading =p;};
	void 	setFontAdvanceTable(SI16* p) { fontAdvanceTable =p;};
	void	setFontBoundsTable(RECT* p) { fontBoundsTable =p;};
	void	setKerningCount(UI16 p) { kerningCount =p;};
	void	setFontKerningTable(KERNINGRECORD* p) { fontKerningTable =p;};
private:
	UI16	fontID;
	bool	fontFlagsHasLayout;
	bool	fontFlagsShiftJIS;
	bool	fontFlagsSmallText;
	bool	fontFlagsANSI;
	bool	fontFlagsWideOffsets;
	bool	fontFlagsWideCodes;
	bool	fontFlagsItalic;
	bool	fontFlagsBold;
	LANGCODE languageCode;
	UI8		fontNameLen;
	UI8	*	fontName;
	UI16	numGlyphs;
	UI32*	offsetTable;
	UI32	codeTableOffset;
	Shape*	glyphShapeTable;
	UI16*	codeTable;
	SI16	fontAscent;
	SI16	fontDescent;
	SI16	fontLeading;
	SI16* 	fontAdvanceTable;
	RECT*	fontBoundsTable;
	UI16	kerningCount;
	KERNINGRECORD* fontKerningTable;
};

/**
 * a define edit text tag (input area)
 **/
class DefineEditTextTag : public BaseTag {
public:
	DefineEditTextTag() { setType(stagDefineEditText); variableName = (STRING*)NULL; initialText = (STRING*)NULL;}
	~DefineEditTextTag() { 
		if (variableName != NULL) delete variableName;
		if (initialText != NULL) delete initialText;
	};
	UI16	getCharacterID() { return characterID;};
	RECT	getBounds() { return bounds;};
	bool	getHasText() { return hasText;};
	bool	getWordWrap() { return wordWrap;};
	bool	getMultiline() { return multiline;};
	bool	getPassword() { return password;};
	bool	getReadOnly() { return readOnly;};
	bool	getHasTextColor() { return hasTextColor;};
	bool	getHasMaxLength() { return hasMaxLength;};
	bool	getHasFont() { return hasFont;};
	bool	getAutoSize() { return autoSize;};
	bool	getHasLayout() { return hasLayout;};
	bool	getNoSelect() { return noSelect;};
	bool	getBorder() { return border;};
	bool	getHTML() { return HTML;};
	bool	getUseOutlines() { return useOutlines;};
	UI16	getFontID() { return fontID;};
	UI16	getFontHeight() { return fontHeight;};
	RGBA	*getTextColor() { return &textColor;};
	UI16	getMaxLength() { return maxLength;};
	UI8		getAlign() { return align;};
	UI16	getLeftMargin() { return leftMargin;};
	UI16	getRightMargin() { return rightMargin;};
	UI16	getIndent() { return indent;};
	UI16	getLeading() { return leading;};
	STRING 	*getVariableName() { return variableName;};
	STRING	*getInitialText() { return initialText;};
	void	setCharacterID(UI16 p) { characterID =p;};
	void	setBounds(RECT p) { bounds =p;};
	void	setHasText(bool p) { hasText =p;};
	void	setWordWrap(bool p) { wordWrap =p;};
	void	setMultiline(bool p) { multiline =p;};
	void	setPassword(bool p) { password =p;};
	void	setReadOnly(bool p) { readOnly =p;};
	void	setHasTextColor(bool p) { hasTextColor =p;};
	void	setHasMaxLength(bool p) { hasMaxLength =p;};
	void	setHasFont(bool p) { hasFont =p;};
	void	setAutoSize(bool p) { autoSize =p;};
	void	setHasLayout(bool p) { hasLayout =p;};
	void	setNoSelect(bool p) { noSelect =p;};
	void	setBorder(bool p) { border =p;};
	void	setHTML(bool p) { HTML =p;};
	void	setUseOutlines(bool p) { useOutlines =p;};
	void	setFontID(UI16 p) { fontID =p;};
	void	setFontHeight(UI16 p) { fontHeight =p;};
	void	setTextColor(RGBA p) { textColor =p;};
	void	setMaxLength(UI16 p) { maxLength =p;};
	void	setAlign(UI8 p) { align =p;};
	void	setLeftMargin(UI16 p) { leftMargin =p;};
	void	setRightMargin(UI16 p) { rightMargin =p;};
	void	setIndent(UI16 p) { indent =p;};
	void	setLeading(UI16 p) { leading =p;};
	void	setVariableName(STRING 	* p) { variableName =p;};
	void	setInitialText(STRING	* p) { initialText =p;};
private:
	UI16	characterID;
	RECT	bounds;
	bool	hasText;
	bool	wordWrap;
	bool	multiline;
	bool	password;
	bool	readOnly;
	bool	hasTextColor;
	bool	hasMaxLength;
	bool	hasFont;
	bool	autoSize;
	bool	hasLayout;
	bool	noSelect;
	bool	border;
	bool	HTML;
	bool	useOutlines;
	UI16	fontID;
	UI16	fontHeight;
	RGBA	textColor;
	UI16	maxLength;
	UI8		align;
	UI16	leftMargin;
	UI16	rightMargin;
	UI16	indent;
	UI16	leading;
	STRING 	*variableName;
	STRING	*initialText;
};

/**
 * a button tag
 **/
class DefineButtonTag : public BaseTag {
public:
	DefineButtonTag() { setType(stagDefineButton); code = NULL;}
	~DefineButtonTag() { 
		if (code != NULL) delete []code;
	}
	UI16	getButtonId() { return buttonId; };
	list<BUTTONRECORD *>* getCharacters() { return &characters; };
	void	setButtonId(UI16 p) { buttonId =p; };
	void	addCharacters(BUTTONRECORD * p) { characters.push_back(p); };
	void 	setCode(UI8* p) { code = p; }
	UI8*	getCode() { return code; }		
private:
	UI16	buttonId;
	list<BUTTONRECORD *>characters;
	UI8*	code;
};


/**
 * a button II tag
 **/
class DefineButton2Tag : public BaseTag {
public:
	DefineButton2Tag() { 
		setType(stagDefineButton2); 
	}
	~DefineButton2Tag() { 
		for (list<BUTTONRECORD *>::iterator it= characters.begin(); it != characters.end(); ++it) delete *it;
		for (list<BUTTONCONDACTION *>::iterator it= actions.begin(); it != actions.end(); ++it) delete *it;
		characters.clear();
		actions.clear();
	}
	UI16	getButtonId() { return buttonId; };
	bool	getTrackAsMenu() { return trackAsMenu; };
	UI16	getActionOffset() { return actionOffset; };
	list<BUTTONRECORD *>* getCharacters() { return &characters; };
	list<BUTTONCONDACTION*>* getActions() { return &actions; };
	void	setButtonId(UI16 p) { buttonId =p; };
	void	setTrackAsMenu(bool p) { trackAsMenu =p; };
	void	setActionOffset(UI16 p ) {actionOffset = p; };
	void	addCharacters(BUTTONRECORD * p) { characters.push_back(p); };
	void	addAction(BUTTONCONDACTION* p) { actions.push_back(p); };
private:
	UI16	buttonId;
	bool	trackAsMenu;
	UI16	actionOffset;
	list<BUTTONRECORD *>characters;
	list<BUTTONCONDACTION*> actions;
};

/**
 * A frame label tag
 **/
class FrameLabelTag : public BaseTag {
public:
	FrameLabelTag() { setType(stagFrameLabel); name=NULL; }
	~FrameLabelTag() { if (name != NULL) delete name; }
	STRING 	*getName() { return name; };
	void	setName(STRING *p) { name = p; }
	UI8		getNamedAnchorFlag() { return namedAnchorFlag; }
	void	setNamedAnchorFlag(UI8 p) { namedAnchorFlag = p ; }
private:
	STRING 	*name;
	UI8		namedAnchorFlag;
};

/**
 * an export assets tag
 **/
class ExportAssetsTag : public BaseTag {
public:
	ExportAssetsTag() { setType(stagExportAssets); tags = NULL; names = NULL; }
	~ExportAssetsTag() { 
		if (tags != NULL) delete []tags; 
		if (names != NULL) {
			for (int i=0; i< count; i++) delete names[i];
			free(names);
		}
	}
	UI16	getCount() { return count;};
	UI16	*getTags() { return tags;};
	STRING  **getNames() { return names;};	
	void	setCount(UI16 p) { count =p;};
	void	setTags(UI16* p) { tags =p;};
	void	setNames(STRING** p) { names =p;};	
private:
	UI16	count;
	UI16	*tags;
	STRING  **names;	
};

/**
 * A do init action tag
 **/
class DoInitActionTag : public BaseTag {
public:
	DoInitActionTag() { 
		setType(stagDoInitAction); 
		code= NULL;
		processed = false;
	}
	~DoInitActionTag() { 
		if (code != NULL) delete []code;
	}
	UI16	getSpriteID() { return spriteID; }
	void	setSpriteID(UI16 p) { spriteID =p; }
	void 	setCode(UI8* p) { code = p; }
	UI8*	getCode() { return code; }
	void	setProcessed(bool p) { processed = p; }
	bool	getProcessed() { return processed; }
private:
	UI16	spriteID;
	UI8*	code;
	bool	processed;
};

/**
 * a define bits lossless tag
 **/
class DefineBitsLossless2Tag: public BaseTag {
public:
	DefineBitsLossless2Tag() { setType(stagDefineBitsLossless2); zlibBitmapData=NULL; };
	~DefineBitsLossless2Tag() { if (zlibBitmapData) delete []zlibBitmapData; }
	UI16	getCharacterID() { return characterID;};
	UI8		getBitmapFormat() { return bitmapFormat;};
	UI16	getBitmapWidth() { return bitmapWidth;};
	UI16	getBitmapHeight() { return bitmapHeight;};
	UI8		getBitmapColorTableSize() { return bitmapColorTableSize;};
	UI8		*getZlibBitmapData() { return zlibBitmapData;};
	UI32	getSize() { return size; }
	void	setCharacterID(UI16 p) { characterID =p;};
	void	setBitmapFormat(UI8 p) { bitmapFormat =p;};
	void	setBitmapWidth(UI16 p) { bitmapWidth =p;};
	void	setBitmapHeight(UI16 p) { bitmapHeight =p;};
	void	setBitmapColorTableSize(UI8 p) { bitmapColorTableSize =p;};
	void	setZlibBitmapData(UI8 *p) { zlibBitmapData =p;};
	void	setSize(UI32 p) { size =p; }
private:
	UI16	characterID;
	UI8		bitmapFormat;
	UI16	bitmapWidth;
	UI16	bitmapHeight;
	UI8		bitmapColorTableSize;
	UI32	size;
	UI8		*zlibBitmapData;
};

/**
 * a JPEG table tag
 **/
class JPEGTablesTag :  public BaseTag {
public:
	JPEGTablesTag() { setType(stagJPEGTables); JPEGData=NULL; }
	~JPEGTablesTag() { if (JPEGData != NULL) delete []JPEGData; }
	UI8*	getJPEGData() { return JPEGData;};
	void	setJPEGData(UI8* p) { JPEGData =p;};
private:
	UI8*	JPEGData;
};

/**
 * a define bits tag
 **/
class DefineBitsTag : public JPEGTablesTag {
public:
	DefineBitsTag() { setType(stagDefineBits);  }
	~DefineBitsTag() {  }
	UI16	getCharacterID() { return characterID;};
	void	setCharacterID(UI16 p) { characterID =p;};
private:
	UI16	characterID;
};

/**
 * a define bits JPEG II tag
 **/
class DefineBitsJPEG2Tag : public JPEGTablesTag {
public:
	DefineBitsJPEG2Tag() { setType(stagDefineBitsJPEG2);  }
	~DefineBitsJPEG2Tag() {  }
	UI16	getCharacterID() { return characterID;};
	void	setCharacterID(UI16 p) { characterID =p;};
private:
	UI16	characterID;
};

/**
 * a define bits JPEG III tag
 **/
class DefineBitsJPEG3Tag : public DefineBitsJPEG2Tag {
public:
	DefineBitsJPEG3Tag() { setType(stagDefineBitsJPEG3); }
	~DefineBitsJPEG3Tag() { }
	UI32	getAlphaDataOffset() { return alphaDataOffset;};
	void	setAlphaDataOffset(UI32 p) { alphaDataOffset =p;};
private:
	UI32	alphaDataOffset;
};

/**
 * a sound stream head definition tag
 **/
class SoundStreamHeadTag : public BaseTag {
public:
	SoundStreamHeadTag () { setType(stagSoundStreamHead); }
	~SoundStreamHeadTag () { }
	UI8		getPlaybackSoundRate() { return playbackSoundRate;};
	bool	getPlaybackSoundSize() { return playbackSoundSize;};
	bool	getPlaybackSoundType() { return playbackSoundType;};
	UI8		getStreamSoundCompression() { return streamSoundCompression;};
	UI8		getStreamSoundRate() { return streamSoundRate;};
	bool	getStreamSoundSize() { return streamSoundSize;};
	bool	getStreamSoundType() { return streamSoundType;};
	UI16	getStreamSoundSampleCount() { return streamSoundSampleCount;};
	SI16	getLatencySeek() { return latencySeek;};
	void	setPlaybackSoundRate(UI8 p) { playbackSoundRate =p;};
	void	setPlaybackSoundSize(bool p) { playbackSoundSize =p;};
	void	setPlaybackSoundType(bool p) { playbackSoundType =p;};
	void	setStreamSoundCompression(UI8 p) { streamSoundCompression =p;};
	void	setStreamSoundRate(UI8 p) { streamSoundRate =p;};
	void	setStreamSoundSize(bool p) { streamSoundSize =p;};
	void	setStreamSoundType(bool p) { streamSoundType =p;};
	void	setStreamSoundSampleCount(UI16 p) { streamSoundSampleCount =p;};
	void	setLatencySeek(SI16 p) { latencySeek =p;};
private:
	UI8		playbackSoundRate;
	bool	playbackSoundSize;
	bool	playbackSoundType;
	UI8		streamSoundCompression;
	UI8		streamSoundRate;
	bool	streamSoundSize;
	bool	streamSoundType;
	UI16	streamSoundSampleCount;
	SI16	latencySeek;
};

/**
 * a sound stream head definition II tag
 **/
class SoundStreamHead2Tag : public SoundStreamHeadTag {
public:
	SoundStreamHead2Tag () { setType(stagSoundStreamHead2); }
	~SoundStreamHead2Tag () { }
};

/**
 * a sound tag
 **/
class DefineSoundTag :  public BaseTag {
public:
	DefineSoundTag() { setType(stagDefineSound); soundData = NULL; }
	~DefineSoundTag() { 
		if (soundData !=NULL) delete []soundData; 
	}
	UI16	getSoundId() { return soundId;}
	UI8		getSoundFormat() { return soundFormat;};
	UI8		getSoundRate() { return soundRate;};
	bool	getSoundSize() { return soundSize;};
	bool	getSoundType() { return soundType;};
	UI32	getSoundSampleCount() { return soundSampleCount;};
	UI8*	getSoundData() { return soundData;};
	UI32	getSoundDataSize() { return dataSize; }
	void	setSoundId(UI16 p) { soundId =p;}
	void	setSoundFormat(UI8 p) { soundFormat =p;};
	void	setSoundRate(UI8 p) { soundRate =p;};
	void	setSoundSize(bool p) { soundSize =p;};
	void	setSoundType(bool p) { soundType =p;};
	void	setSoundSampleCount(UI32 p) { soundSampleCount =p;};
	void	setSoundData(UI8* p) { soundData =p;};
	void	setSoundDataSize(UI32 p) { dataSize = p; }
private:
	UI16	soundId;
	UI8		soundFormat;
	UI8		soundRate;
	bool	soundSize;
	bool	soundType;
	UI32	soundSampleCount;
	UI8*	soundData;
	UI32	dataSize;
};

/**
 * a define sound button tag
 **/ 
class DefineButtonSoundTag :  public BaseTag {
public:
	DefineButtonSoundTag() { 
		setType(stagDefineButtonSound); 
		buttonSoundInfo0 = NULL; 
		buttonSoundInfo1 = NULL; 
		buttonSoundInfo2 = NULL; 
		buttonSoundInfo3 = NULL; 
	}
	~DefineButtonSoundTag() {	
		if (buttonSoundInfo0 != NULL) delete buttonSoundInfo0;
		if (buttonSoundInfo1 != NULL) delete buttonSoundInfo1;
		if (buttonSoundInfo2 != NULL) delete buttonSoundInfo2;
		if (buttonSoundInfo3 != NULL) delete buttonSoundInfo3;
	}
	UI16	getButtonId() { return buttonId;};
	UI16	getButtonSoundChar0() { return buttonSoundChar0;};
	SOUNDINFO *getButtonSoundInfo0() { return buttonSoundInfo0;};
	UI16	getButtonSoundChar1() { return buttonSoundChar1;};
	SOUNDINFO *getButtonSoundInfo1() { return buttonSoundInfo1;};
	UI16	getButtonSoundChar2() { return buttonSoundChar2;};
	SOUNDINFO *getButtonSoundInfo2() { return buttonSoundInfo2;};
	UI16	getButtonSoundChar3() { return buttonSoundChar3;};
	SOUNDINFO *getButtonSoundInfo3() { return buttonSoundInfo3;};
	void	setButtonId(UI16 p) { buttonId =p;};
	void	setButtonSoundChar0(UI16 p) { buttonSoundChar0 =p;};
	void	setButtonSoundInfo0(SOUNDINFO *p) { buttonSoundInfo0 =p;};
	void	setButtonSoundChar1(UI16 p) { buttonSoundChar1 =p;};
	void	setButtonSoundInfo1(SOUNDINFO *p) { buttonSoundInfo1 =p;};
	void	setButtonSoundChar2(UI16 p) { buttonSoundChar2 =p;};
	void	setButtonSoundInfo2(SOUNDINFO *p) { buttonSoundInfo2 =p;};
	void	setButtonSoundChar3(UI16 p) { buttonSoundChar3 =p;};
	void	setButtonSoundInfo3(SOUNDINFO *p) { buttonSoundInfo3 =p;};
private:
	UI16	buttonId;
	UI16	buttonSoundChar0;
	SOUNDINFO *buttonSoundInfo0;
	UI16	buttonSoundChar1;
	SOUNDINFO *buttonSoundInfo1;
	UI16	buttonSoundChar2;
	SOUNDINFO *buttonSoundInfo2;
	UI16	buttonSoundChar3;
	SOUNDINFO *buttonSoundInfo3;
};

/**
 * a start sound tag
 **/
class StartSoundTag : public BaseTag {
public:
	StartSoundTag() { setType(stagStartSound); soundInfo = NULL;}
	~StartSoundTag() { if (soundInfo != NULL) delete soundInfo; }
	UI16	getSoundId() { return soundId;};
	SOUNDINFO *getSoundInfo() { return soundInfo;};
	void	setSoundId(UI16 p) { soundId =p;};
	void	setSoundInfo(SOUNDINFO *p) { soundInfo =p;};
private:
	UI16	soundId;
	SOUNDINFO *soundInfo;
};

/**
 * a define morph shape tag
 **/
class DefineMorphShapeTag : public BaseTag {
public:
	DefineMorphShapeTag() { setType(stagDefineMorphShape); morphFillStyles=NULL; morphLineStyles= NULL; startEdges=NULL; endEdges=NULL; }
	~DefineMorphShapeTag() { 
		if (morphFillStyles != NULL) delete morphFillStyles;
		if (morphLineStyles != NULL) delete morphLineStyles;
		if (startEdges != NULL) delete startEdges;
		if (endEdges != NULL) delete endEdges;
	}
	UI16	getCharacterID() { return characterID;};
	RECT	*getStartBounds() { return &startBounds;};
	RECT	*getEndBounds() { return &endBounds;};
	UI32	getOffset() { return offset;};
	MORPHFILLSTYLES *getMorphFillStyles() { return morphFillStyles;};
	MORPHLINESTYLES *getMorphLineStyles() { return morphLineStyles;};
	Shape	*getStartEdges() { return startEdges;};
	Shape	*getEndEdges() { return endEdges;};
	void	setCharacterID(UI16 p) { characterID =p;};
	void	setStartBounds(RECT p) { startBounds =p;};
	void	setEndBounds(RECT p) { endBounds =p;};
	void	setOffset(UI32 p) { offset =p;};
	void	setMorphFillStyles(MORPHFILLSTYLES *p) { morphFillStyles =p;};
	void	setMorphLineStyles(MORPHLINESTYLES *p) { morphLineStyles =p;};
	void	setStartEdges(Shape	*p) { startEdges =p;};
	void	setEndEdges(Shape *p) { endEdges =p;};
private:
	UI16	characterID;
	RECT	startBounds;
	RECT	endBounds;
	UI32	offset;
	MORPHFILLSTYLES *morphFillStyles;
	MORPHLINESTYLES *morphLineStyles;
	Shape	*startEdges;
	Shape	*endEdges;
};

/**
 * a sound stream block
 **/
class SoundStreamBlockTag : public BaseTag {
public:
	SoundStreamBlockTag()
	:streamSoundData(NULL),
	nextSoundBlock(NULL)
	{ setType(stagSoundStreamBlock); }
	~SoundStreamBlockTag() { if (streamSoundData != NULL) delete []streamSoundData;}
	UI8*	getStreamSoundData() { return streamSoundData; }
	void	setStreamSoundData(UI8* p) { streamSoundData =p; }
	void	setDataSize(int p) { dataSize = p; }
	int		getDataSize() { return dataSize; }
	void	setSampleSize(int p) { sampleSize = p; }
	int		getSampleSize() { return sampleSize; }
	void	setSkipSize(int p) { skipSize = p; }
	int		getSkipSize() { return skipSize; }
	// To get and set frame numbers for coordinating with showFrameTag
	int	getFrameNumber(){ return frameNumber; }
	void	setFrameNumber(int num) { frameNumber = num; }
	// Getting and setting the next soundstreamblock in sequence
	SoundStreamBlockTag* getNextBlock(){ return nextSoundBlock; }
	void	setNextBlock(SoundStreamBlockTag* next) { nextSoundBlock = next; }
	// The sampleCount
	void	setSampleCount(UI16 c){ sampleCount = c; }
	UI16	getSampleCount(){ return sampleCount; }
	void	setStreamId(int id){ streamSoundId = id; }
	int	getStreamId(){ return streamSoundId; }
	
private:
	UI8*	streamSoundData;
	int		dataSize;
	UI16	sampleCount;					// From the MP3 Stream Sound Data
	int 	sampleSize;			
	int		skipSize;					// From the MP3 Sound Data
	SoundStreamBlockTag*  nextSoundBlock;	// Links to the next tag if there is one. 
	int	frameNumber;			// Corresponds to the number in the showFrameTag
	int	streamSoundId;
};

/**
 * a file protected flag tag
 **/
class ProtectTag: public BaseTag {
public:
	ProtectTag() { setType(stagProtect); password = NULL;}
	~ProtectTag() { 
		if (password != NULL) delete password;
	 }
	STRING *getPassword() { return password; }
	void	setPassword(STRING *p) { password =p; }
private:
	STRING *password;
};

/**
 * A name tag
 **/
class NameCharacterTag : public BaseTag {
public:
	NameCharacterTag() { setType(stagNameCharacter); name = NULL; }
	~NameCharacterTag() { if (name != NULL) delete name; }
	STRING *getName() { return name; }
	UI16	getTagId() { return tagId; };
	void	setName(STRING *p) { name =p; }
	void 	setTagId(UI16 p) { tagId =p; };
private:
	UI16	tagId;
	STRING *name;
};

/**
 *  a remove object tag
 **/
class RemoveObjectTag : public BaseTag {
public:
	RemoveObjectTag() { setType(stagRemoveObject); }
	~RemoveObjectTag() { };
	UI16	getCharacterId() { return characterId;};
	UI16	getDepth() { return depth;};
	void	setCharacterId(UI16 p) { characterId =p;};
	void	setDepth(UI16 p) { depth =p;};
private:
	UI16	characterId;
	UI16	depth;
};

/**
 * a define button with cxform tag
 **/
class DefineButtonCxformTag : public BaseTag {
public:
	DefineButtonCxformTag() { setType(stagDefineButtonCxform); }
	~DefineButtonCxformTag() { }
	UI16	getButtonId() { return buttonId; }
	CXFORM	*getButtonColorTransform() { return &buttonColorTransform; }
	void	setButtonId(UI16 p) { buttonId =p; }
	void	setButtonColorTransform(CXFORM p) { buttonColorTransform =p; }
private:
	UI16	buttonId;
	CXFORM	buttonColorTransform;
};

/**
 * an import asset tag
 **/
class ImportAssetsTag : public BaseTag {
public:
	ImportAssetsTag() { setType(stagImportAssets);  URL=NULL; tags= NULL; names=NULL; }
	~ImportAssetsTag() { 
		if (URL != NULL) delete URL;
		if (tags != NULL) delete []tags;
		if (names != NULL) delete []names;
	}
	STRING *getURL() { return URL;};
	UI16	getCount() { return count;};
	UI16	*getTags() { return tags;};
	STRING  *getNames() { return names;};
	void	setURL(STRING * p) { URL =p;};
	void	setCount(UI16 p) { count =p;};
	void	setTags(UI16 *p) { tags =p;};
	void	setNames(STRING  *p) { names =p;};
private:
	STRING *URL;
	UI16	count;
	UI16	*tags;
	STRING  *names;
};

/**
 * a password enabled debug tag
 **/
class EnableDebuggerTag : public BaseTag {
public:
	EnableDebuggerTag() { setType(stagEnableDebugger);  password = NULL; }
	virtual ~EnableDebuggerTag() { if (password != NULL) delete password; }
	STRING 	*getPassword() { return password; };
	void	setPassword(STRING *p) { password =p; };
private:
	STRING 	*password;
};

/**
 * a password enabled debug II tag
 **/
class EnableDebugger2Tag : public EnableDebuggerTag {
public:
	EnableDebugger2Tag() { setType(stagEnableDebugger2); }
private:
};

/**
 * a define video stream tag
 **/
class DefineVideoStreamTag : public BaseTag {
public:
	DefineVideoStreamTag() { setType(stagDefineVideoStream); }
	~DefineVideoStreamTag() { };
	UI16	getCharacterID() { return characterID; };
	UI16	getNumFrames() { return numFrames; };
	UI16	getWidth() { return width; };
	UI16	getHeight() { return height; };
	UI8		getVideoFlagsDeblocking() { return videoFlagsDeblocking; };
	bool	getVideoFlagsSmoothing() { return videoFlagsSmoothing; };
	UI8		getCodecID() { return codecID; };
	void	setCharacterID(UI16 p) { characterID =p; };
	void	setNumFrames(UI16 p) { numFrames =p; };
	void	setWidth(UI16 p) { width =p; };
	void	setHeight(UI16 p) { height =p; };
	void	setVideoFlagsDeblocking(UI8 p) { videoFlagsDeblocking =p; };
	void	setVideoFlagsSmoothing(bool p) { videoFlagsSmoothing =p; };
	void	setCodecID(UI8 p) { codecID =p; };
private:
	UI16	characterID;
	UI16	numFrames;
	UI16	width;
	UI16	height;
	UI8		videoFlagsDeblocking;
	bool	videoFlagsSmoothing;
	UI8		codecID;
};

/**
 * a video frame definition tag
 **/
class VideoFrameTag : public BaseTag {
public:
	VideoFrameTag() { setType(stagVideoFrame); videoData = NULL; }
	~VideoFrameTag() { if (videoData != NULL) delete videoData; }
	UI16	getStreamID() { return streamID;};
	UI16	getFrameNum() { return frameNum;};
	UI8*	getVideoData() { return videoData;};
	UI32	getSize() { return size; }
	void	setStreamID(UI16 p) { streamID =p;};
	void	setFrameNum(UI16 p) { frameNum =p;};
	void	setVideoData(UI8* p) { videoData =p;};
	void	setSize(UI32 p) { size = p; }
private:
	UI16	streamID;
	UI16	frameNum;
	UI8*	videoData;
	UI32	size;
};

/**
 * a script limit tag
 **/
class ScriptLimitsTag : public BaseTag {
public:
	ScriptLimitsTag() { setType(stagScriptLimits); }
	~ScriptLimitsTag() { }
	UI16	getMaxRecursionDepth() { return maxRecursionDepth; };
	UI16	getScriptTimeoutSeconds() { return scriptTimeoutSeconds; };
	void	setMaxRecursionDepth(UI16 p) { maxRecursionDepth =p; };
	void	setScriptTimeoutSeconds(UI16 p) { scriptTimeoutSeconds =p; };
private:
	UI16	maxRecursionDepth;
	UI16	scriptTimeoutSeconds;
};


#endif
